-----------------------------------------------------------------------------
--
--		Super Character Rig
--		FK IK Arm Setup
--
-----------------------------------------------------------------------------
global SSR_part

struct SSR_part
(
	function doPart charName extParams:#("R", true, 4, 3) boneSize:1 debug:false=
	(
		-- External parameters
		local side = extParams[1] 
		local doSoft = 1 -- Should be optional, slows down the rig?
		local doTwist = extParams[2]
		local numTwistLowArm = extParams[3]
		local numTwistUpArm = extParams[4]
		
		local helperList = SSR_loc.collectLocators charName ("*Locator_Arm_" + side +"_*")
		local posClavicle, posShoulder, posElbow, posHand = undefined
		local armConTop
		local prefix = (charName + "_" + side + "_arm_")
		
		-- Look for helper objects that give locations to arm bone positions
		if helperList != false then for h in helperList do
		(
			if matchPattern h.name pattern:("*_Arm_" + side + "_Clavicle*") == true then posClavicle = h.pos
			if matchPattern h.name pattern:("*_Arm_" + side + "_Shoulder*") == true then posShoulder = h.pos
			if matchPattern h.name pattern:("*_Arm_" + side + "_Elbow*") == true then posElbow = h.pos
			if matchPattern h.name pattern:("*_Arm_" + side + "_Hand*") == true then posHand = h.pos
		)
		
		-- If one of these helpers are not found, we will exit
		if posClavicle == undefined or posShoulder == undefined or posElbow == undefined \
			or posHand == undefined \
		then (
			SSR_gen.verbose ("Missing locators for " + side as string + " arm.")
			return false
		) else (
			SSR_gen.verbose ("All locators are found, now creating " + side as string + " arm.")
		)
		
	-----------------------------------------------------------------------------
	--
	--		Create FK and IK system
	--
	-----------------------------------------------------------------------------
	
		-- Bone zAxis will be perpendicular to the bone plane
		local vecToHand = posHand - posElbow
		local vecToElbow = posElbow - posShoulder
		local zAxis = -cross (normalize vecToHand) (normalize vecToElbow)
		
		-- returns #( #(boneStart, boneEnd),#(IK_root, IK_target, midMover), limit, #(midLookat, midStartEnd, midNull) )
		local IKarm = SSR_rig.IKSetup prefix "upArm" "elbow" "lowArm" posShoulder posElbow posHand boneSize:boneSize zAxis:zAxis debug:debug 
		
		local elbowPinControl = IKarm[1][1]
		local elbowLookatControl = IKarm[1][2]
		local boneUpArm_IK 	= IKarm[2][1]
		local boneLowArm_IK = IKarm[2][2]
		local IKroot 		= IKarm[3][1]
		local IKend 		= IKarm[3][2]
		local softIKHelper 	= IKarm[4]
		local IKlookatTop	= IKarm[5][4]
		
		-- Rename the IK objects
		elbowPinControl.name	= charName + "_" + side + "_elbow_Pin_CTL"
		elbowLookatControl.name	= charName + "_" + side + "_elbow_IK_CTL"
		boneUpArm_IK.name 		= (prefix + "IKBone_UpperArm")
		boneLowArm_IK.name 		= (prefix + "IKBone_LowerArm")
		
		--! seems like the lower arm bone need a different upnode alignment when created individually. 
		--! Beware, figure out why
		
		-- The FK arm system
		local flipSide = false
		if side == "L" then flipSide = true
		local FKarm = SSR_rig.FKSetup prefix "UpArm" "Elbow" "LowArm" posShoulder posElbow posHand boneSize:boneSize zAxis:zAxis flipSide:flipSide debug:debug 
		
		local shoulderControl 	= FKarm[1][1]
		local elbowFKControl 	= FKarm[1][2]
		local boneUpArm_FK 		= FKarm[2][1]
		local boneLowArm_FK 	= FKarm[2][2]
		
		-- Rename the FK objects
		shoulderControl.name = charName + "_" + side + "_shoulder_CTL"
		elbowFKControl.name = charName + "_" + side + "_elbow_CTL"
		boneUpArm_FK.name 	= (prefix + "FKBone_UpperArm")
		boneLowArm_FK.name	= (prefix + "FKBone_LowerArm")
		
		
	-----------------------------------------------------------------------------
	--
	--		Create armbones and link them up
	--
	-----------------------------------------------------------------------------
		
		-- Create our main bones
		local boneClavicle	= SSR_rig.makeBone (prefix + "skinBone_Clavicle") posClavicle posShoulder [0, 0, 1] boneSize
		local boneUpArm		= SSR_rig.makeBone (prefix + "skinBone_UpperArm") posShoulder posElbow zAxis boneSize
		local boneLowArm	= SSR_rig.makeBone (prefix + "skinBone_LowerArm") posElbow posHand zAxis boneSize
		local boneLowArmEnd = SSR_rig.makeBone (prefix + "bone_LowerArmEnd") posHand (posHand + [0, 0, boneSize]) [1,0,0] boneSize
		
		local boneHand		= SSR_rig.makeBone (prefix + "skinBone_Hand") posHand (posHand + [0, 0, boneSize*3]) [1,0,0] boneSize
		local boneHandEnd	= SSR_rig.makeBone (prefix + "bone_HandEnd") (posHand + [0, 0, boneSize*3]) (posHand + [0, 0, boneSize*4]) [1,0,0] boneSize
		
		-- Alignments and parenting
		SSR_rig.objAlign boneLowArmEnd boneLowArm
		
		boneLowArmEnd.parent = boneLowArm
		boneLowArm.parent = boneUpArm
		boneUpArm.parent = boneClavicle
		
		-- Hand bone parenting
		boneHandEnd.parent = boneHand
		SSR_rig.objAlign boneHand boneLowArm
		boneHand.parent = boneLowArm

	-----------------------------------------------------------------------------
	--
	--		Controller Objects
	--
	-----------------------------------------------------------------------------
		
		-- Hand position controller, will also be used as the IK controller
		local handIKControl = SSR_rig.crossCTL posHand (boneSize*10) alignAxis:"X"
		handIKControl.name = charName + "_" + side + "_hand_IK_CTL"
		setTransformLockFlags handIKControl #{7..9}
		
		IKend.name = prefix + "IK"
		IKend.parent = handIKControl
		
		-- Hand rotation controller, this will also receive the finger controls
		-- Perhaps the L side should be flipped?
		local handFKControl = SSR_rig.ringCTL posHand (boneSize*10) alignAxis:"X"
		handFKControl.name = charName + "_" + side + "_hand_FK_CTL"
		SSR_rig.objAlign handFKControl boneLowArm
		setTransformLockFlags handFKControl #{1..3,7..9}
		
		-- Clavicle controller, independent of IK and FK
		-- The clavicle control must be reversed on the other side of the rig
		local posInv
		if side == "R" then posInv = (1) else posInv = (-1)
		
		local clavicleControl = SSR_rig.pinCTL posClavicle (posInv*boneSize*10) (boneSize*2)
		clavicleControl.name = charName + "_" + side + "_clavicle_CTL"
		SSR_rig.objAlign clavicleControl boneClavicle
		setTransformLockFlags clavicleControl #{1..3,7..9}
		
		
	-----------------------------------------------------------------------------
	--
	--		FK parenting and IK 
	--
	-----------------------------------------------------------------------------	

		-- Arm root helper, this should be parented to the main rig
		local armConTop = SSR_rig.makeHelper [0, posClavicle.y, posClavicle.z] boneSize (prefix + "RootCon") cross:true
		boneClavicle.parent = clavicleControl
		clavicleControl.parent = armConTop
		IKlookatTop.parent = armConTop
		
		-- Arm root, this node will be parented to the main rig
		local armTop = SSR_rig.makeHelper [0, posClavicle.y, posClavicle.z] (boneSize*2) (prefix + "Root") box:true
		SSR_rig.parentCon armConTop armTop 

		-- Wedge the FK controllers into the FK hierarchy
		handFKControl.parent = boneLowArm
		
		-- Parent up our IK and FK system
		IKroot.parent = clavicleControl
		shoulderControl.parent.parent = boneClavicle
		handIKControl.parent = boneLowArm
		
		
		
	-----------------------------------------------------------------------------
	--
	--		Position local/world constraints for hand IK
	--
	-----------------------------------------------------------------------------

		-- Helpers that will be aligned with local or world
		local handWorld = SSR_rig.makeHelper posHand boneSize (prefix + "handConstraint_World") box:true
		local handLocal = SSR_rig.makeHelper posHand (boneSize*1.5) (prefix + "handConstraint_Local") box:true
		
		SSR_rig.objAlign handWorld handIKControl
		SSR_rig.objAlign handLocal handIKControl
		
		handLocal.parent = armConTop
		
		-- Nulls that the controllers will be parented to, and that will be constrained
		local handNull = SSR_rig.makeHelper posHand boneSize (prefix + "handConstraint") cross:true
		handIKControl.parent = handNull
		handNull.parent = armConTop
		
		-- The constraint controllers
		local handCon = SSR_rig.posConTwo handNull handWorld handLocal
		local handRotCon = SSR_rig.rotConTwo handNull handWorld handLocal
		
	-----------------------------------------------------------------------------
	--	
	--		Rotation local/world constraints for shoulder FK
	--
	-----------------------------------------------------------------------------	

		local shoulderWorld = SSR_rig.makeHelper shoulderControl.pos (boneSize) (prefix + "shoulderConstraint_World") box:true
		local shoulderLocal = SSR_rig.makeHelper shoulderControl.pos (boneSize*1.5) (prefix + "shoulderConstraint_Local") box:true
		
		SSR_rig.objAlign shoulderWorld shoulderControl
		SSR_rig.objAlign shoulderLocal shoulderControl
		
		shoulderLocal.parent = boneClavicle
		
		local shoulderNull = SSR_rig.makeHelper shoulderControl.pos boneSize (prefix + "shoulderConstraint") cross:true
		SSR_rig.objAlign shoulderNull shoulderControl
		shoulderControl.parent = shoulderNull
		shoulderNull.parent = boneClavicle	
		
		local shoulderRotCon = SSR_rig.rotConTwo shoulderNull shoulderWorld shoulderLocal
		
	-----------------------------------------------------------------------------
	--
	--		Hand orientation Controller and constraints
	--
	-----------------------------------------------------------------------------

		local handRotHelper = SSR_rig.makeHelper posHand (boneSize*2.5) (prefix + "handRotate_Constraint") box:true
		SSR_rig.objAlign handRotHelper boneHand
		
		local handRotWorld = SSR_rig.makeHelper posHand (boneSize*3) (prefix + "handRotate_World") box:true
		SSR_rig.objAlign handRotWorld boneHand
		handRotWorld.parent = handWorld
		
		handOrientCon = SSR_rig.rotConTwo handRotHelper boneLowArm handRotWorld
		
		handRotHelper.parent = boneLowArmEnd --boneLowArm
		handFKControl.parent = handRotHelper
		--local handRotCon = SSR_rig.rotCon boneHand handFKControl
		boneHand.parent = handFKControl
		
		-- Seems like the hand might work better with a different axis order
		handFKControl.rotation.controller.axisOrder = 3
		
	-----------------------------------------------------------------------------
	--
	--		FK/IK switch and constraints controllers
	--
	-----------------------------------------------------------------------------	
		
		local shoulderSwitchCon = SSR_rig.rotConTwo boneUpArm boneUpArm_IK boneUpArm_FK
		local elbowSwitchCon= SSR_rig.rotConTwo boneLowArm boneLowArm_IK boneLowArm_FK
		
		shoulderSwitchCon.weight[1].controller = Bezier_Float()
		elbowSwitchCon.weight[1].controller = Bezier_Float()
		
		-- Expose helper to measure the IK upper arm length
		local exposeIKUpArm = exposeTM pos:posElbow size:(boneSize*3) name:(prefix + "IK_exposeUpArm")
		exposeIKUpArm.parent = boneUpArm_IK
		exposeIKUpArm.useParent = false
		exposeIKUpArm.localReferenceNode = boneLowArm_IK
		exposeIKUpArm.exposeNode = boneUpArm_IK
		
		-- Expose helper to measure the FK upper arm length
		local exposeFKUpArm = exposeTM pos:posElbow size:(boneSize*3) name:(prefix + "FK_exposeUpArm")
		exposeFKUpArm.parent = boneUpArm_FK
		exposeFKUpArm.useParent = false
		exposeFKUpArm.localReferenceNode = boneLowArm_FK
		exposeFKUpArm.exposeNode = boneUpArm_FK
		
		-- Expression controller to adjust the skinned upper arm length
		local UpArmStretchCon = Float_Expression()
		boneLowArm.Position.controller.X_position.controller = UpArmStretchCon
		UpArmStretchCon.AddScalarTarget "ik_len" exposeIKUpArm.distance.controller
		UpArmStretchCon.AddScalarTarget "fk_len" exposeFKUpArm.distance.controller
		UpArmStretchCon.AddScalarTarget "weight" boneUpArm.rotation.controller.weight[1].controller
		UpArmStretchCon.SetExpression "(ik_len*(weight*0.01))+(fk_len*((100-weight)*0.01))"
		
		-- Expose helper to measure the IK lower arm length
		local exposeIKLowArm = exposeTM pos:posElbow size:(boneSize*3) name:(prefix + "IK_exposeLowArm")
		exposeIKLowArm.parent = boneLowArm_IK
		exposeIKLowArm.useParent = false
		exposeIKLowArm.localReferenceNode = (SSR_rig.getChildBone boneLowArm_IK)
		exposeIKLowArm.exposeNode = boneLowArm_IK
		
		-- Expose helper to measure the FK lower arm length
		local exposeFKLowArm = exposeTM pos:posElbow size:(boneSize*3) name:(prefix + "FK_exposeLowArm")
		exposeFKLowArm.parent = boneLowArm_FK
		exposeFKLowArm.useParent = false
		exposeFKLowArm.localReferenceNode = (SSR_rig.getChildBone boneLowArm_FK)
		exposeFKLowArm.exposeNode = boneLowArm_FK -- Illegal self ref?
		
		-- Expression controller to adjust the skinned UpArm length
		local LowArmStretchCon = Float_Expression()
		boneLowArmEnd.Position.controller.X_position.controller = LowArmStretchCon
		LowArmStretchCon.AddScalarTarget "ik_len" exposeIKLowArm.distance.controller
		LowArmStretchCon.AddScalarTarget "fk_len" exposeFKLowArm.distance.controller
		LowArmStretchCon.AddScalarTarget "weight" boneLowArm.rotation.controller.weight[1].controller
		LowArmStretchCon.SetExpression "(ik_len*(weight*0.01))+(fk_len*((100-weight)*0.01))"

	-----------------------------------------------------------------------------
	--
	--		Controller custom attributes
	--
	-----------------------------------------------------------------------------
		
		-- Arm FK/IK switch attributes, assigned to most of the FK and IK controllers
		local ca_IK = attributes FKIK
		(
			parameters params rollout:armfkik
			(
				FK_IK_switch 	type:#float ui:(armIKSp, armIKSl)			
			)

			rollout armfkik "Arm FK IK Switch"
			(
				label lbl02 	"Arm set to FK <> IK" align:#centre offset:[0,5]
				spinner armIKSp	"" range:[0, 100, 0] fieldwidth:40 align:#left offset:[10, 10] across:2
				slider armIKSl 	"" range:[0, 100, 0]			
			)
		)
		
		-- Xform to scale the FK and IK controllers up and down, depending on which one is weighted
		local fkScale = xform name:"fk_scale"
		local ikScale = xform name:"ik_scale"
		
		--	Hand attributes
		local ca_H = attributes transforms
		(
			parameters params rollout:handR
			(
				X_Rot		type:#angle ui:RXSp
				Y_Rot		type:#angle ui:RYSp
				Z_Rot		type:#angle ui:RZSp
				handRot_WL 	type:#float ui:(handRotSp, handRotSl)
			)
			rollout handR "Hand Rotation"
			(
				spinner RXSp 	"Rotate X " 	range:[-1000000, 1000000, 0] align:#right
				spinner RYSp 	"Rotate Y " 	range:[-1000000, 1000000, 0] align:#right
				spinner RZSp 	"Rotate Z " 	range:[-1000000, 1000000, 0] align:#right
				label lbl02 		"Hand orients Arm <> World" align:#centre offset:[0,10]
				spinner handRotSp 	"" range:[0, 100, 0] fieldwidth:40 align:#left offset:[10, 10] across:2
				slider handRotSl 	"" range:[0, 100, 0]
			)
		)
		
		--	Hand IK attributes
		local ca_H_IK = attributes IK_transforms
		(
			parameters params rollout:handP
			(
				X_Pos		type:#float ui:PXSp
				Y_Pos		type:#float ui:PYSp
				Z_Pos		type:#float ui:PZSp
				handPos_WL	type:#float ui:(handSp, handSl)
				soft_IK		type:#float ui:softSp
			)		
			rollout handP "Hand IK Transforms"
			(
				spinner PXSp 		"Translate X " 	range:[-1000000, 1000000, 0] align:#right
				spinner PYSp 		"Translate Y " 	range:[-1000000, 1000000, 0] align:#right
				spinner PZSp 		"Translate Z " 	range:[-1000000, 1000000, 0] align:#right
				spinner softSp		"Soft IK "		range:[0, 100000, 0] offset:[0,10]
				label lbl01 		"Hand follows Chest <> World" align:#centre offset:[0,10]
				spinner handSp 		"" range:[0, 100, 0] fieldwidth:40 align:#left offset:[10, 10] across:2
				slider handSl 		"" range:[0, 100, 0]
				
			)
		)
		
		-- Add custom attributes to the hand controller
		local penMod_H = 'PEN_Attribute_Holder 2'()
		addModifier handFKControl penMod_H
		custAttributes.add penMod_H ca_H
		custAttributes.add penMod_H ca_IK
		
		-- Add custom attributes to the hand IK controller
		local penMod_H_IK = 'PEN_Attribute_Holder 2'()
		addModifier handIKControl ikScale
		addModifier handIKControl penMod_H_IK
		custAttributes.add penMod_H_IK ca_H_IK
		custAttributes.add penMod_H_IK ca_IK
		
		-- FK hand connections
		penMod_H.transforms.X_Rot.controller = \
			handFKControl.rotation.controller.X_rotation.controller
		penMod_H.transforms.Y_Rot.controller = \
			handFKControl.rotation.controller.Y_rotation.controller
		penMod_H.transforms.Z_Rot.controller = \
			handFKControl.rotation.controller.Z_rotation.controller
			
		-- Important, the controller connection must be made before the wiring
		penMod_H.transforms.handRot_WL.controller = bezier_float()
		
		paramWire.connect penMod_H.transforms[#handRot_WL] \
			handRotHelper.rotation.controller[#Orientation_Weight_0] \
			"100-handRot_WL"
		paramWire.connect penMod_H.transforms[#handRot_WL] \
			handRotHelper.rotation.controller[#Orientation_Weight_1] \
			"handRot_WL"		

		-- IK hand connections
		penMod_H_IK.IK_transforms.X_Pos.controller = \
			handIKControl.pos.controller.X_Position.controller
		penMod_H_IK.IK_transforms.Y_Pos.controller = \
			handIKControl.pos.controller.Y_Position.controller	
		penMod_H_IK.IK_transforms.Z_Pos.controller = \
			handIKControl.pos.controller.Z_Position.controller
			
		
		paramWire.connect penMod_H_IK.IK_transforms[#handPos_WL] \
			handNull.position.controller[#Position_Weight_0] \
			"handPos_WL"
		paramWire.connect penMod_H_IK.IK_transforms[#handPos_WL] \
			handNull.position.controller[#Position_Weight_1] \
			"100-handPos_WL"
		
		paramWire.connect penMod_H_IK.IK_transforms[#handPos_WL] \
			handNull.rotation.controller[#Orientation_Weight_0] \
			"handPos_WL"
		paramWire.connect penMod_H_IK.IK_transforms[#handPos_WL] \
			handNull.rotation.controller[#Orientation_Weight_1] \
			"100-handPos_WL"
			
		-- Soft IK
		penMod_H_IK.IK_transforms.soft_IK.controller = \
			softIKHelper[#Limits][#Upper_Smoothing].controller = \
			Bezier_float()
			
		-- Clavicle attributes
		local ca_C = attributes transforms
		(
			parameters params rollout:elbow
			(
				X_Rot		type:#angle ui:RXSp
				Y_Rot		type:#angle ui:RYSp
				Z_Rot		type:#angle ui:RZSp
			)
			rollout elbow "Clavicle Transforms"
			(
				spinner RXSp 	"Rotation X " 	range:[-1000000, 1000000, 0]
				spinner RYSp 	"Rotation Y " 	range:[-1000000, 1000000, 0]
				spinner RZSp 	"Rotation Z " 	range:[-1000000, 1000000, 0]
			)
		)
		
		-- Add custom attributes to the clavicle controller
		local penMod_C = 'PEN_Attribute_Holder 2'()
		addModifier clavicleControl penMod_C
		custAttributes.add penMod_C ca_C
		
		-- Add scale modifier to the elbow lookat controller
		addModifier elbowLookatControl ikScale before:1
		
		-- Add scale modifier to the elbow pin controller
		addModifier elbowPinControl ikScale before:1
		
		-- Add custom attributes to the elbow controller
		local penMod_E = elbowFKControl.modifiers[1]
		addModifier elbowFKControl fkScale before:1
		custAttributes.add penMod_E ca_IK
		
		--	Shoulder attributes
		local ca_S = attributes constraints
		(
			parameters params rollout:shoulder
			(
				shoulderRot_WL 	type:#float ui:(shoulderRotSp, shoulderRotSl)
			)
			
			rollout shoulder "Shoulder Constraints"
			(
				label lbl02 		"Shoulder orients Clavicle <> World" align:#centre offset:[0,10]
				spinner shoulderRotSp 	"" range:[0, 100, 0] fieldwidth:40 align:#left offset:[10, 10] across:2
				slider shoulderRotSl 	"" range:[0, 100, 0]
			)
		)
		
		-- Add custom attributes to the shoulder controller
		local penMod_S = shoulderControl.modifiers[1]
		addModifier shoulderControl fkScale before:1
		custAttributes.add penMod_S ca_S
		custAttributes.add penMod_S ca_IK
		
		
	-----------------------------------------------------------------------------
	--
	--		Parameter wiring
	--
	-----------------------------------------------------------------------------

		-- Shoulder connections
		paramWire.connect penMod_S.constraints[#shoulderRot_WL] \
			shoulderNull.rotation.controller[#Orientation_Weight_0] \
			"shoulderRot_WL"
		paramWire.connect penMod_S.constraints[#shoulderRot_WL] \
			shoulderNull.rotation.controller[#Orientation_Weight_1] \
			"100-shoulderRot_WL"
		
		-- Clavicle connections
		-- Freeze the clavicle transforms with a little helper
		local storeParent = clavicleControl.parent
		local clavicleNull = SSR_rig.makeHelper posClavicle (boneSize*0.5) (prefix + "clavicleNull") cross:true
		SSR_rig.objAlign clavicleNull clavicleControl
		clavicleNull.parent = clavicleControl.parent
		clavicleControl.parent = clavicleNull

		penMod_C.transforms.X_Rot.controller = \
			clavicleControl.rotation.controller.X_Rotation.controller
		penMod_C.transforms.Y_Rot.controller = \
			clavicleControl.rotation.controller.Y_Rotation.controller
		penMod_C.transforms.Z_Rot.controller = \
			clavicleControl.rotation.controller.Z_Rotation.controller
		
		-- IK connections
		penMod_H_IK.FKIK.FK_IK_switch.controller = \
			penMod_E.FKIK.FK_IK_switch.controller = \
			penMod_H.FKIK.FK_IK_switch.controller = \
			penMod_S.FKIK.FK_IK_switch.controller = bezier_float()
		
		paramWire.connect penMod_H_IK.FKIK[#FK_IK_switch] \
			boneUpArm.rotation.controller[#Orientation_Weight_0] \
			"FK_IK_switch"		
		paramWire.connect penMod_H_IK.FKIK[#FK_IK_switch] \
			boneUpArm.rotation.controller[#Orientation_Weight_1] \
			"100-FK_IK_switch"
			
		paramWire.connect penMod_H_IK.FKIK[#FK_IK_switch] \
			boneLowArm.rotation.controller[#Orientation_Weight_0] \
			"FK_IK_switch"
		paramWire.connect penMod_H_IK.FKIK[#FK_IK_switch] \
			boneLowArm.rotation.controller[#Orientation_Weight_1] \
			"100-FK_IK_switch"
		
		-- Xforms are used to scale down the FK and IK controls when not used
		paramWire.connect penMod_H_IK.FKIK[#FK_IK_switch] \
			fkScale.Gizmo.controller[#Scale] \
			"[1,1,1]-[FK_IK_switch, FK_IK_switch, FK_IK_switch]*0.01"
		paramWire.connect penMod_H_IK.FKIK[#FK_IK_switch] \
			ikScale.Gizmo.controller[#Scale] \
			"[FK_IK_switch, FK_IK_switch, FK_IK_switch]*0.01"
				
		-- Null out a couple of the attributes, they have some weird offset
		shoulderControl.modifiers[1].transforms.X_Rot = 0
		shoulderControl.modifiers[1].transforms.Y_Rot = 0
		shoulderControl.modifiers[1].transforms.Z_Rot = 0


	-----------------------------------------------------------------------------
	--
	--		Optional Twist Bones
	--
	-----------------------------------------------------------------------------

		if doTwist == true then
		(
			twistReturn = SSR_rig.stretchTwistBones boneLowArm boneHand numTwistLowArm twistUp:false debug:debug
			
			twistReturn = SSR_rig.stretchTwistBones boneUpArm boneClavicle numTwistUpArm twistUp:true debug:debug

			boneUpArm.name = prefix + "bone_UpperArm"
			boneLowArm.name = prefix + "bone_LowerArm"
		)
		
		return #(handIKControl, handFKControl, elbowFKControl, elbowPinControl, elbowLookatControl, shoulderControl, clavicleControl)
	)
)

SSR_part = SSR_part()