-----------------------------------------------------------------------------
--
--		Super Character Rig
--		IK Arm Setup
--
--		v1.0: Basic setup
--		v1.1: World/arm constraint to hand
--		v1.2: New orientation constraints in hand
--		v1.3: Failed version
--		v1.4: Back to previous arm
--		v1.5: Stretchy bones
--		v1.6: Twist bones
--		v1.7: Soft IK
--		v1.8: Named all objects and put them into layers
--		v1.9: Moving hand rotations to separate controller
--		v2.0: Renamed attributes to match FK/IK setup
--		v2.1: List controllers for hand rotations
--
-----------------------------------------------------------------------------
global SSR_part

struct SSR_part
(
	function doPart charName extParams:#("R", true, 4, 3) boneSize:1 =
	(
		-- External parameters
		local side = extParams[1] 
		local doSoft = 1 -- Should be optional, slows down the rig?
		local doTwist = extParams[2]
		local numTwistLowArm = extParams[3]
		local numTwistUpArm = extParams[4]
		
		local helperList = SSR_loc.collectLocators charName ("*Locator_Arm_" + side +"_*")
		local posClavicle, posShoulder, posElbow, posHand = undefined
		local armConTop
		local prefix = (charName + "_" + side + "_arm_")

		-- Look for helper objects that give locations to arm bone positions
		if helperList != false then for h in helperList do
		(
			if matchPattern h.name pattern:("*_Arm_" + side + "_Clavicle*") == true then posClavicle = h.pos
			if matchPattern h.name pattern:("*_Arm_" + side + "_Shoulder*") == true then posShoulder = h.pos
			if matchPattern h.name pattern:("*_Arm_" + side + "_Elbow*") == true then posElbow = h.pos
			if matchPattern h.name pattern:("*_Arm_" + side + "_Hand*") == true then posHand = h.pos
		)
		
		-- If one of these helpers are not found, we will exit
		if posClavicle == undefined or posShoulder == undefined or posElbow == undefined \
			or posHand == undefined \
		then (
			SSR_gen.verbose ("Missing locators for " + side as string + " arm.")
			return false
		) else (
			SSR_gen.verbose ("All locators are found, now creating " + side as string + " arm.")
		)

		
	-----------------------------------------------------------------------------
	--
	--		Create armbones and link them up
	--
	-----------------------------------------------------------------------------
		
		-- Bone zAxis will be perpendicular to the bone plane
		local vecToHand = posHand - posElbow
		local vecToElbow = posElbow - posShoulder
		local zAxis = -cross (normalize vecToHand) (normalize vecToElbow)
		
		-- Create our main bones
		local boneClavicle	= SSR_rig.makeBone (prefix + "skinBone_Clavicle") posClavicle posShoulder [0, 0, 1] boneSize
		local boneUpArm		= SSR_rig.makeBone (prefix + "skinBone_UpperArm") posShoulder posElbow zAxis boneSize
		local boneLowArm	= SSR_rig.makeBone (prefix + "skinBone_LowerArm") posElbow posHand zAxis boneSize
		local boneLowArmEnd = SSR_rig.makeBone (prefix + "bone_LowerArmEnd") posHand (posHand + [0, 0, boneSize]) [1,0,0] boneSize
		
		local boneHand		= SSR_rig.makeBone (prefix + "skinBone_Hand") posHand (posHand + [0, 0, boneSize*3]) [1,0,0] boneSize
		local boneHandEnd	= SSR_rig.makeBone (prefix + "bone_HandEnd") (posHand + [0, 0, boneSize*3]) (posHand + [0, 0, boneSize*4]) [1,0,0] boneSize
		
		-- Alignments and parenting
		SSR_rig.objAlign boneLowArmEnd boneLowArm
		
		boneLowArmEnd.parent = boneLowArm
		boneLowArm.parent = boneUpArm
		boneUpArm.parent = boneClavicle
		
		boneHandEnd.parent = boneHand
		SSR_rig.objAlign boneHand boneLowArm
		
		armConTop = SSR_rig.makeHelper [0, posClavicle.y, posClavicle.z] boneSize (prefix + "Root") cross:true

	-----------------------------------------------------------------------------
	--
	--		Control Objects and IK
	--
	-----------------------------------------------------------------------------
		
		-- Hand position controller, will also be used as the IK controller
		local handControl = SSR_rig.crossCTL posHand (boneSize*10) alignAxis:"X"
		handControl.name = charName + "_" + side + "_hand_IK_CTL"
		setTransformLockFlags handControl #{4..9}
		
		-- Clavicle controller, independent of IK and FK
		-- The clavicle control must be reversed on the other side of the rig
		local posInv
		if side == "R" then posInv = (1)
		else posInv = (-1)
			
		local clavicleControl = SSR_rig.pinCTL posClavicle (posInv*boneSize*10) (boneSize*2)
		clavicleControl.name = charName + "_" + side + "_clavicle_CTL"
		SSR_rig.objAlign clavicleControl boneClavicle
		setTransformLockFlags clavicleControl #{1..3,7..9}
		
		-- Elbow lookat controller, will be used as the IK lookat target
		local elbowLookatControl = SSR_rig.crossCTL posElbow (boneSize*6)	
		elbowLookatControl.name = charName + "_" + side + "_elbow_IK_CTL"
		in coordsys boneLowArm move elbowLookatControl [0,-(distance posHand posShoulder),0]
		setTransformLockFlags elbowLookatControl #{4..9}
		
		boneClavicle.parent = clavicleControl
		boneHand.parent = boneLowArmEnd
		
		-- IK for the hand, with adjustable hand constraining
		local armIK = IKSys.ikChain boneUpArm boneLowArmEnd "IKHISolver"
		armIK.transform.controller.goalSize = boneSize * 2
		armIK.transform.controller.VHTarget = elbowLookatControl
		armIK.name = prefix + "IK"
		armIK.parent = handControl
		boneHand.parent = handControl
		
		-- Helpers and constraints for our hand and elbow constraints
		local handWorld = SSR_rig.makeHelper posHand boneSize (prefix + "handConstraint_World") cross:true
		local handLocal = SSR_rig.makeHelper posHand (boneSize*1.5) (prefix + "handConstraint_Local") cross:true
		local elbowWorld = SSR_rig.makeHelper elbowLookatControl.pos (boneSize) (prefix + "elbowConstraint_World") cross:true
		local elbowLocal = SSR_rig.makeHelper elbowLookatControl.pos (boneSize*1.5) (prefix + "elbowConstraint_Local") cross:true
		SSR_rig.objAlign handWorld handControl
		SSR_rig.objAlign handLocal handControl
		
		handLocal.parent = armConTop
		elbowLocal.parent = armConTop
		clavicleControl.parent = armConTop
		
		local handNull = SSR_rig.makeHelper posHand boneSize (prefix + "handConstraint") cross:true
		handControl.parent = handNull
		handNull.parent = armConTop
		local elbowNull = SSR_rig.makeHelper elbowLookatControl.pos boneSize (prefix + "elbowConstraint") cross:true
		elbowLookatControl.parent = elbowNull
		elbowNull.parent = armConTop
		
		local elbowCon = SSR_rig.posConTwo elbowNull elbowWorld elbowLocal
		local elbowRotCon = SSR_rig.rotConTwo elbowNull elbowWorld elbowLocal
		local handCon = SSR_rig.posConTwo handNull handWorld handLocal
		local handRotCon = SSR_rig.rotConTwo handNull handWorld handLocal
		
	-----------------------------------------------------------------------------
	--
	--		Hand orientation Controller and constraints
	--
	-----------------------------------------------------------------------------
		handRotControl = SSR_rig.ringCTL posHand (boneSize*10) alignAxis:"X"
		handRotControl.name = charName + "_" + side + "_hand_FK_CTL"
		SSR_rig.objAlign handRotControl boneHand
		
		handRotHelper = SSR_rig.makeHelper posHand (boneSize*1.5) (prefix + "handRotate_Constraint") box:true
		SSR_rig.objAlign handRotHelper boneHand
		
		handRotWorld = SSR_rig.makeHelper posHand (boneSize*2) (prefix + "handRotate_World") box:true
		SSR_rig.objAlign handRotWorld boneHand
		handRotWorld.parent = handWorld
		
		handOrientCon = SSR_rig.rotConTwo handRotHelper boneLowArm handRotWorld
		
		handRotHelper.parent = handControl
		boneHand.parent = handRotControl
		handRotControl.parent = handRotHelper
		
		-- Seems like the hand might work better with a different axis order
		handRotControl.rotation.controller.axisOrder = 3
		
		-- We will add an attribute holder to the rotation controller, used to hold fingers
		local penFingMod = 'PEN_Attribute_Holder 2'()
		addModifier handRotControl penFingMod
		setTransformLockFlags handRotControl #{1..3,7..9}
		
	-----------------------------------------------------------------------------
	--
	--		Controller custom attributes
	--
	-----------------------------------------------------------------------------
		
		-- Hand attributes
		local ca_H = attributes hand_transforms
		(
			parameters params rollout:handR
			(
				RotX		type:#angle ui:RXSp
				RotY		type:#angle ui:RYSp
				RotZ		type:#angle ui:RZSp
				handRot_WL 	type:#float ui:(handRotSp, handRotSl)
			)
			rollout handR "Hand Rotation"
			(
				spinner RXSp 	"Rotate X " 	range:[-1000000, 1000000, 0] align:#right
				spinner RYSp 	"Rotate Y " 	range:[-1000000, 1000000, 0] align:#right
				spinner RZSp 	"Rotate Z " 	range:[-1000000, 1000000, 0] align:#right
				label lbl02 		"Hand orients Arm <> World" align:#centre offset:[0,10]
				spinner handRotSp 	"" range:[0, 100, 0] fieldwidth:40 align:#left offset:[10, 10] across:2
				slider handRotSl 	"" range:[0, 100, 0]
			)
		)
		
		-- Hand IK attributes
		local ca_H_IK = attributes hand_IK_transforms
		(
			parameters params rollout:handP
			(
				PosX		type:#float ui:PXSp
				PosY		type:#float ui:PYSp
				PosZ		type:#float ui:PZSp
				handPos_WL	type:#float ui:(handSp, handSl)
				soft_IK		type:#float ui:softSp
			)		
			rollout handP "Hand IK Transforms"
			(
				spinner PXSp 	"Translate X " 	range:[-1000000, 1000000, 0] align:#right
				spinner PYSp 	"Translate Y " 	range:[-1000000, 1000000, 0] align:#right
				spinner PZSp 	"Translate Z " 	range:[-1000000, 1000000, 0] align:#right
				spinner softSp	"Soft IK "		range:[0, 100000, 0] offset:[0,10]
				label lbl01 		"Hand follows Chest <> World" align:#centre offset:[0,10]
				spinner handSp 		"" range:[0, 100, 0] fieldwidth:40 align:#left offset:[10, 10] across:2
				slider handSl 		"" range:[0, 100, 0]
				
			)
		)

		-- Elbow_Lookat attributes
		local ca_E_IK = attributes elbow_ik_transforms
		(
			parameters params rollout:elbow
			(
				PosX		type:#float ui:PXSp
				PosY		type:#float ui:PYSp
				PosZ		type:#float ui:PZSp
			)
			parameters paramsC rollout:elbowC
			(
				elbowPos_WL	type:#float ui:(elbowSp, elbowSl)			
			)
			rollout elbow "Elbow Lookat Transforms"
			(
				spinner PXSp 	"Position X " 	range:[-1000000, 1000000, 0]
				spinner PYSp 	"Position Y " 	range:[-1000000, 1000000, 0]
				spinner PZSp 	"Position Z " 	range:[-1000000, 1000000, 0]
			)
			rollout elbowC "Elbow Lookat Constraints"
			(
				label lbl02 	"Elbow follows Chest <> World" align:#centre offset:[0,5]
				spinner elbowSp	"" range:[0, 100, 0] fieldwidth:40 align:#left offset:[10, 10] across:2
				slider elbowSl 	"" range:[0, 100, 0]			
			)
		)
		
		
		-- Add custom attributes to the hand IK controller
		local penMod_H_IK = 'PEN_Attribute_Holder 2'()
		addModifier handControl penMod_H_IK
		custAttributes.add penMod_H_IK ca_H_IK
		custAttributes.add penMod_H_IK ca_H
		
		-- Add custom attributes to the hand FK controller
		local penMod_H_FK = 'PEN_Attribute_Holder 2'()
		addModifier handRotControl penMod_H_FK
		custAttributes.add penMod_H_FK ca_H

		-- Add custom attributes to the elbow lookat controller
		local penMod_E_IK = 'PEN_Attribute_Holder 2'()
		addModifier elbowLookatControl penMod_E_IK
		custAttributes.add penMod_E_IK ca_E_IK
		
	-----------------------------------------------------------------------------
	--
	--		Parameter wiring
	--
	-----------------------------------------------------------------------------

		-- FK hand connections
		penMod_H_IK.hand_transforms.RotX.controller = \
			penMod_H_FK.hand_transforms.RotX.controller = \
			handRotControl.rotation.controller.X_rotation.controller
		penMod_H_IK.hand_transforms.RotY.controller = \
			penMod_H_FK.hand_transforms.RotY.controller = \
			handRotControl.rotation.controller.Y_rotation.controller
		penMod_H_IK.hand_transforms.RotZ.controller = \
			penMod_H_FK.hand_transforms.RotZ.controller = \
			handRotControl.rotation.controller.Z_rotation.controller
			
		-- Important, the controller connection must be made before the wiring
		penMod_H_IK.hand_transforms.handRot_WL.controller = \
			penMod_H_FK.hand_transforms.handRot_WL.controller = \
			bezier_float()
		
		paramWire.connect penMod_H_IK.hand_transforms[#handRot_WL] \
			handRotHelper.rotation.controller[#Orientation_Weight_0] \
			"100-handRot_WL"
		paramWire.connect penMod_H_IK.hand_transforms[#handRot_WL] \
			handRotHelper.rotation.controller[#Orientation_Weight_1] \
			"handRot_WL"		

		-- IK hand connections
		penMod_H_IK.hand_IK_transforms.PosX.controller = \
			handControl.pos.controller.X_Position.controller
		penMod_H_IK.hand_IK_transforms.PosY.controller = \
			handControl.pos.controller.Y_Position.controller	
		penMod_H_IK.hand_IK_transforms.PosZ.controller = \
		handControl.pos.controller.Z_Position.controller
		
		paramWire.connect penMod_H_IK.hand_IK_transforms[#handPos_WL] \
			handNull.position.controller[#Position_Weight_0] \
			"handPos_WL"
		paramWire.connect penMod_H_IK.hand_IK_transforms[#handPos_WL] \
			handNull.position.controller[#Position_Weight_1] \
			"100-handPos_WL"
		
		paramWire.connect penMod_H_IK.hand_IK_transforms[#handPos_WL] \
			handNull.rotation.controller[#Orientation_Weight_0] \
			"handPos_WL"
		paramWire.connect penMod_H_IK.hand_IK_transforms[#handPos_WL] \
			handNull.rotation.controller[#Orientation_Weight_1] \
			"100-handPos_WL"
			
		-- IK Elbow connections
		--penMod_H_IK.elbow_ik_transforms.elbowPos_WL.controller = \
			penMod_E_IK.elbow_ik_transforms.elbowPos_WL.controller = bezier_float()
		
		paramWire.connect penMod_E_IK.elbow_ik_transforms[#elbowPos_WL] \
			elbowNull.position.controller[#Position_Weight_0] \
			"elbowPos_WL"
		paramWire.connect penMod_E_IK.elbow_ik_transforms[#elbowPos_WL] \
			elbowNull.position.controller[#Position_Weight_1] \
			"100-elbowPos_WL"
			
		paramWire.connect penMod_E_IK.elbow_ik_transforms[#elbowPos_WL] \
			elbowNull.rotation.controller[#Orientation_Weight_0] \
			"elbowPos_WL"
		paramWire.connect penMod_E_IK.elbow_ik_transforms[#elbowPos_WL] \
			elbowNull.rotation.controller[#Orientation_Weight_1] \
			"100-elbowPos_WL"
		

		-- Clavicle attributes
		local ca_C = attributes clavicle_transforms
		(
			parameters params rollout:elbow
			(
				RotX		type:#angle ui:RXSp
				RotY		type:#angle ui:RYSp
				RotZ		type:#angle ui:RZSp
			)
			rollout elbow "Clavicle Transforms"
			(
				spinner RXSp 	"Rotation X " 	range:[-1000000, 1000000, 0]
				spinner RYSp 	"Rotation Y " 	range:[-1000000, 1000000, 0]
				spinner RZSp 	"Rotation Z " 	range:[-1000000, 1000000, 0]
			)
		)
		
		-- Add custom attributes to the clavicle controller
		local penMod_C = 'PEN_Attribute_Holder 2'()
		addModifier clavicleControl penMod_C
		custAttributes.add penMod_C ca_C
		
		-- Clavicle connections
		-- Freeze the clavicle transforms with a little helper
		local storeParent = clavicleControl.parent
		local clavicleNull = SSR_rig.makeHelper posClavicle (boneSize*0.5) (prefix + "clavicleNull") cross:true
		SSR_rig.objAlign clavicleNull clavicleControl
		clavicleNull.parent = clavicleControl.parent
		clavicleControl.parent = clavicleNull

		penMod_C.clavicle_transforms.RotX.controller = \
			clavicleControl.rotation.controller.X_Rotation.controller
		penMod_C.clavicle_transforms.RotY.controller = \
			clavicleControl.rotation.controller.Y_Rotation.controller
		penMod_C.clavicle_transforms.RotZ.controller = \
			clavicleControl.rotation.controller.Z_Rotation.controller

		
	-----------------------------------------------------------------------------
	--
	--		Stretchy arm bones
	--
	-----------------------------------------------------------------------------
		
		-- Simple soft IK
		if doSoft == 1 then
		(
			-- The original starting length of the arm
			local lowArmLength = distance posElbow posHand
			local upArmLength = distance posShoulder posElbow
			local armLength = lowArmLength + upArmLength

			-- Lookat helper at the root
			local softLookat = SSR_rig.makeHelper posShoulder boneSize (prefix + "softLookat") cross:true
			softLookat.parent = boneClavicle
			SSR_rig.objAlign softLookat boneUpArm
			
			-- Lookat constraint to the IK controller
			local lookCon = SSR_rig.lookCon softLookat handControl
			lookCon.StoUp_axis = 1
			lookCon.upnode_axis = 1
			
			-- The helpers and expose transforms that do the soft IK
			local exposeLength = exposeTM pos:(posShoulder+[0,boneSize*4,0]) size:(boneSize*2) name:(prefix + "exposeLength")
			exposeLength.parent = softLookat
			--exposeLength.exposeNode = boneHand
			exposeLength.exposeNode = handControl
			exposeLength.useParent = false
			exposeLength.localReferenceNode = softLookat
			
			-- Soft limited point
			local softPoint = SSR_rig.makeHelper posHand (boneSize*3) (prefix + "softIK") box:true
			SSR_rig.objAlign softPoint softLookat
			softPoint.parent = softLookat
			
			-- A float_limit controller will stop the helper from moving beyond the arm length
			local limit = float_limit()
			softPoint.pos.controller.X_Position.controller = limit
			limit.SetLimitedControl exposeLength.distance.controller
			limit.upper_limit = armLength
			
			-- And link the smooth control to our hand and Elbow controller object
			penMod_H_IK.hand_IK_transforms.soft_IK.controller = \
				softPoint.pos.controller.X_Position.controller[#Limits][#Upper_Smoothing].controller = \
				Bezier_float()
			
			-- Finally, parent the IK target to our new soft helper
			armIK.parent = softPoint
			handRotHelper.parent = softPoint
		)
		
		if doSoft == 2 then
		(
			boneLowArm.boneFreezeLength = false
			boneUpArm.boneFreezeLength = false
			
			-- The original starting length of the arm
			local lowArmLength = distance posElbow posHand
			local upArmLength = distance posShoulder posElbow
			local armLength = lowArmLength + upArmLength
			
			local shoulderHelper = SSR_rig.makeHelper posShoulder boneSize (prefix + "exposeHelper") cross:true
			shoulderHelper.parent = boneClavicle
			SSR_rig.objAlign shoulderHelper boneUpArm
			
			-- The helpers and expose transforms that do the soft IK
			local exposeLength = exposeTM pos:posShoulder side:(boneSize*2) name:(prefix + "exposeLength")
			exposeLength.parent = shoulderHelper
			--exposeLength.exposeNode = boneHand
			exposeLength.exposeNode = armIK
			exposeLength.useParent = false
			exposeLength.localReferenceNode = shoulderHelper
			
			local softLookat = SSR_rig.makeHelper posShoulder (boneSize*3) (prefix + "softLookat") axistripod:true
			local lookCon = SSR_rig.lookCon softLookat handControl
			lookCon.StoUp_axis = 1
			lookCon.upnode_axis = 1
			softLookat.parent = shoulderHelper
			
			local softPoint = SSR_rig.makeHelper posHand (boneSize*3) (prefix + "softPoint") box:true
			SSR_rig.objAlign softPoint softLookat
			softPoint.parent = softLookat
			
			local limit = float_limit()
			softPoint.pos.controller.X_Position.controller = limit
			limit.SetLimitedControl exposeLength.distance.controller
			
			-- Link the float limit input to the expose length helper
			limit.upper_limit = armLength
			
			-- And link the smooth control to our Hand Controller object
			penMod_H_IK.hand_IK_transforms.soft_IK.controller = \
					softPoint.pos.controller.X_Position.controller[#Limits][#Upper_Smoothing].controller = \
					Bezier_float()
			
			-- One more expose transform to measure the distance from the smooth object to the controller
			local exposeDist = exposeTM pos:posShoulder size:(boneSize*2) name:(prefix + "exposeDistance")
			exposeDist.parent = exposeLength
			--exposeDist.exposeNode = boneHand
			exposeDist.exposeNode = armIK
			exposeDist.useParent = false
			exposeDist.localReferenceNode = softPoint
			
			-- Expression controller to set the lower arm bone length
			SSR_rig.freezeTransform boneLowArmEnd
			local lowArmController = Float_Expression ()
							
			boneLowArmEnd.transform.controller.FK_Sub_Control.controller.Position.controller.Zero_Pos_XYZ.controller.X_Position.controller = lowArmController
			-- With some math, the leg will extend matching its proportions
			lowArmController.AddScalarConstant "con" (0.01*upArmLength*(100/armLength))
			lowArmController.AddScalarTarget "dist" exposeDist.distance.controller
			lowArmController.SetExpression "(dist*con)"
			
			-- Expression controller to set the upper arm bone length
			SSR_rig.freezeTransform boneLowArm
			local upArmController = Float_Expression ()
			
			boneLowArm.transform.controller.FK_Sub_Control.controller.Position.controller.Zero_Pos_XYZ.controller.X_Position.controller = upArmController
			upArmController.AddScalarConstant "con" (0.01*lowArmLength*(100/armLength))
			upArmController.AddScalarTarget "dist" exposeDist.distance.controller
			upArmController.SetExpression "(dist*con)"
		)
		
	-----------------------------------------------------------------------------
	--
	--		Twist Bones
	--
	-----------------------------------------------------------------------------

		if doTwist == true then
		(
			twistReturn = SSR_rig.twistBones boneLowArm boneHand numTwistLowArm
			
			twistReturn = SSR_rig.twistBones boneUpArm boneClavicle numTwistUpArm twistUp:true
			if side == "L" then twistReturn[1].StoUP_axisFlip = on
				
			-- The original skinning bones should not be skinBones anymore
			boneUpArm.name = prefix + "bone_UpperArm"
			boneLowArm.name = prefix + "bone_LowerArm"
		)

		
		gc()
		return #(handControl, handRotControl, elbowLookatControl, clavicleControl)
	)
)

SSR_part = SSR_part()