-----------------------------------------------------------------------------
--
--		Standard stretching leg setup
--		
--		v1.0:	Basic setup
--		v1.1:	Reactor removed, additional attributes
--		v1.2:	Foot bank controls
--		v1.3:	New knee lookat setup
--		v1.4:	Minor fixes to parenting
--		v1.5:	kneeHip helper unconnected to thigh End, 
--			 	created errors when undoing hip translate
--		v1.6:	Stretchy soft IK, yeah baby!!
--		v1.7:	Twist bones
--		v1.8:	Fixed a bug in the stretchy leg
--			 	Exposed rotations are now instanced and not wired
--				Named all objects and put them into layers
--		v1.9:	Better orientation on leg bones
--				Local/World switch on leg
--				Controller for toe rotations
--		v2.0:	Improved (hopefully) connection point to spine
--		v3.0:	Replaced IK with modular method
--
-----------------------------------------------------------------------------
global SSR_part

struct SSR_part
(
	function doPart charName extParams:#("R", true, 3, 3) boneSize:1 debug:false =
	(
		-- External parameters
		local side = extParams[1] 
		local doSoft = 1 -- Should be optional, slows down the rig?
		local doTwist = extParams[2]
		local numTwistShin = extParams[3]
		local numTwistThigh = extParams[4]
		
		local helperList = SSR_loc.collectLocators charName ("*Locator_Leg_" + side +"_*")
		local posHip, posKnee, posAnkle, posHeel, posBall, posToe, posSideOut, posSideIn
		local legConTop
		local prefix = (charName + "_" + side + "_leg_")

		-- Look for helper objects that give locations to leg bone positions
		if helperList != false then for h in helperList do
		(
			if matchPattern h.name pattern:("*_Leg_" + side + "_Hip*") 		== true then posHip = h.pos
			if matchPattern h.name pattern:("*_Leg_" + side + "_Knee*") 	== true then posKnee = h.pos
			if matchPattern h.name pattern:("*_Leg_" + side + "_Ankle*") 	== true then posAnkle = h.pos
			if matchPattern h.name pattern:("*_Leg_" + side + "_Heel*") 	== true then posHeel = h.pos
			if matchPattern h.name pattern:("*_Leg_" + side + "_Toe*") 		== true then posBall = h.pos
			if matchPattern h.name pattern:("*_Leg_" + side + "_Tip*") 		== true then posToe = h.pos
			if matchPattern h.name pattern:("*_Leg_" + side + "_SideOut*") 	== true then posSideOut = h.pos
			if matchPattern h.name pattern:("*_Leg_" + side + "_SideIn*") 	== true then posSideIn = h.pos
		)
		
		-- If one of these helpers are not found, we will exit
		if posHip == undefined or posKnee == undefined or posAnkle == undefined \
			or posHeel == undefined or posBall == undefined or posToe == undefined \
			or posSideIn == undefined or posSideOut == undefined
		then (
			SSR_gen.verbose ("Missing locators for " + side as string + " leg.")
			return false
		) else (
			SSR_gen.verbose ("All locators are found, creating " + side as string + " leg.")
		)
		
	-----------------------------------------------------------------------------
	--
	--		Create IK System
	--
	-----------------------------------------------------------------------------
		
		-- returns #( #(boneStart, boneEnd),#(IK_root, IK_target, midMover), limit, #(midLookat, midStartEnd, midNull) )
		local IKleg = SSR_rig.IKSetup prefix "Thigh" "Knee" "Shin" posHip posKnee posAnkle boneSize:boneSize debug:debug 
		
		local boneThigh = IKleg[1][1]
		local boneShin 	= IKleg[1][2]
		local IKroot 	= IKleg[2][1]
		local IKend 	= IKleg[2][2]
		local kneePin 	= IKleg[2][3]
		local limitCon 	= IKleg[3]
		local kneeLookat = IKleg[4][1]
		local kneeHipFoot = IKleg[4][2]
		local kneeWorldNull = IKleg[4][3]
		
	-----------------------------------------------------------------------------
	--
	--		Create Foot Assembly
	--
	-----------------------------------------------------------------------------
		
		-- Bone zAxis for the leg will be perpendicular to the bone plane
		local legAxis = normalize ( cross (posHip-posKnee) (posAnkle-posKnee) )
		-- Bone zAxis for the foot will be along the ground
		local zAxis = normalize (cross (posToe-posHeel) [0,0,10])
		
		local boneFoot 		= SSR_rig.makeBone (prefix + "IKbone_Foot") posHeel posToe legAxis boneSize
		local boneBall 		= SSR_rig.makeBone (prefix + "IKbone_Ball") posToe posBall zAxis boneSize
		local boneAnkle 	= SSR_rig.makeBone (prefix + "skinBone_Foot") posBall posAnkle zAxis boneSize
		local boneAnkleEnd	= SSR_rig.makeBone (prefix + "IKbone_AnkleEnd") posAnkle (posAnkle + [0,0,-boneSize]) zAxis boneSize
		local boneToe 		= SSR_rig.makeBone (prefix + "skinBone_Toe") posBall posToe zAxis boneSize
		local boneToeEnd 	= SSR_rig.makeBone (prefix + "IKbone_ToeEnd") posToe (posToe + [0,0,boneSize]) zAxis boneSize
		
		-- Alignments and parenting
		SSR_rig.objAlign boneAnkleEnd boneAnkle
		SSR_rig.objAlign boneToeEnd boneToe
		
		boneBall.parent = boneFoot
		boneAnkle.parent = boneBall
		boneAnkleEnd.parent = boneAnkle
		boneToe.parent = boneBall
		boneToeEnd.parent = boneToe
		
		-- Parent the IK system onto the foot
		IKend.parent = boneAnkleEnd
		
		--	Helpers we will need to twist and roll the foot
		local heelTwister = SSR_rig.nullTransform boneFoot
		heelTwister.name = prefix + "heelTwister"
		local toeTwister = SSR_rig.nullTransform boneBall
		toeTwister.name = prefix + "toeTwister"
		
		-- Leg root constraint, the entire leg will be parented to this
		legConTop = SSR_rig.makeHelper [0, posHip.y, posHip.z] boneSize (prefix + "RootCon") cross:true
		IKroot.parent = legConTop
		
		-- Leg root, this node will be parented to the main rig
		legTop = SSR_rig.makeHelper [0, posHip.y, posHip.z] (boneSize*2) (prefix + "Root") box:true
		SSR_rig.parentCon legConTop legTop
		
		
	-----------------------------------------------------------------------------
	--
	--		Create leg controller, aligned to foot bones
	--
	-----------------------------------------------------------------------------
		
		-- Calculate leg controller alignment
		local sideMult = undefined
		if side == "R" then sideMult = 1 else sideMult = -1

		local v1 = [posToe.x, posToe.y, 0] - [posHeel.x, posHeel.y, 0]  
		local theAngle = sideMult * acos (dot (normalize v1) [0,-1,0])
		--SSR_gen.verbose ("Angle between world and leg is: " + theAngle as string)
		
		local controlLength = (distance posHeel posToe) * 1.3 
		local controlWidth = (distance posSideOut posSideIn) * 1.3

		local footIKControl = Rectangle length:controlLength width:controlWidth 
		footIKControl.rotation = (eulerAngles 0 0 theAngle)
		footIKControl.position = (posHeel + posToe) * 0.5
		
		footIKControl.pivot = posAnkle
		footIKControl.name = charName + "_" + side + "_foot_IK_CTL"
		
		setTransformLockFlags footIKControl #{7..9}
		
		-- Parent up the heel null object to the controller
		heelTwister.parent = footIKControl

		
	-----------------------------------------------------------------------------
	--
	--		Additional helpers that will twist and bank the leg
	--
	-----------------------------------------------------------------------------

		local legBankOut = SSR_rig.makeHelper posSideOut boneSize (prefix + "bankOut") box:true
		local legBankIn = SSR_rig.makeHelper posSideIn boneSize (prefix + "bankIn") box:true
		
		SSR_rig.objAlign legBankOut footIKControl
		SSR_rig.objAlign legBankIn footIKControl
		--	We add a null transform to avoid the wired bank helpers freaking out
		local bankNull = SSR_rig.nullTransform legBankOut
		bankNull.parent = toeTwister
		legBankIn.parent = legBankOut
		
		-- Next, we add a null transform to the ball joint, and parent them to the banks
		local ballNull = SSR_rig.nullTransform boneBall
		ballNull.parent = legBankIn
		
	-----------------------------------------------------------------------------
	--
	--		IK foot following torso/world constraint setup
	--
	-----------------------------------------------------------------------------
	
		-- Helpers that will be aligned with local or world
		local legWorld = SSR_rig.makeHelper posAnkle boneSize (prefix + "footConstraint_World") box:true
		local legLocal = SSR_rig.makeHelper posAnkle (boneSize*1.5) (prefix + "footConstraint_Local") box:true
		SSR_rig.objAlign legWorld footIKControl
		SSR_rig.objAlign legLocal footIKControl
		
		legLocal.parent = legConTop
		
		-- Null that the controller will be parented to, and that will be constrained
		local legNull = SSR_rig.makeHelper posAnkle boneSize (prefix + "footConstraint") cross:true
		SSR_rig.objAlign legNull footIKControl
		footIKControl.parent = legNull
		legNull.parent = legConTop
		
		-- The constraint controllers
		local legCon = SSR_rig.posConTwo legNull legWorld legLocal
		local legRotCon = SSR_rig.rotConTwo legNull legWorld legLocal
		
	-----------------------------------------------------------------------------
	--
	--		Knee IK controller objects
	--
	-----------------------------------------------------------------------------	
		
		-- The spline knee control object
		local kneeLookatControl = SSR_rig.crossCTL kneeLookat.pos (boneSize*4) alignAxis:"X"
		kneeLookatControl.name = charName + "_" + side + "_knee_IK_CTL"
		SSR_rig.objAlign kneeLookatControl footIKControl
		kneeLookatControl.parent = kneeWorldNull
		kneeLookat.parent = kneeLookatControl
		setTransformLockFlags kneeLookatControl #{4..9}
		
		-- The knee pin control object
		local kneePinControl = SSR_rig.crossCTL posKnee (boneSize*4) alignAxis:"X"
		kneePinControl.name = charName + "_" + side + "_kneePin_IK_CTL"
		SSR_rig.objAlign kneePinControl footIKControl
		kneePinControl.parent = kneePin.parent
		kneePin.parent = kneePinControl
		setTransformLockFlags kneePinControl #{4..9}	


	-----------------------------------------------------------------------------
	--
	--		Controller custom attributes
	--
	-----------------------------------------------------------------------------
		
		-- Knee Pin Controller
		local ca = attributes transforms
		(
			parameters params rollout:extraTrans
			(
				X_Pos		type:#float ui:PXSp
				Y_Pos		type:#float ui:PYSp
				Z_Pos		type:#float ui:PZSp
			)
			rollout extraTrans "Transforms"
			(
				spinner PXSp 	"Translate X " 	range:[-1000000, 1000000, 0] align:#right
				spinner PYSp 	"Translate Y " 	range:[-1000000, 1000000, 0] align:#right
				spinner PZSp 	"Translate Z " 	range:[-1000000, 1000000, 0] align:#right
			)
		)
		
		local penMod = 'PEN_Attribute_Holder 2'()
		addModifier kneePinControl penMod
		custAttributes.add penMod ca
		
		
		-- Knee and Foot Controller
		penMod.transforms.X_Pos.controller = \
			kneePinControl.position.controller.X_position.controller
		penMod.transforms.Y_Pos.controller = \
			kneePinControl.position.controller.Y_position.controller
		penMod.transforms.Z_Pos.controller = \
			kneePinControl.position.controller.Z_position.controller
		
		local ca = attributes footControls
		(
			parameters params rollout:footR
			(
				PosX	type:#float ui:PXSp
				PosY	type:#float ui:PYSp
				PosZ	type:#float ui:PZSp
				RotX	type:#angle ui:RXSp
				RotY	type:#angle ui:RYSp
				RotZ	type:#angle ui:RZSp
				legPos_WL	type:#float ui:(legSp, legSl)
			)
			parameters params2 rollout:attR
			(
				footRoll 	type:#angle ui:footRollSp
				ballRoll	type:#angle ui:ballRollSp
				sideRoll	type:#angle ui:sideRollSp
				toeWiggle	type:#angle ui:toeWiggleSp
				heelTwist	type:#angle ui:heelTwistSp
				toeTwist 	type:#angle ui:toeTwistSp
			)
			rollout footR "Leg Transforms"
			(
				spinner PXSp 	"Translate X " 	range:[-1000000, 1000000, 0]
				spinner PYSp 	"Translate Y " 	range:[-1000000, 1000000, 0]
				spinner PZSp 	"Translate Z " 	range:[-1000000, 1000000, 0]
				spinner RXSp 	"Rotation X " 	range:[-1000000, 1000000, 0]
				spinner RYSp 	"Rotation Y " 	range:[-1000000, 1000000, 0]
				spinner RZSp 	"Rotation Z " 	range:[-1000000, 1000000, 0]
				label lbl02 	"Leg follows World <> Hip" align:#centre offset:[0,10]
				spinner legSp 	"" range:[0, 100, 0] fieldwidth:40 align:#left offset:[10, 10] across:2
				slider legSl 	"" range:[0, 100, 0]
			)
			rollout attR "Foot Attributes"
			(
				spinner footRollSp 		"Foot Roll " 		range:[-90, 90, 0]
				spinner ballRollSp		"Ball Roll "		range:[-90, 90, 0]
				spinner sideRollSp		"Foot Bank "	range:[-90, 90, 0]
				spinner toeWiggleSp 	"Toes Wiggle " 	range:[-90, 90, 0]
				spinner heelTwistSp		"Heel Twist "	range:[-90, 90, 0]
				spinner toeTwistSp		"Toe Twist "	range:[-90, 90, 0]
			)
		)
		
		local ca_K = attributes kneeLookatControls
		(
			parameters params3 rollout:kneeR
			(
				PosX 		type:#float ui:kPXSp
				PosY		type:#float ui:kPYSp
				PosZ		type:#float ui:kPZSp
				soft_IK		type:#float ui:softSp
			)
			parameters params4 rollout:conR
			(
				knee_Hip_Foot 	type:#float ui:(legHipSp, legHipSl)
				knee_World		type:#float ui:(worldSp, worldSl)
			)
			rollout kneeR "Knee Transforms"
			(
				spinner kPXSp 	"Translate X " 	range:[-1000000, 1000000, 0]
				spinner kPYSp 	"Translate Y " 	range:[-1000000, 1000000, 0]
				spinner kPZSp 	"Translate Z " 	range:[-1000000, 1000000, 0]
				spinner softSp	"Soft IK "		range:[0.0, 1000.0, 0]
			)
			rollout conR "Knee Constraints"
			(
				label lbl01 		"Knee follows Hip <> Foot" align:#centre offset:[0,5]
				spinner legHipSp	"" range:[-100, 100, 0] fieldwidth:40 align:#left offset:[10, 10] across:2
				slider legHipSl 	"" range:[-100, 100, 0]
				
				label lbl02 		"Force knee to world" align:#centre offset:[0,5]
				spinner worldSp		"" range:[0, 100, 0] fieldwidth:40 align:#left offset:[10, 10] across:2
				slider worldSl 		"" range:[0, 100, 0]
			)
		)		
		
	-----------------------------------------------------------------------------
	--
	--		Foot parameter connections
	--
	-----------------------------------------------------------------------------
		-- Add custom attributes for the controller
		local penMod = 'PEN_Attribute_Holder 2'()
		addModifier footIKControl penMod
		custAttributes.add penMod ca
		custAttributes.add penMod ca_K
		
		-- Transform connections
		penMod.footControls.PosX.controller = \
		footIKControl.pos.controller.X_Position.controller
		penMod.footControls.PosY.controller = \
		footIKControl.pos.controller.Y_Position.controller
		penMod.footControls.PosZ.controller = \
		footIKControl.pos.controller.Z_Position.controller
		
		penMod.footControls.RotX.controller = \
		footIKControl.rotation.controller.X_Rotation.controller
		penMod.footControls.RotY.controller = \
		footIKControl.rotation.controller.Y_Rotation.controller
		penMod.footControls.RotZ.controller = \
		footIKControl.rotation.controller.Z_Rotation.controller
		
		-- Leg follow hip or world connections
		paramWire.connect penMod.footControls[#legPos_WL] \
			legNull.position.controller[#Position_Weight_0] \
			"100-legPos_WL"
		paramWire.connect penMod.footControls[#legPos_WL] \
			legNull.position.controller[#Position_Weight_1] \
			"legPos_WL"
		paramWire.connect penMod.footControls[#legPos_WL] \
			legNull.rotation.controller[#Orientation_Weight_0] \
			"100-legPos_WL"
		paramWire.connect penMod.footControls[#legPos_WL] \
			legNull.rotation.controller[#Orientation_Weight_1] \
			"legPos_WL"
		
		-- Freeze transform on foot, ball and ankle
		SSR_rig.freezeTransform heelTwister
		SSR_rig.freezeTransform toeTwister
		SSR_rig.freezeTransform boneFoot
		SSR_rig.freezeTransform boneBall
		SSR_rig.freezeTransform boneAnkle
		SSR_rig.freezeTransform boneToe

		-- Set up the necessary wiring for the foot
		-- Foot roll (first heel, then toe)
		paramWire.connect penMod.footControls[#footRoll] \
			heelTwister.rotation.controller.Zero_Euler_XYZ.controller[#Z_Rotation] \
			"if footRoll < 0 then -footRoll else 0"
		
		paramWire.connect penMod.footControls[#footRoll] \
			toeTwister.rotation.controller.Zero_Euler_XYZ.controller[#Z_Rotation] \
			"if footRoll > 0 then -footRoll else 0"
		
		-- Ball roll
		paramWire.connect penMod.footControls[#ballRoll] \
			boneAnkle.rotation.controller.Zero_Euler_XYZ.controller[#Z_Rotation] \
			"-ballRoll"
		
		-- Foot side roll (first outside, then inside)
		if side == "R" then
		(
			paramWire.connect penMod.footControls[#sideRoll] \
				legBankOut.rotation.controller[#Y_Rotation] \
				"if sideRoll < 0 then sideRoll else 0"
			
			paramWire.connect penMod.footControls[#sideRoll] \
				legBankIn.rotation.controller[#Y_Rotation] \
				"if sideRoll > 0 then sideRoll else 0"
		) else (
			paramWire.connect penMod.footControls[#sideRoll] \
				legBankOut.rotation.controller[#Y_Rotation] \
				"if sideRoll > 0 then sideRoll else 0"
			
			paramWire.connect penMod.footControls[#sideRoll] \
				legBankIn.rotation.controller[#Y_Rotation] \
				"if sideRoll < 0 then sideRoll else 0"	
		)
		
		-- Heel Twist
		paramWire.connect penMod.footControls[#heelTwist] \
			heelTwister.rotation.controller.Zero_Euler_XYZ.controller[#Y_Rotation] \
			"heelTwist"
		
		-- Toe Twist
		paramWire.connect penMod.footControls[#toeTwist] \
			toeTwister.rotation.controller.Zero_Euler_XYZ.controller[#Y_Rotation] \
			"toeTwist"

		-- Toe wiggle
		paramWire.connect penMod.footControls[#toeWiggle] \
			boneToe.rotation.controller.Zero_Euler_XYZ.controller[#Z_Rotation] \
			"toeWiggle"
			
		-- Knee Transformations	
		penMod.kneeLookatControls.PosX.controller = 	kneeLookatControl.pos.controller.X_Position.controller
		penMod.kneeLookatControls.PosY.controller = 	kneeLookatControl.pos.controller.Y_Position.controller
		penMod.kneeLookatControls.PosZ.controller = 	kneeLookatControl.pos.controller.Z_Position.controller
		
		-- Knee Attributes, follow hip or foot
		paramWire.connect penMod.kneeLookatControls[#knee_Hip_Foot] \
			kneeHipFoot.position.controller[#Position_Weight_0] \
			"(knee_Hip_Foot*0.5)+50"
		paramWire.connect penMod.kneeLookatControls[#knee_Hip_Foot] \
			kneeHipFoot.position.controller[#Position_Weight_1] \
			"50-(knee_Hip_Foot*0.5)"
		
		-- Follow world
		paramWire.connect penMod.kneeLookatControls[#knee_world] \
			kneeWorldNull.position.controller[#Position_Weight_0] \
			"100-knee_world"
		paramWire.connect penMod.kneeLookatControls[#knee_world] \
			kneeWorldNull.position.controller[#Position_Weight_1] \
			"knee_world"	

		-- And link the smooth control to our Leg Controller object
		penMod.kneeLookatControls[#soft_ik].controller = \
			limitCon[#Limits][#Upper_Smoothing].controller = \
			Bezier_float()
		
	-----------------------------------------------------------------------------
	--
	--		Knee parameter connections
	--
	-----------------------------------------------------------------------------

		-- Add our existing knee attributes to the knee controller
		penMod_K = 'PEN_Attribute_Holder 2'()
		addModifier kneeLookatControl penMod_K
		custAttributes.add penMod_K ca_K
		
		-- Connect up the sliders
		penMod_K.kneeLookatControls.PosX.controller = 			penMod.kneeLookatControls.PosX.controller
		penMod_K.kneeLookatControls.PosY.controller = 			penMod.kneeLookatControls.PosY.controller
		penMod_K.kneeLookatControls.PosZ.controller = 			penMod.kneeLookatControls.PosZ.controller
		penMod_K.kneeLookatControls.knee_Hip_Foot.controller = 	penMod.kneeLookatControls.knee_Hip_Foot.controller
		penMod_K.kneeLookatControls.knee_world.controller = 	penMod.kneeLookatControls.knee_world.controller

		penMod_K.kneeLookatControls[#soft_ik].controller = 	penMod.kneeLookatControls[#soft_ik].controller
	
	-----------------------------------------------------------------------------
	--
	--		Twist Bones
	--
	-----------------------------------------------------------------------------
		
		if doTwist == true then
		(
			twistReturn = SSR_rig.stretchTwistBones boneShin boneAnkle numTwistShin twistUp:false debug:debug
			for c in twistReturn[1] do controlLayer.addNode c
			for o in twistReturn[2] do skinLayer.addNode o
			
			twistReturn = SSR_rig.stretchTwistBones boneThigh (boneThigh.parent.parent) numTwistThigh twistUp:true debug:debug
			for c in twistReturn[1] do controlLayer.addNode c
			for o in twistReturn[2] do skinLayer.addNode o

			-- The previous skinBones are no longer skinning bones
			boneThigh.name = prefix + "bone_Thigh"
			boneShin.name = prefix + "bone_Shin"
			rigLayer.addNode boneThigh
			rigLayer.addNode boneShin
		)	
		
		return #(footIKControl, kneeLookatControl)
	)
)

SSR_part = SSR_part()