-----------------------------------------------------------------------------
--
--		FK spine setup
--
--		v1.0:	Standard setup
--		v1.1:	Chest controllers follow bones loosely through expressions
--		v1.2:	Slider to align chest to world/local
--		v1.3:	Named all objects and put them into layers
--		v1.4:	Breathe attributes
--		v1.5:	Spine controllers are now aligned with the bones
--		v1.6:	Controllers are rings instead of pins
--		v1.7:	Optional spine root aligned to world
--
-----------------------------------------------------------------------------
global SSR_part

struct SSR_part
(
	function doPart charName extParams:#(true) boneSize:1 =
	(
		-- External parameters
		local rootAlign = extParams[1] 
		
		local helperList = SSR_loc.collectLocators charName ("*Locator_Spine_*")
		local posHip, posLBack, posMBack, posHBack, posChest, posRoot, posChestR, posChestL, posBelly
		local prefix = (charName + "_spine_")
		local doBreathe = false

		-- Look for helper objects that give locations to spine bone positions
		if helperList != false then for h in helperList do
		(
			if matchPattern h.name pattern:("*_Spine_Hip") == true then 	posHip = h.pos
			if matchPattern h.name pattern:("*_Spine_LowBack") == true then posLBack = h.pos
			if matchPattern h.name pattern:("*_Spine_MidBack") == true then posMBack = h.pos
			if matchPattern h.name pattern:("*_Spine_UpperBack") == true then posHBack = h.pos
			if matchPattern h.name pattern:("*_Spine_Chest") == true then posChest = h.pos
			if matchPattern h.name pattern:("*_Leg_R_Hip") == true then posRoot = [0, h.pos.y, h.pos.z]
			if matchPattern h.name pattern:("*_Chest_R_Breathe") == true then posChestR = h.pos
			if matchPattern h.name pattern:("*_Chest_L_Breathe") == true then posChestL = h.pos
			if matchPattern h.name pattern:("*_Belly_Breathe") == true then posBelly = h.pos
		)
		
		-- If one of these helpers are not found, we will exit
		if posHip == undefined or posLBack == undefined \
			or posMBack == undefined or posHBack == undefined or posChest == undefined \
		then (
			SSR_gen.verbose ("Missing locators for spine, operation aborted.")
			return false
		) else (
			SSR_gen.verbose ("All locators are found, now creating spine.")
		)
		
		-- If all chest breathe attributes are found, we will add breathe attributes
		if posChestR != undefined and posChestL != undefined and posBelly != undefined then 
			doBreathe = true
		
	-----------------------------------------------------------------------------
	--
	--		Bones
	--
	-----------------------------------------------------------------------------
		
		-- Create our main bones
		--local boneSize = (distance posHip poschest) / 20
		local zAxis = [ -1, 0, 0]
		local boneHip		= SSR_rig.makeBone (prefix + "skinBone_Hip") posLBack posHip zAxis boneSize
		local boneHipEnd	= SSR_rig.makeBone (prefix + "bone_HipEnd") posHip (posHip + [0, 0, boneSize]) zAxis boneSize
		local boneLBack		= SSR_rig.makeBone (prefix + "skinBone_LowerBack") posLBack posMBack zAxis boneSize
		local boneUBack		= SSR_rig.makeBone (prefix + "skinBone_UpperBack") posMBack posHBack zAxis boneSize
		local boneChest		= SSR_rig.makeBone (prefix + "skinBone_Chest") posHBack poschest zAxis boneSize
		local boneChestEnd	= SSR_rig.makeBone (prefix + "bone_ChestEnd") poschest (poschest + [0, 0, boneSize]) zAxis boneSize
		
		boneHip.boneScaleType = boneLBack.boneScaleType = boneUBack.boneScaleType = boneChest.boneScaleType = #scale
		
		boneHip.frontfin 	= true
		boneLBack.frontfin 	= true
		boneUBack.frontfin 	= true
		boneChest.frontfin 	= true
		
		-- Alignments and parenting
		SSR_rig.objAlign boneChestEnd boneChest
		SSR_rig.objAlign boneHipEnd boneHip
		boneChestEnd.parent = boneChest
		boneHipEnd.parent = boneHip
		boneChest.parent = boneUBack
		boneUBack.parent = boneLBack

		-- Create stretchy bone chains
		local spineStretch = SSR_rig.stretchyChain #(boneLBack, boneUBack, boneChest, boneChestEnd) helperSize:(boneSize*2) doParenting:true
		local boneLBackStretch = spineStretch[1]
		local boneMBackStretch = spineStretch[2]
		local boneUBackStretch = spineStretch[3]
		local boneChestStretch = spineStretch[4]
		
		local hipStretch = SSR_rig.stretchyChain #(boneHip, boneHipEnd) helperSize:(boneSize*2) doParenting:true
		local boneHipStretch = hipStretch[1]
		local boneHipEndStretch = hipStretch[2]

		
	-----------------------------------------------------------------------------
	--
	--		Breathe Bones and helpers
	--
	-----------------------------------------------------------------------------	
		if doBreathe == true then
		(
			-- Chest breathers
			local breatheRRoot = posChestR - (normalize (posChestR - posHBack) * boneSize * 2)
			local boneBreatheR		= SSR_rig.makeBone (prefix + "skinBone_BreatheChestR") breatheRRoot posChestR [0,0,1] boneSize
			local boneBreatheREnd	= SSR_rig.endBone boneBreatheR
			boneBreatheREnd.name = prefix + "bone_BreatheChestREnd"
			
			local breatheLRoot = posChestL - (normalize (posChestL - posHBack) * boneSize * 2)
			local boneBreatheL		= SSR_rig.makeBone (prefix + "skinBone_BreatheChestL") breatheLRoot posChestL [0,0,1] boneSize
			local boneBreatheLEnd	= SSR_rig.endBone boneBreatheL
			boneBreatheLEnd.name = prefix + "bone_BreatheChestLEnd"
			
			local breatheRNull = SSR_rig.nullTransform boneBreatheR
			breatheRNull.name = substituteString (breatheRNull.name) "_skinBone_" "_"
			local breatheLNull = SSR_rig.nullTransform boneBreatheL
			breatheLNull.name = substituteString (breatheLNull.name) "_skinBone_" "_"
			
			-- Chest breathers are constrained between chest bone and upper back bone
			local chestBreatheConUp = SSR_rig.makeHelper [0, posChestL.y, posChestL.z] (boneSize*1.2) (prefix + "breatheCon_Up") box:true
			local chestBreatheConDown = SSR_rig.makeHelper [0, posChestL.y, posChestL.z] boneSize (prefix + "breatheCon_Down") box:true
			local chestBreatheCon = SSR_rig.makeHelper [0, posChestL.y, posChestL.z] boneSize (prefix + "breatheCon") cross:true
			
			chestBreatheConUp.parent = boneChest
			chestBreatheConDown.parent = boneUBack
			
			local chestPosCon = SSR_rig.posConTwo chestBreatheCon chestBreatheConUp chestBreatheConDown
			local chestRotCon = SSR_rig.rotConTwo chestBreatheCon chestBreatheConUp chestBreatheConDown
			
			breatheRNull.parent = breatheLNull.parent = chestBreatheCon
			
			-- Belly breather
			local breatheBellyRoot = posBelly - (normalize (posBelly - posLBack) * boneSize * 2)
			local boneBreatheBelly		= SSR_rig.makeBone (prefix + "skinBone_BreatheBelly") breatheBellyRoot posBelly [0,0,1] boneSize
			local boneBreatheBellyEnd	= SSR_rig.endBone boneBreatheBelly
			boneBreatheBellyEnd.name = prefix + "bone_BreatheBellyEnd"
			
			local breatheBellyNull = SSR_rig.nullTransform boneBreatheBelly
			breatheBellyNull.name = substituteString (breatheBellyNull.name) "_skinBone_" "_"
			
			-- Belly breather is constrained between chest bone and upper back bone
			local bellyBreatheConUp = SSR_rig.makeHelper posBelly (boneSize*1.2) (prefix + "bellyBreatheCon_Up") box:true
			local bellyBreatheConDown = SSR_rig.makeHelper posBelly boneSize (prefix + "bellyBreatheCon_Down") box:true
			local bellyBreatheCon = SSR_rig.makeHelper posBelly boneSize (prefix + "bellyBreatheCon") cross:true
			
			bellyBreatheConUp.parent = boneLBack
			bellyBreatheConDown.parent = boneHip
			
			local bellyPosCon = SSR_rig.posConTwo bellyBreatheCon bellyBreatheConUp bellyBreatheConDown
			local bellyRotCon = SSR_rig.rotConTwo bellyBreatheCon bellyBreatheConUp bellyBreatheConDown
			
			breatheBellyNull.parent = bellyBreatheCon	
			
			skinLayer.addNode boneBreatheR
			skinLayer.addNode boneBreatheL
			skinLayer.addNode boneBreatheBelly
		) -- if doBreathe == true then
		
		-- Clavicle breather
		-- This will be created anyway, as this is the default connection point for the arms to the chest
		local clavicleBreathe = SSR_rig.makeHelper (posChest + [0,-boneSize,0]) boneSize (prefix + "clavicleParent") cross:true
		SSR_rig.objAlign clavicleBreathe boneChest
		local clavicleBreatheNull = SSR_rig.nullTransform clavicleBreathe
		clavicleBreatheNull.parent = boneChestStretch

	-----------------------------------------------------------------------------
	--
	--		Control objects
	--
	-----------------------------------------------------------------------------

		-- Control objects
		local hipControl 	= SSR_rig.ringCTL posLBack (boneSize*20) alignAxis:"X"
		hipControl.name		= charName + "_hip_CTL" 
		SSR_rig.objAlign hipControl boneLBack
		setTransformLockFlags hipControl #{1..3,7..9}
		
		local backControl 	= SSR_rig.ringCTL posLBack (boneSize*30) alignAxis:"X"
		backControl.name 	= charName + "_back_CTL"
		SSR_rig.objAlign backControl boneLBack
		setTransformLockFlags backControl #{1..3,7..9}
		
		local chestControl	= SSR_rig.ringCTL posMBack (boneSize*30) alignAxis:"X"
		chestControl.name 	= charName + "_chest_CTL"
		SSR_rig.objAlign chestControl boneUBack
		setTransformLockFlags chestControl #{1..3,7..9}
		
		/*local chestStretchControl 	= SSR_rig.pinCTL posChest (boneSize*30) (boneSize*10) alignAxis:"Y"
		chestStretchControl.name 	= charName + "_chestStretch_CTL"
		--SSR_rig.objAlign chestStretchControl boneChest
		setTransformLockFlags chestStretchControl #{4..9}*/
		
		local torsoControl 	= SSR_rig.ringCTL posHip (boneSize*30) alignAxis:"X"
		
		-- Move torso controller to center of legs, if they exist
		if posRoot != undefined then torsoControl.pos = posRoot
		torsoControl.name = charName + "_torso_CTL"
		SSR_rig.objAlign torsoControl boneLBack closestAxis:rootAlign -- Optional align to world axis
		setTransformLockFlags torsoControl #{7..9}
		
		--chestTopControl.parent = chestControl
		hipControl.parent = backControl.parent = torsoControl
		chestControl.parent = backControl
		
		boneHipStretch.parent = boneLBackStretch.parent = torsoControl
		
		-- A null for later constraint setup
		chestNull = SSR_rig.nullTransform chestControl
		
	-----------------------------------------------------------------------------
	--
	--		Helpers and constraints
	--
	-----------------------------------------------------------------------------	
		-- The helpers our bones will be aligned to
		local alignHip = SSR_rig.makeHelper posLBack boneSize (prefix + "alignHip") box:true
		SSR_rig.objAlign alignHip boneLBackStretch
		alignHip.parent = torsoControl
		
		local alignLow = SSR_rig.makeHelper posLBack boneSize (prefix + "alignLow") box:true
		SSR_rig.objAlign alignLow boneHipStretch
		alignLow.parent = hipControl
		
		local alignBack = SSR_rig.makeHelper posMBack boneSize (prefix + "alignBack") box:true
		SSR_rig.objAlign alignBack boneLBackStretch
		alignBack.parent = backControl
		
		local alignChest = SSR_rig.makeHelper posChest boneSize (prefix + "alignChest") box:true
		SSR_rig.objAlign alignChest boneLBackStretch
		alignChest.parent = chestControl
		--chestStretchControl.parent = alignChest
		
		-- Constraints to slide the chest between world and local orientation
		local chestRoot = SSR_rig.makeHelper posChest boneSize (prefix + "chestRoot") cross:true
		SSR_rig.objAlign chestRoot backControl
		chestAlignCon = SSR_rig.rotConTwo chestNull backControl chestRoot
		
		-- Rotation constraints
		SSR_rig.rotCon boneHipStretch alignLow
		
		local lBackCon = SSR_rig.rotConTwo boneLBackStretch alignHip alignBack
		lBackCon.weight[1] = 35
		lBackCon.weight[2] = 65
		local mBackCon = SSR_rig.rotConTwo boneMBackStretch alignBack alignChest
		mBackCon.relative = on
		local chestCon = SSR_rig.rotCon boneUBackStretch alignChest
		chestCon.relative = on
		
		-- A wire controller to ease off the back twist
		SSR_rig.nullTransform hipControl
		
		local rotList = rotation_list()
		boneLBack.rotation.controller = rotList
		rotList.available.controller = Euler_XYZ()
		paramWire.connect \
			hipControl.rotation.controller[#X_Rotation] \
			rotList.Euler_XYZ.controller[#X_Rotation] "X_Rotation*0.5"
		
		-- Stretchable list controllers for the bone stretchers
		local hipListCon = boneHipEndStretch.position.controller.X_position.controller = Float_list() 			
		local hipStretcher = hipListCon.available.controller = Bezier_float() 	
		hipListCon.setname 1 "FrozenXPos" 	
		hipListCon.setname 2 "StretchyXPos" 			
		hipListCon.SetActive 2 	
		
		local lowBackListCon = boneMBackStretch.position.controller.X_position.controller = Float_list() 			
		local lowBackStretcher = lowBackListCon.available.controller = Bezier_float() 	
		lowBackListCon.setname 1 "FrozenXPos" 	
		lowBackListCon.setname 2 "StretchyXPos" 			
		lowBackListCon.SetActive 2 	
		
		local backListCon = boneUBackStretch.position.controller.X_position.controller = Float_list() 			
		local instancedStretcher = backListCon.available.controller = Bezier_float() 	
		backListCon.setname 1 "FrozenXPos" 	
		backListCon.setname 2 "StretchyXPos" 			
		backListCon.SetActive 2 	
		
		local chestListCon = boneChestStretch.position.controller.X_position.controller = Float_list() 			
		chestListCon.available.controller = instancedStretcher 
		chestListCon.setname 1 "FrozenXPos" 	
		chestListCon.setname 2 "StretchyXPos" 			
		chestListCon.SetActive 2 
		
	
	
	-----------------------------------------------------------------------------
	--
	--		Controller custom attributes
	--
	-----------------------------------------------------------------------------
		
		-- Add custom attributes for the hip controller
		local lowMod = 'PEN_Attribute_Holder 2'()
		addModifier hipControl lowMod
		
		local stretchAtt = attributes boneStretch
		(
			parameters params rollout:handR
			(
				X_Pos		type:#float ui:X_PosSp
			)
			rollout handR "Stretch"
			(
				spinner X_PosSp 	"Stretch " 	range:[-1000000, 1000000, 0] align:#right
			)
		)		
		
		local lowAtt = attributes transforms
		(
			parameters params rollout:handR
			(
				X_Rot		type:#angle ui:X_RotSp
				Y_Rot		type:#angle ui:Y_RotSp
				Z_Rot		type:#angle ui:Z_RotSp
			)
			rollout handR "Transforms"
			(
				spinner X_RotSp 	"Rotate X " 	range:[-1000000, 1000000, 0] align:#right
				spinner Y_RotSp 	"Rotate Y " 	range:[-1000000, 1000000, 0] align:#right
				spinner Z_RotSp 	"Rotate Z " 	range:[-1000000, 1000000, 0] align:#right
			)
		)		
		custAttributes.add lowMod lowAtt
		custAttributes.add lowMod stretchAtt

		-- Connect up custom attributes		
		lowMod.transforms.X_Rot.controller = \
			hipControl.rotation.controller.X_Rotation.controller
		lowMod.transforms.Y_Rot.controller = \
			hipControl.rotation.controller.Y_Rotation.controller
		lowMod.transforms.Z_Rot.controller = \
			hipControl.rotation.controller.Z_Rotation.controller
			
		lowMod.boneStretch.X_Pos.controller = \
			hipListCon[2].controller
		
		-- Custom attributes for the back controller
		local backMod = 'PEN_Attribute_Holder 2'()
		addModifier backControl backMod
		custAttributes.add backMod lowAtt
		custAttributes.add backMod stretchAtt
		
		-- Connect up custom attributes
		SSR_rig.nullTransform backControl

		backMod.transforms.X_Rot.controller = \
			backControl.rotation.controller.X_Rotation.controller
		backMod.transforms.Y_Rot.controller = \
			backControl.rotation.controller.Y_Rotation.controller
		backMod.transforms.Z_Rot.controller = \
			backControl.rotation.controller.Z_Rotation.controller
			
		backMod.boneStretch.X_Pos.controller = \
			lowBackListCon[2].controller

		-- Custom attributes for the chest controller
		local chestMod = 'PEN_Attribute_Holder 2'()
		addModifier chestControl chestMod
			
		local chestAtt = attributes constraints
		(
			parameters params rollout:chestRoll
			(
				Chest_Follow	type:#float ui:(chestSp, chestSl)
			)
			rollout chestRoll "Constraints"
			(
				label lbl01 	"Chest follows Spine <> World" align:#centre
				spinner chestSp "" range:[0, 100, 0] fieldwidth:40 align:#left offset:[10, 10] across:2
				slider chestSl 	"" range:[0, 100, 0]
			)
		)
		
		custAttributes.add chestMod lowAtt
		custAttributes.add chestMod stretchAtt
		custAttributes.add chestMod chestAtt
		
		-- Connect up custom attributes
		SSR_rig.nullTransform chestControl

		chestMod.transforms.X_Rot.controller = \
			chestControl.rotation.controller.X_Rotation.controller
		chestMod.transforms.Y_Rot.controller = \
			chestControl.rotation.controller.Y_Rotation.controller
		chestMod.transforms.Z_Rot.controller = \
			chestControl.rotation.controller.Z_Rotation.controller
		
		chestMod.boneStretch.X_Pos.controller = \
			chestListCon[2].controller
		
		paramWire.connect chestMod.constraints[#Chest_Follow] \
			chestNull.rotation.controller[#Orientation_Weight_0] \
			"100-Chest_Follow"
		paramWire.connect chestMod.constraints[#Chest_Follow] \
			chestNull.rotation.controller[#Orientation_Weight_1] \
			"Chest_Follow"	
			
		-- Custom attributes for the torso controller
		local hipMod = 'PEN_Attribute_Holder 2'()
		addModifier torsoControl hipMod
		
		local hipAtt = attributes transforms
		(
			parameters params rollout:handR
			(
				X_Pos		type:#float ui:X_PosSp
				Y_Pos		type:#float ui:Y_PosSp
				Z_Pos		type:#float ui:Z_PosSp
				X_Rot		type:#angle ui:X_RotSp
				Y_Rot		type:#angle ui:Y_RotSp
				Z_Rot		type:#angle ui:Z_RotSp
			)
			rollout handR "Transforms"
			(
				spinner X_PosSp 	"Translate X " 	range:[-1000000, 1000000, 0] align:#right
				spinner Y_PosSp 	"Translate Y " 	range:[-1000000, 1000000, 0] align:#right
				spinner Z_PosSp 	"Translate Z " 	range:[-1000000, 1000000, 0] align:#right
				spinner X_RotSp 	"Rotate X " 	range:[-1000000, 1000000, 0] align:#right
				spinner Y_RotSp 	"Rotate Y " 	range:[-1000000, 1000000, 0] align:#right
				spinner Z_RotSp 	"Rotate Z " 	range:[-1000000, 1000000, 0] align:#right
			)
		)		
		
		custAttributes.add hipMod hipAtt
		
		-- Connect up custom attributes
		SSR_rig.nullTransform torsoControl -- Results in 360 rotations?
		
		hipMod.transforms.X_Pos.controller = \
			torsoControl.pos.controller.X_Position.controller
		hipMod.transforms.Y_Pos.controller = \
			torsoControl.pos.controller.Y_Position.controller
		hipMod.transforms.Z_Pos.controller = \
			torsoControl.pos.controller.Z_Position.controller

		hipMod.transforms.X_Rot.controller = \
			torsoControl.rotation.controller.X_Rotation.controller
		hipMod.transforms.Y_Rot.controller = \
			torsoControl.rotation.controller.Y_Rotation.controller
		hipMod.transforms.Z_Rot.controller = \
			torsoControl.rotation.controller.Z_Rotation.controller
		
		if doBreathe == true then
		(
			-- Custom attributes for the breathe bones
			local breatheAtt = attributes breathe_controls
			(
				parameters params rollout:handR
				(
					clavBreathe		type:#float ui:clavSp
					chestBreathe	type:#float ui:chestSp
					bellyBreathe	type:#float ui:bellySp
				)
				rollout handR "Breathe Controls"
				(
					spinner clavSp 		"Clavicle Breathe " range:[-1000000, 1000000, 0] align:#right fieldWidth:50
					spinner chestSp 	"Chest Breathe " 	range:[-1000000, 1000000, 0] align:#right fieldWidth:50
					spinner bellySp 	"Belly Breathe " 	range:[-1000000, 1000000, 0] align:#right fieldWidth:50
					label lbl01			"Hold ALT for fine control" offset:[0,10]
				)
			)	

			custAttributes.add chestMod breatheAtt
			chestMod.breathe_controls.clavBreathe.controller = \
				clavicleBreathe.pos.controller.X_Position.controller
			chestMod.breathe_controls.chestBreathe.controller = \
				boneBreatheR.pos.controller.X_Position.controller = \
				boneBreatheL.pos.controller.X_Position.controller
			chestMod.breathe_controls.bellyBreathe.controller = \
				boneBreatheBelly.pos.controller.X_Position.controller	
		) -- if doBreathe == true then

		
		return #(hipControl, backControl, chestControl, chestTopControl, torsoControl)
	)
)

SSR_part = SSR_part()