-----------------------------------------------------------------------------
--
--		Stretchy IK spine setup
--
--		v1.0: Standard setup
--		v1.1: Added root bone at the bottom that controls entire spine
--		v1.2:	Hip bone is stabilized a bit
--				Changed rotation wiring to controller instancing
--		v1.3:	Spine end End constrained to chest controller
--				so now the FK and IK spine will have the arms connected to the End 
--		v1.4 	Named all objects and put them into layers
--		v1.5	Optional spine root aligned to world
--
-----------------------------------------------------------------------------
global SSR_part

struct SSR_part
(
	function doPart charName extParams:#(true) boneSize:1 =
	(
		-- External parameters
		local rootAlign = extParams[1] 
		
		local helperList = SSR_loc.collectLocators charName ("*Locator_Spine_*")
		local posHip, posLBack, posMBack, posHBack, posChest, posRoot
		local prefix = (charName + "_spine_")
		
		-- Look for helper objects that give locations to spine bone positions
		if helperList != false then for h in helperList do
		(
			if matchPattern h.name pattern:("*_Spine_Hip") == true then 	posHip = h.pos
			if matchPattern h.name pattern:("*_Spine_LowBack") == true then posLBack = h.pos
			if matchPattern h.name pattern:("*_Spine_MidBack") == true then posMBack = h.pos
			if matchPattern h.name pattern:("*_Spine_UpperBack") == true then posHBack = h.pos
			if matchPattern h.name pattern:("*_Spine_Chest") == true then posChest = h.pos
			if matchPattern h.name pattern:("*_Leg_R_Hip") == true then posRoot = [0, h.pos.y, h.pos.z]
		)
		
		-- If one of these helpers are not found, we will exit
		if posHip == undefined or posLBack == undefined \
			or posMBack == undefined or posHBack == undefined or posChest == undefined \
		then (
			SSR_gen.verbose ("Missing locators for spine, operation aborted.")
			return false
		) else (
			SSR_gen.verbose ("All locators are found, now creating spine.")
		)
		
	-----------------------------------------------------------------------------
	--
	--		Spline IK and Bones
	--
	-----------------------------------------------------------------------------
		
		-- Spline with IK helpers
		local spineShape = SplineShape pos:[0, 0, 0]
		addNewSpline spineShape
		addKnot spineShape 1 #smooth #curve posHip
		addKnot spineShape 1 #smooth #curve posLBack
		addKnot spineShape 1 #smooth #curve posMBack
		addKnot spineShape 1 #smooth #curve posHBack
		addKnot spineShape 1 #smooth #curve poschest
		updateShape spineShape
		spineShape.name = prefix + "IKShape"
		
		local spineIK= Spline_IK_Control linkTypes:2 helper_size:(boneSize*10.0) box:true helper_axistripod:false helper_cross:false helper_centermarker:false
		addModifier spineShape spineIK
		for i = 1 to 5 do 
		(
			spineIK.createHelper i
			spineIK.helper_list[i].name = prefix + "IKHelper0" + i as string
		)
		
		-- Create our main bones
		--local boneSize = (distance posHip poschest) / 20
		local zAxis = [ -1, 0, 0]
		local boneHip			= SSR_rig.makeBone (prefix + "skinBone_Hip") posHip posLBack zAxis boneSize
		local boneLBack		= SSR_rig.makeBone (prefix + "skinBone_LowerBack") posLBack posMBack zAxis boneSize
		local boneMBack		= SSR_rig.makeBone (prefix + "skinBone_UpperBack") posMBack posHBack zAxis boneSize
		local boneHBack		= SSR_rig.makeBone (prefix + "skinBone_Chest") posHBack poschest zAxis boneSize
		local boneHBackEnd	= SSR_rig.makeBone (prefix + "bone_ChestEnd") poschest (poschest + [0, 0, boneSize]) zAxis boneSize
		
		boneHip.frontfin = true
		boneLBack.frontfin = true
		boneMBack.frontfin = true
		boneHBack.frontfin = true
		
		-- Alignments and parenting
		-- SSR_rig.objAlign target source
		SSR_rig.objAlign boneHBackEnd boneHBack
		boneHBackEnd.parent = boneHBack
		boneHBack.parent = boneMBack
		boneMBack.parent = boneLBack
		boneLBack.parent = boneHip

	-----------------------------------------------------------------------------
	--
	--		Control objects
	--
	-----------------------------------------------------------------------------

		-- Control objects
		local lowControl 	= SSR_rig.pinCTL posHip (boneSize*20) (bonesize*4)
		lowControl.name		= charName + "_hip_CTL" 
		
		local backControl 	= SSR_rig.pinCTL posMBack (boneSize*20) (bonesize*4)
		backControl.name 	= charName + "_back_CTL"
		
		local chestControl	= SSR_rig.pinCTL posChest (boneSize*20) (bonesize*4)
		chestControl.name 	= charName + "_chest_CTL"
		
		local hipControl 	= SSR_rig.ringCTL posHip (boneSize*30) alignAxis:"X"
		if posRoot != undefined then hipControl.pos = posRoot
		addModifier hipControl (xform())
		hipControl.xform.gizmo.rotation = eulerAngles 0 90 0
		convertTo hipControl SplineShape
		hipControl.name = charName + "_torso_CTL"
			
		SSR_rig.pivotAlign lowControl posHip
		SSR_rig.pivotAlign backControl posMBack
		SSR_rig.pivotAlign chestControl poschest
		
		-- Constrain the objects so that the rotation and not the position is inherited
		local backRot = SSR_rig.makeHelper lowControl.pos boneSize (prefix + "backRotate") cross:true
		local chestRot = SSR_rig.makeHelper backControl.pos boneSize (prefix + "chestRotate") cross:true
		
		SSR_rig.objAlign backRot lowControl
		SSR_rig.objAlign chestRot lowControl
		
		backControl.parent = backRot
		chestControl.parent = chestRot
		
		local hipParent = SSR_rig.makeHelper lowControl.pos boneSize (charName + "_torso_CTL_null") box:true
		
		hipControl.parent = hipParent
		lowControl.parent = hipControl
		backRot.parent = hipControl
		chestRot.parent = backRot
		
		SSR_rig.rotCon backRot hipControl
		SSR_rig.rotCon chestRot backControl


		-- Clavicle breather
		-- No breathe on the Spline IK rig, but this will be created anyway, as this is the default connection point for the arms to the chest
		local clavicleBreathe = SSR_rig.makeHelper (posChest + [0,-boneSize,0]) boneSize (prefix + "clavicleParent") cross:true
		SSR_rig.objAlign clavicleBreathe boneHBack --boneChest
		local clavicleBreatheNull = SSR_rig.nullTransform clavicleBreathe
		clavicleBreatheNull.parent = chestControl --boneChest

	-----------------------------------------------------------------------------
	--
	--		Splike IK constraints
	--
	-----------------------------------------------------------------------------

		-- Helper objects to interpolate the rest of the spine
		local hInterpH = point pos:posHBack name:(prefix + "upperBack_ConstraintUp") size:(boneSize)
		local hInterpL = point pos:posHBack name:(prefix + "upperBack_ConstraintDown") size:(boneSize)
		hInterpH.parent = chestControl
		hInterpL.parent = backControl
		
		local lInterpH = point pos:posLBack name:(prefix + "lowerBack_ConstraintUp") size:(boneSize)
		local lInterpL = point pos:posLBack name:(prefix + "lowerBack_ConstraintDown") size:(boneSize)
		lInterpH.parent = backControl
		lInterpL.parent = lowControl
		
		-- Subhelpers that will be interpolated between the ones above
		local hBackHelper = point name:(prefix + "upperBack_Constraint") size:(boneSize * 0.5)
		local lBackHelper = point name:(prefix + "lowerBack_Constraint") size:(boneSize * 0.5)
		
		SSR_rig.posConTwo hBackHelper hInterpH hInterpL
		lowCon = SSR_rig.posConTwo lBackHelper lInterpH lInterpL
		lowCon.weight[1] = 25
		lowCon.weight[2] = 75
		
		-- Parent the IK Spline helpers to our objects
		spineIK.helper_list[1].parent = lowControl
		spineIK.helper_list[3].parent = backControl
		spineIK.helper_list[5].parent = chestControl 
		spineIK.helper_list[2].parent = lBackHelper
		spineIK.helper_list[4].parent = hBackHelper
		
	-----------------------------------------------------------------------------
	--
	--		Connecting up the bones using constraints
	--
	-----------------------------------------------------------------------------
		
		-- We create the helpers the bones will be locked to
		boneHipH 	= SSR_rig.curveHelper (boneSize * 0.5) spineShape posHip
		boneLBackH 	= SSR_rig.curveHelper (boneSize * 0.5) spineShape posLBack
		boneMBackH 	= SSR_rig.curveHelper (boneSize * 0.5) spineShape posMBack
		boneHBackH 	= SSR_rig.curveHelper (boneSize * 0.5) spineShape posHBack
		boneChestH 	= SSR_rig.curveHelper (boneSize * 0.5) spineShape posChest
		boneChestH.pos.controller.percent = 100
		boneHipH.name = prefix + "hip_constraint"
		boneLBackH.name = prefix + "lowerBack_constraint"
		boneMBackH.name = prefix + "upperBack_constraint"
		boneHBackH.name = prefix + "chest_constraint"
		boneChestH.name = prefix + "chestEnd_constraint"
		
		-- Constraint fix
		--! Hack, does not give exact positions!
		if boneSize < 1.0 then
		(
			boneLBackH.pos.controller.percent = 25.0
			boneMBackH.pos.controller.percent = 50.0
			boneHBackH.pos.controller.percent = 75.0
		)
		
		-- The bones are then constrained to these helpers
		SSR_rig.posCon boneHip boneHipH
		SSR_rig.posCon boneLback boneLBackH
		SSR_rig.posCon boneMback boneMBackH
		SSR_rig.posCon boneHback boneHBackH
		SSR_rig.posCon boneHBackEnd boneChestH
		
		-- Front fin should point towards the belly, check the flip setting
		lookCon1 = SSR_rig.lookCon boneHip boneLBackH
		lookCon1.StoUp_axis = 1 ; lookCon1.StoUP_axisFlip = true
		lookCon2 = SSR_rig.lookCon boneLBack boneMBackH
		lookCon2.StoUp_axis = 1 ; lookCon2.StoUP_axisFlip = false
		lookCon3 = SSR_rig.lookCon boneMBack boneHBackH
		lookCon3.StoUp_axis = 1 ; lookCon3.StoUP_axisFlip = false
		lookCon4 = SSR_rig.lookCon boneHBack bonechestH
		lookCon4.StoUp_axis = 1 ; lookCon4.StoUP_axisFlip = false
		
		-- Next, we need to stabilize the orientation of the bones
		SSR_rig.rotCon boneHipH lowControl 
		SSR_rig.rotCon boneLBackH backControl 
		SSR_rig.rotCon bonechestH chestControl 
		
		local boneMBackHCon = Orientation_Constraint()
		boneMBackHCon.appendTarget chestControl 33.5
		boneMBackHCon.appendTarget backControl 66.5
		boneMBackH.rotation.controller = boneMBackHCon
		
		local boneHBackHCon = Orientation_Constraint()
		boneHBackHCon.appendTarget chestControl 66.5
		boneHBackHCon.appendTarget backControl 33.5
		boneHBackH.rotation.controller = boneHBackHCon
		
		-- A couple more helpers for the upnodes of the interpolated helpers
		local hBackStable = point name:(prefix + "chest_upVector") pos:(posHBack + [0, boneSize*30.0, 0] ) size:(boneSize * 0.2)
		local mBackStable = point name:(prefix + "upperBack_upVector") pos:(posMBack + [0, boneSize*30.0, 0] ) size:(boneSize * 0.2)
		local lBackStable = point name:(prefix + "lowerBack_upVector") pos:(posLBack + [0, boneSize*30.0, 0] ) size:(boneSize * 0.2)
		local hipStable = point name:(prefix + "hip_upVector") pos:(posHip + [0, boneSize*30.0, 0] ) size:(boneSize * 0.2)
		
		--SSR_rig.objAlign target source
		SSR_rig.objAlign hBackStable boneHBackH
		SSR_rig.objAlign mBackStable boneMBackH
		SSR_rig.objAlign lBackStable boneLBackH
		SSR_rig.objAlign hipStable boneHipH
		
		hBackStable.parent = boneHBackH
		mBackStable.parent = boneMBackH
		lBackStable.parent = bonelBackH
		hipStable.parent = boneHipH
		
		SSR_rig.setUpNode boneHip.rotation.controller hipStable
		SSR_rig.setUpNode boneLBack.rotation.controller lBackStable
		SSR_rig.setUpNode boneMBack.rotation.controller mBackStable
		SSR_rig.setUpNode boneHBack.rotation.controller hBackStable
		
		--	Constrain end End to Chest controller
		EndCon = SSR_rig.rotCon boneHBackEnd chestControl
		EndCon.relative = on
		
	-----------------------------------------------------------------------------
	--
	--		Controller custom attributes
	--
	-----------------------------------------------------------------------------
		
		-- Add custom attributes for the hip controller
		local hipMod = 'PEN_Attribute_Holder 2'()
		addModifier lowControl hipMod
		
		local hipAtt = attributes controls
		(
			parameters params rollout:handR
			(
				X_Pos		type:#float ui:PXSp
				Y_Pos		type:#float ui:PYSp
				Z_Pos		type:#float ui:PZSp
				X_Rot		type:#angle ui:RXSp
				Y_Rot		type:#angle ui:RYSp
				Z_Rot		type:#angle ui:RZSp
			)
			rollout handR "Transforms"
			(
				spinner PXSp 	"Translate X " 	range:[-1000000, 1000000, 0] align:#right
				spinner PYSp 	"Translate Y " 	range:[-1000000, 1000000, 0] align:#right
				spinner PZSp 	"Translate Z " 	range:[-1000000, 1000000, 0] align:#right
				spinner RXSp 	"Rotate X " 	range:[-1000000, 1000000, 0] align:#right
				spinner RYSp 	"Rotate Y " 	range:[-1000000, 1000000, 0] align:#right
				spinner RZSp 	"Rotate Z " 	range:[-1000000, 1000000, 0] align:#right
				label lbl01			"Hold ALT for fine control" offset:[0,10]
			)
		)		
		custAttributes.add hipMod hipAtt

		-- Connect up custom attributes
		SSR_rig.freezeTransform lowControl
		
		hipMod.controls.X_Pos.controller = \
		lowControl.pos.controller.Zero_POS_XYZ.controller.X_Position.controller
		hipMod.controls.Y_Pos.controller = \
		lowControl.pos.controller.Zero_POS_XYZ.controller.Y_Position.controller
		hipMod.controls.Z_Pos.controller = \
		lowControl.pos.controller.Zero_POS_XYZ.controller.Z_Position.controller
		
		hipMod.controls.X_Rot.controller = \
		lowControl.rotation.controller.Zero_Euler_XYZ.controller.X_Rotation.controller
		hipMod.controls.Y_Rot.controller = \
		lowControl.rotation.controller.Zero_Euler_XYZ.controller.Y_Rotation.controller
		hipMod.controls.Z_Rot.controller = \
		lowControl.rotation.controller.Zero_Euler_XYZ.controller.Z_Rotation.controller
			

			-- Custom attributes for the back controller
		local backMod = 'PEN_Attribute_Holder 2'()
		addModifier backControl backMod
		custAttributes.add backMod hipAtt
		
		-- Connect up custom attributes
		SSR_rig.freezeTransform backControl
		
		backMod.controls.X_Pos.controller = \
		backControl.pos.controller.Zero_POS_XYZ.controller.X_Position.controller
		backMod.controls.Y_Pos.controller = \
		backControl.pos.controller.Zero_POS_XYZ.controller.Y_Position.controller
		backMod.controls.Z_Pos.controller = \
		backControl.pos.controller.Zero_POS_XYZ.controller.Z_Position.controller

		backMod.controls.X_Rot.controller = \
		backControl.rotation.controller.Zero_Euler_XYZ.controller.X_Rotation.controller
		backMod.controls.Y_Rot.controller = \
		backControl.rotation.controller.Zero_Euler_XYZ.controller.Y_Rotation.controller
		backMod.controls.Z_Rot.controller = \
		backControl.rotation.controller.Zero_Euler_XYZ.controller.Z_Rotation.controller

		-- Custom attributes for the chest controller
		local chestMod = 'PEN_Attribute_Holder 2'()
		addModifier chestControl chestMod
		custAttributes.add chestMod hipAtt
		
		-- Connect up custom attributes
		SSR_rig.freezeTransform chestControl
		
		chestMod.controls.X_Pos.controller = \
		chestControl.pos.controller.Zero_POS_XYZ.controller.X_Position.controller
		chestMod.controls.y_Pos.controller = \
		chestControl.pos.controller.Zero_POS_XYZ.controller.Y_Position.controller
		chestMod.controls.Z_Pos.controller = \
		chestControl.pos.controller.Zero_POS_XYZ.controller.Z_Position.controller

		chestMod.controls.X_Rot.controller = \
		chestControl.rotation.controller.Zero_Euler_XYZ.controller.X_Rotation.controller
		chestMod.controls.Y_Rot.controller = \
		chestControl.rotation.controller.Zero_Euler_XYZ.controller.Y_Rotation.controller
		chestMod.controls.Z_Rot.controller = \
		chestControl.rotation.controller.Zero_Euler_XYZ.controller.Z_Rotation.controller
			
		-- Custom attributes for the hip controller
		local hipMod = 'PEN_Attribute_Holder 2'()
		addModifier hipControl hipMod
		custAttributes.add hipMod hipAtt
		
		-- Connect up custom attributes
		SSR_rig.freezeTransform hipControl
		
		hipMod.controls.X_Pos.controller = \
		hipControl.pos.controller.Zero_POS_XYZ.controller.X_Position.controller
		hipMod.controls.y_Pos.controller = \
		hipControl.pos.controller.Zero_POS_XYZ.controller.Y_Position.controller
		hipMod.controls.Z_Pos.controller = \
		hipControl.pos.controller.Zero_POS_XYZ.controller.Z_Position.controller

		hipMod.controls.X_Rot.controller = \
		hipControl.rotation.controller.Zero_Euler_XYZ.controller.X_Rotation.controller
		hipMod.controls.Y_Rot.controller = \
		hipControl.rotation.controller.Zero_Euler_XYZ.controller.Y_Rotation.controller
		hipMod.controls.Z_Rot.controller = \
		hipControl.rotation.controller.Zero_Euler_XYZ.controller.Z_Rotation.controller
	

		return #(lowControl, backControl, chestControl, hipControl)
	)
)

SSR_part = SSR_part()