global SSR_gen

struct SSR_gen
(
	----------------------------------------------------------
	-- verbose: 
	--	writes text to the Listener, adding preceeding "--" and line break
	----------------------------------------------------------
	function verbose verboseText =
	(
		format "-- %\n" verboseText
	),
	
	
	-----------------------------------------------------------------------------
	-- Finds all characters in the scene by root nodes
	-----------------------------------------------------------------------------
	function getCharArray =
	(
		local characters = #()
		local sceneChars = $'*_root_CTL' 
		
		for o in sceneChars do
		(
			-- strip away _root_CTL
			local charName = substituteString o.name "_root_CTL" "" 
			append characters charName
		)
		
		return characters
	),
	
	----------------------------------------------------------
	-- collectObjects: 
	--	Collects objects according to string pattern, with optional parent object
	----------------------------------------------------------
	function collectObjects searchString parent:undefined = 
	(
		local objList = #()
		for o in objects do
		(
			if (matchPattern o.name pattern:searchString ignoreCase:true) then append objList o
		)
		return objList
	),
	
	----------------------------------------------------------
	-- collectControllers: 
	--	Returns controllers according to search string
	----------------------------------------------------------
	function collectControllers searchString = 
	(
		local controllerList = #()
		local objList = SSR_gen.collectObjects searchString
		
		if objList.count > 0 then controllerList = SSR_gen.filterControllers objList strict:true
		
		return controllerList
	),
	
	----------------------------------------------------------
	-- filterControllers: 
	--	Returns only controller objects from the input object array
	----------------------------------------------------------
	function filterControllers objList strict:true = 
	(
		local controllerList = #()
		for o in objList do 
		(
			if (matchPattern o.name pattern:"*_CTL" ignoreCase:true) then
			(
				if strict == true then
				(
					if (classOf o == splineShape) and (classOf o.modifiers[1] as string == "PEN_Attribute_Holder 2") then
					(
						append controllerList o
					)
				) else (
					append controllerList o
				)
			)
		)
		return controllerList
	),
	
	----------------------------------------------------------
	-- scriptedTransforms: 
	--	Prints current transforms to the listener, can be copied back to a script
	----------------------------------------------------------
	function scriptedTransforms objList =
	(
		for o in objList do
		(
			local str = ("-- Transforms for $'" + o.name + "':\n\n")
			str = str + ("$'" + o.name + "'.parent = $'" + o.parent.name + "'\n\n")
			str = str + ("$'" + o.name + "'.position.controller.X_position = " + o.position.controller.X_position as string + "\n" )
			str = str + ("$'" + o.name + "'.position.controller.Y_position = " + o.position.controller.Y_position as string + "\n" )
			str = str + ("$'" + o.name + "'.position.controller.Z_position = " + o.position.controller.Z_position as string + "\n\n" )
			str = str + ("$'" + o.name + "'.rotation.controller.X_rotation = " + o.rotation.controller.X_rotation as string + "\n" )
			str = str + ("$'" + o.name + "'.rotation.controller.Y_rotation = " + o.rotation.controller.Y_rotation as string + "\n" )
			str = str + ("$'" + o.name + "'.rotation.controller.Z_rotation = " + o.rotation.controller.Z_rotation as string + "\n\n" )
			str = str + ("$'" + o.name + "'.scale = " + o.scale as string + "\n\n")
			format "%" str
		)
		return true
	),
	
	----------------------------------------------------------
	-- clearFolder: 
	--	removes all files with matching wildcard from folder
	----------------------------------------------------------
	function eraseFolder thePath wildCard = 
	(
		local thePath = trimRight thepath "\\"
		thePath = thePath + "\\"
				
		local allFiles = getFiles (thePath + wildCard)
		SSR_gen.verbose ("Erasing " + allFiles.count as string + " animation files in folder: " + thePath) 
		
		for theFile in allFiles do deleteFile theFile
			
		return true
	),
	
	----------------------------------------------------------
	-- incrementVersion: 
	--	Returns incremented filename if versioning is "_v###_"
	----------------------------------------------------------
	function incrementVersion fileName includeDot:false returnVersion:false =
	(
		local fileName
		local versionString = ""
		local doBreak = false
		
		local digits = #("1","2","3","4","5","6","7","8","9","0",".")
		local searchLength = digits.count
		-- includedot (fraction/sub-versions such as v01.02) is not implemented properly yet
		if includeDot == false then searchLength = digits.count-1
		
		local nameElements = filterString fileName "_"
		
		for n in nameElements do
		(
			-- Look for an element starting with "v"
			if n[1] == "v" then
			(
				-- Found a "_v*" element, are the rest of the characters digits?
				-- ! this check could be done as ((substring n 2 n.count) as float), investigate that option
				for c = 2 to n.count do
				(
					numCheck = false
					for d = 1 to searchLength do
					(
						matchCheck = matchPattern (n[c]) pattern:(digits[d])
						if matchCheck == true then numCheck = true
					)
					if numCheck == false then doBreak = true
				)
				if doBreak == false then versionString = n
			)
		)
		
		if doBreak == true then
		(
			local warn = "Could not find version number in filename.\n"
			warn = warn + "The file version should be formatted like this: _v##_\n\n"
			warn = warn + "Please re-save your file with a version number."
			
			messageBox warn title:"Error!"
			return false	
		)
		
		-- Create incremented string
		local oldVersion = substring versionString 2 versionString.count
		if returnVersion == true then return oldVersion
		
		local newVersion = formattedPrint (oldVersion as integer + 1) format:("." + (versionString.count-1) as string + "d")
		fileName = substitutestring filename versionString ("v" + newVersion)
		
		return fileName
	),
	
	----------------------------------------------------------
	-- alienProbe: 
	--	checks and reports if the rig nodes has "alien" rigs as parents or as children
	----------------------------------------------------------
	function alienProbe prefix =
	(
		local alienList = #()
		local objList = execute ("$'" + prefix + "_*'")
		
		SSR_gen.verbose ("Checking " + objList.count as string + " object with prefix '" + prefix + "'.")
		
		for o in objList do
		(
			-- alien parent?
			if o.parent != undefined then
			(
				local filterName = filterString o.parent.name "_"
				if (toLower filterName[1]) != (toLower prefix) then SSR_gen.verbose ("$'" + o.name + "' parented to alien $'" + o.parent.name + "'.")
			)
			-- alien children?
			if o.children.count > 0 then for c in o.children do
			(
				local filterName = filterString c.name "_"
				if (toLower filterName[1]) != (toLower prefix) then SSR_gen.verbose ("$'" + o.name + "' has alien child $'" + c.name + "'.")
			)
			-- alien transform constraints?
			if classOf o.transform.controller == Link_Constraint do
			(
				SSR_gen.verbose ("$'" + o.name + "' has the following link constraints:")
				local linkCon = o.transform.controller
				local numTargets = linkCon.getNumTargets()
				
				for c = 1 to numTargets do
				(
					if linkCon.getNode c != undefined then
						format "%\n" ("$'" + o.name +"'.transform.controller.addTarget $'" + ((linkCon.getNode c).name) + "' " + (linkCon.getFrameNo c) as string)
					else 
						format "%\n" ("$'" + o.name +"'.transform.controller.addWorld frameNo:" + (linkCon.getFrameNo c) as string)
				)
			)
			-- alien position constraints?
			/*if classOf o.transform.controller == Link_Constraint do
			(
				SSR_gen.verbose ("$'" + o.name + "' has the following link constraints:")
				local linkCon = o.transform.controller
				local numTargets = linkCon.getNumTargets()
				
				for c = 1 to numTargets do
				(
					if linkCon.getNode c != undefined then
						format "%\n" ("$'" + o.name +"'.transform.controller.addTarget $'" + ((linkCon.getNode c).name) + "' " + (linkCon.getFrameNo c) as string)
					else 
						format "%\n" ("$'" + o.name +"'.transform.controller.addWorld frameNo:" + (linkCon.getFrameNo c) as string)
				)
			)*/
		)
		
		return true
	),
	
	----------------------------------------------------------
	-- createLayer: 
	--	creates a named layer, returns existing named layer
	----------------------------------------------------------
	function createLayer layerName activeLayer:true =
	(
		local newLayer = layermanager.getlayerfromname layerName
		if newLayer == undefined then newLayer = layermanager.newLayerfromname layerName
		if activeLayer == true then newLayer.current = true
		return newLayer
	),
	
	----------------------------------------------------------
	-- putToLayer: 
	--	puts objects to layer, creating the layer if it doesn't exist
	----------------------------------------------------------
	function putToLayer objList layerName =
	(
		local newLayer = SSR_gen.createLayer layerName activeLayer:false
		for o in objList do newLayer.addNode o
	),
	
	
	----------------------------------------------------------
	-- renameLayer: 
	--	renames Rig Locators Layer
	----------------------------------------------------------
	function renameLayer oldName newName =
	(
		local wasNamed = false
		local theLayer = undefined
		if oldName == " Rig Locators" then
		(
			try (
				theLayer = layermanager.getlayerfromname "Rig Locators"
				wasNamed = theLayer.setName newName
			) catch SSR_gen.verbose (getCurrentException())
		) else (
			try (
				theLayer = layermanager.getlayerfromname oldName
				wasNamed = theLayer.setName newName
			) catch print (getCurrentException())
		)

		if wasNamed == true then return true
		else verbose ("Could not rename layer from: " + oldName + " to: " + newName)
	),
	
	----------------------------------------------------------
	-- renameLayers: 
	--	changes prefix of scene layers
	----------------------------------------------------------
	function renameLayers oldName newName =
	(
		if newName != "" then
		(
			for i = 0 to layerManager.count-1 do
			(
				curlayer = layerManager.getLayer i
				if (filterString curLayer.name " ")[1] == oldName then
				(
					curLayer.setName (substituteString curLayer.name (oldName + " ") (newName + " "))
				)
			)
		) else ( return false )
		return true
	),
	
	----------------------------------------------------------
	-- deleteLayer: 
	--	deletes layers matching current string
	----------------------------------------------------------
	function deleteLayer deleteString deleteContent:false =
	(
		-- Parse through layers, see if their names match our string
		for c = (LayerManager.count-1) to 1 by -1 do
		(
			iLayer = LayerManager.getLayer c
			if iLayer != undefined then
			(
				if matchPattern iLayer.name pattern:deleteString then
				(
					-- Is this the current layer? If it is, let the default layer be current
					if (LayerManager.getLayer c).current == true then (LayerManager.getLayer 0).current = true
					
					-- Are there any objects on the layer?
					iLayer.nodes &layerNodes
					
					if layerNodes.count != 0 and deleteContent == false then
					(
						SSR_gen.verbose ("Layer '" + iLayer.name + "' has objects. Cannot delete!")
						return false
					) else if layerNodes.count != 0 and deleteContent == true then
					(
						for o in layerNodes do delete o
					)
					
					SSR_gen.verbose ("Deleting layer '" + iLayer.name + "'.")
					LayerManager.deleteLayerByName iLayer.name
				)
			)
		)
		return true
	),
	
	----------------------------------------------------------
	-- deleteAsset: 
	--	deletes all objects that belong to named asset
	----------------------------------------------------------
	function deleteAsset assetName = 
	(
		SSR_gen.deleteLayer (assetName + " *") deleteContent:true
		execute ("select $'" + assetName + "_*'")
		delete $selection
	),
	
	----------------------------------------------------------
	-- renameObjects: 
	--	changes prefix of all objects in the scene
	----------------------------------------------------------
	function renameObjects oldName newName =
	(
		if newName != "" then
		(
			for o in $objects do
			(
				if (filterString o.name "_")[1] == oldName then
				(
					--print o.name
					o.name = substituteString o.name (oldName + "_") (newName + "_")
				)
			)
		) else ( return false ) 
		return true
	),
		
	----------------------------------------------------------
	-- renameRig: 
	--	renames character rig, including objects and layers
	----------------------------------------------------------
	function renameRig charName =
	(
		rollout renameRoll "Rename Character Rig"
		(
			local charName
			
			edittext newCharName "New Name: " text:""
			button doRename	"Rename"
			
			on doRename pressed do
			(
				--print (newCharName.text + " " + charName )
				isItDone = SSR_gen.renameObjects charName newCharName.text
				isItDone = SSR_gen.renameLayers charName newCharName.text
				
				-- Fix rig window
				SSR_charName = newCharName.text
				try (superRigMain_UI.editCharName.text = newCharName.text)
				catch()
				
				--if isItDone == true then DestroyDialog renameRoll
			)
		)
		CreateDialog renameRoll 250 50
		renameroll.newcharname.text = charname
		renameroll.charname = charname
	)
	
)

SSR_gen = SSR_gen()