-----------------------------------------------------------------------------
--
--	Functions related to rig locators
--
-----------------------------------------------------------------------------

global SSR_loc

struct SSR_loc
(
	----------------------------------------------------------
	-- collectLocators: 
	--	Collects locators according to string pattern, with parent charName
	----------------------------------------------------------
	function collectLocators charName stringPattern = 
	(
		local objList = #()
		local childObjects = execute ("$'" + charName + "_Locators_TOP'.children")
		
		for o in childObjects do
		(
			if matchPattern o.name pattern:stringPattern then append objList o
		)
			
		if objList.count == 0 then return false
		return objList
	),
	
	----------------------------------------------------------
	-- getGUI: 
	--	fetches rig ui options into an array
	----------------------------------------------------------
	function getGUI debug:false = 
	(	
		-- Check if Builder window is open
		if superRigOptions_UI != undefined then
		(
			local settingArray = #()
			
			-- Parse through controls, check type and store values
			for c = 1 to superRigOptions_UI.controls.count do
			(
				local element = superRigOptions_UI.controls[c]
				
				-- Collect options and store in string
				case (classOf element) of
				(
					comboBoxControl: 
					(
						append settingArray #(element.name,element.selected)
					)
					checkBoxControl:
					(
						append settingArray #(element.name,element.state)
					)
					spinnerControl:
					(
						append settingArray #(element.name,element.value)
					)
					default: 
					(
						if debug == true then SSR_gen.verbose ("Skipping element " + classOf element as string)
					)
				)
			)
			return settingArray
		) else (
			SSR_gen.verbose "getGUI: No rollout found."
			return false
		)
		
		return true
	),
	
	----------------------------------------------------------
	-- storeGUI: 
	--	stores GUI settings to root object
	----------------------------------------------------------
	function storeGUI charName debug:false =
	(
		-- Is there a locator top that belongs to the character?
		local locObj = execute ("$'" + charName + "_Locators_TOP")
		
		if locObj != undefined then
		(
			SSR_gen.verbose ("Storing GUI settings to root locator of '" + charName + "'.")
			
			local settingArray = SSR_loc.getGUI debug:debug
			for a in settingArray do
			(
				setUserProp locObj a[1] a[2]
			)
		) else (
			SSR_gen.verbose "Locator object not found, can not store GUI settings."
		)	
	),
	
	----------------------------------------------------------
	-- setGUI: 
	--	puts GUI settings from root object and stores them in the GUI
	----------------------------------------------------------
	function setGUI charName debug:false =
	(
		-- Is there a locator top that belongs to the character?
		local locObj = execute ("$'" + charName + "_Locators_TOP")
		
		if locObj != undefined and superRigOptions_UI != undefined then
		(
			-- Are there settings stored in this character?
			if (getUserPropBuffer locObj).count > 5 then
			(
				--SSR_gen.verbose ((getUserPropBuffer locObj).count as string)
				-- Parse through controls, check type and get values
				for c = 1 to superRigOptions_UI.controls.count do
				(
					local element = superRigOptions_UI.controls[c]
					
					case (classOf element) of
					(
						comboBoxControl: 
						(
							for c = 1 to element.items.count do
							(
								if (getUserProp locObj element.name) == (element.items[c]) do
								(
									element.selection = c
									if debug == true then SSR_gen.verbose ("Found comboBox match: " + (getUserProp locObj element.name))
								)
							)
						)
						checkBoxControl:
						(
							if debug == true then SSR_gen.verbose (element.name + " set to " + \
								(getUserProp locObj element.name) as string)
							element.state = getUserProp locObj element.name
						)
						spinnerControl:
						(
							if debug == true then SSR_gen.verbose (element.name + " set to " + \
								(getUserProp locObj element.name) as string)
							element.value = getUserProp locObj element.name
						)
						default: 
						(
							if debug == true then SSR_gen.verbose ("Skipping element " + \
								classOf element as string)
						)
					)
				)
			) else (
				SSR_gen.verbose ("No settings found on '" + locObj.name + "'. Could not restore.")
			)

			superRigOptions_UI.doState()
			
		) else (
			SSR_gen.verbose "Locator object or GUI not found, can not fetch GUI settings."
		)
	),
	
	-----------------------------------------------------------------------------
	--	resizeLocators: 
	--	resizes all locators
	-----------------------------------------------------------------------------
	function resizeLocators charName newSize:0 =
	(	
		--print newSize
		local locatorHead = (SSR_loc.collectLocators charName "Locator_Head_HeadTop")[1]
		local scaleSize = newSize/locatorHead.pos.z
		local locatorList = (SSR_loc.collectLocators charName "Locator_*")
		for o in locatorList do o.pos = o.pos * scaleSize
	),
	
	-----------------------------------------------------------------------------
	--	loadLocators: 
	--	creates rig locators based on information in a rig file
	-----------------------------------------------------------------------------
	function loadLocators rigFileName charName charSize:0 =
	(	
		SSR_gen.verbose ("Charsize: " + charSize as string)
			
		local storedCurrent = LayerManager.current
		local charName
		
		local rigSize = 0.0
		local fileCharSize = 0.0 -- Size loaded from .rig file
		local charSize -- Size the function is called from, 0 means use .rig file size
		if charSize != 0 then rigSize = charSize
		
		local rigFile = openFile rigFileName
		if rigFile != undefined then
		(
			-- First, delete existing locators
			SSR_gen.verbose ("Deleting existing locators for " + charName)
			SSR_loc.killLocators charName
			
			-- Get the new character name
			local curLine = readLine rigFile
			local filteredString = filterString curLine ":'"
			charName = filteredString[2]

			-- We create a rig layer, and make it active. 
			local locatorLayerName = charName + " Rig Locators"
			local locatorLayer = layermanager.getlayerfromname locatorLayerName
			if locatorLayer == undefined then locatorLayer = layermanager.newLayerfromname locatorLayerName
			locatorLayer.current = true
			
			-- This bit could be more elegant - search for some variables (such as rig size) before doing anything else
			while not eof rigFile do
			(
				-- Parse our rigFile, line by line
				local curLine = readLine rigFile
				--SSR_gen.verbose curLine
				
				-- Check to see which string starts current line
				local filteredString = filterString curLine ":'"
				case filteredString[1] of
				(
					"size": (
						fileCharSize = filteredString[2] as float
						if charSize == 0 then rigSize = fileCharSize
					)
					"locator": (
						local ex = "SSR_rig.makeHelper (" + filteredString[3] + "*" + rigSize as string + ") " + \
							(rigSize*0.02) as string + " \"" + filteredString[2] + "\" cross:true drawontop:true"
						execute ex
					)
					"line": (
						local startObj = execute ("$'" + filteredString[3])
						local endObj = execute ("$'" + filteredString[4])
						
						if startObj != undefined and endObj != undefined then SSR_loc.connectLine startObj endObj filteredString[2]
					)
					default: (
						SSR_gen.verbose "Unknown Line!"
					)
				) -- case filteredString[1] of
			) -- while not eof rigFile do
			
			-- All locators will be parented under a single point helper
			local locators = #()
			for o in $'Locator_*' do append locators o
			for o in $'LocatorLine_*' do append locators o
			locatorParent = SSR_rig.makeHelper [0,0,0] 20.0 (charName + "_Locators_TOP") centerMarker:true
			for o in locators do o.parent = locatorParent
			
			close rigFile
		) else (
			--	Restore original current layer
			storedCurrent.current = true
			return false
		)
		--	Restore original current layer
		storedCurrent.current = true
		completeRedraw() 
		
		return charName
	),
	
	-----------------------------------------------------------------------------
	--	saveLocators: 
	--	saves rig locators to a rig file
	-----------------------------------------------------------------------------
	function saveLocators rigFileName charName =
	(
		-- Get character locators
		local charLocators = execute ("$'" + charName + "_Locators_TOP")
		
		-- Get locators
		local Locators = #()
		for o in charLocators do 			
			if matchPattern o.name pattern:"Locator_*" then append Locators o
		
		-- Calculate height of character
		local charSize= 0
		for o in Locators do if o.pos.z > charSize then charSize = o.pos.z
		--SSR_gen.verbose ("Character size: " + charSize as string)
		
		-- Get lines
		local LocatorLines = #()
		for o in charLocators do
			if matchPattern o.name pattern:"LocatorLine_*" then append LocatorLines o
		
		-- If file exists, delete it
		if doesFileExist rigFileName == true then deleteFile rigFileName
		local rigFile = createFile rigFileName
		
		-- Save out character name
		format "name:%\n" charName to:rigFile
		
		-- Save out height
		format "size:%\n" (charSize as string) to:rigFile
		
		-- Save out locators
		for o in Locators do 
		(
			local helperName = "'" + o.name + "'"
			local helperPos = o.pos / charSize
			format "%\n" ("locator:" + helperName + helperPos as string) to:rigFile
		)
		
		-- Save out lines
		for o in LocatorLines do 
		(
			if classOf o == SplineShape then
			(
				local helperName = "'" + o.name + "'"
				
				local knot1pos = getKnotPoint o 1 1
				local knot1name = undefined
				for l in Locators do 
				(
					if l.pos == knot1pos then knot1name = "'" + l.name + "'"
				)
				
				local knot2pos = getKnotPoint o 1 2
				local knot2name = undefined
				for l in locators do 
				(
					if l.pos == knot2pos then knot2name = "'" + l.name + "'"
				)
				
				if knot1name != undefined or knot2name != undefined then
				(
					format "%\n" ("line:" + helperName + knot1name + knot2name) to:rigFile
				)
			)
			if classOf o == Null then
			(
				local helperName = "'" + o.name + "'"
				local parentName = "'" + o.parent.name + "'"
				local constraintName = "'" + (o.pos.controller.getNode 1).name + "'"
				format "%\n" ("line:" + helperName + parentName + constraintName) to:rigFile
			)
		)
		
		close rigFile
		
		SSR_gen.verbose ("Finished saving rig at: " + rigFileName)
		
		return true
	),

	-----------------------------------------------------------------------------
	--	createLocators: 
	--	creates locators for specific parts of the rig
	-----------------------------------------------------------------------------
	function createLocators bodyPart rigSize charName =
	(
		local headHelpers = #( \
			#("Locator_Head_Neck", [0,0.0396166,0.837608]), 
			#("Locator_Head_MidNeck", [0,0.0169785,0.871565]),
			#("Locator_Head_Head", [0,0.011319,0.92816]),
			#("Locator_Head_Jaw", [0,-0.011319,0.933819]),
			#("Locator_Head_Mouth", [0,-0.0509356,0.905522]),
			#("Locator_Head_InMouth", [0,-0.0311273,0.908352]),
			#("Locator_Head_HeadTop", [0,0.011319,1]),
			#("Locator_Head_R_Eye", [-0.0169785,-0.0339571,0.950798]),
			#("Locator_Head_L_Eye", [0.0169785,-0.0339571,0.950798]) 	)
		
		local headLines = #( \
			#("LocatorLine_Head", "Locator_Head_Neck", "Locator_Head_MidNeck"),
			#("LocatorLine_Head", "Locator_Head_MidNeck", "Locator_Head_Head"),
			#("LocatorLine_Head", "Locator_Head_Head", "Locator_Head_R_Eye"),
			#("LocatorLine_Head", "Locator_Head_Head", "Locator_Head_L_Eye"),
			#("LocatorLine_Head", "Locator_Head_Head", "Locator_Head_Jaw"),
			#("LocatorLine_Head", "Locator_Head_Head", "Locator_Head_HeadTop"),
			#("LocatorLine_Head", "Locator_Head_Jaw", "Locator_Head_InMouth"),
			#("LocatorLine_Head", "Locator_Head_InMouth", "Locator_Head_Mouth")		)
		
		local spineHelpers = #( \
			#("Locator_Spine_Hip", [0,0.0452761,0.537654]),
			#("Locator_Spine_LowBack", [0,0.0169785,0.605568]),
			#("Locator_Spine_MidBack", [0,0.022638,0.667822]),
			#("Locator_Spine_UpperBack", [0,0.0565951,0.735736]),
			#("Locator_Spine_Chest", [0,0.0565951,0.797991]),
			#("Locator_Chest_R_Breathe", [-0.0563452,-0.0466538,0.763435]),
			#("Locator_Chest_L_Breathe", [0.0563452,-0.0466538,0.763435]),
			#("Locator_Belly_Breathe", [0,-0.0458522,0.608716])		)
		
		local spineLines = #( \	
			#("LocatorLine_Spine", "Locator_Spine_Hip", "Locator_Spine_LowBack"),
			#("LocatorLine_Spine", "Locator_Spine_LowBack", "Locator_Spine_MidBack"),
			#("LocatorLine_Spine", "Locator_Spine_MidBack", "Locator_Spine_UpperBack"),
			#("LocatorLine_Spine", "Locator_Spine_UpperBack", "Locator_Spine_Chest"),
			#("LocatorLine_Chest", "Locator_Spine_UpperBack", "Locator_Chest_R_Breathe"),
			#("LocatorLine_Chest", "Locator_Spine_UpperBack", "Locator_Chest_L_Breathe"),
			#("LocatorLine_Chest", "Locator_Spine_LowBack", "Locator_Belly_Breathe")		)
		
		local armHelpers = #( \	
			#("Locator_Arm_L_Clavicle", [0.022638,0.0169785,0.837608]),
			#("Locator_Arm_L_Shoulder", [0.0848927,0.0169785,0.820629]),
			#("Locator_Arm_L_Elbow", [0.208779,0.022638,0.6907]),
			#("Locator_Arm_L_Hand", [0.316933,0,0.57727]),
			#("Locator_Arm_R_Clavicle", [-0.022638,0.0169785,0.837608]),
			#("Locator_Arm_R_Shoulder", [-0.0848927,0.0169785,0.820629]),
			#("Locator_Arm_R_Elbow", [-0.208779,0.022638,0.6907]),
			#("Locator_Arm_R_Hand", [-0.316933,0,0.57727])			)
			
		local armLines = #( \
			#("LocatorLine_Arm_L", "Locator_Arm_L_Clavicle", "Locator_Arm_L_Shoulder"),
			#("LocatorLine_Arm_L", "Locator_Arm_L_Shoulder", "Locator_Arm_L_Elbow"),
			#("LocatorLine_Arm_L", "Locator_Arm_L_Elbow", "Locator_Arm_L_Hand"),
			#("LocatorLine_Arm_R", "Locator_Arm_R_Clavicle", "Locator_Arm_R_Shoulder"),
			#("LocatorLine_Arm_R", "Locator_Arm_R_Shoulder", "Locator_Arm_R_Elbow"),
			#("LocatorLine_Arm_R", "Locator_Arm_R_Elbow", "Locator_Arm_R_Hand")			)
		
		local legHelpers = #( \	
			#("Locator_Leg_R_Ankle", [-0.101871,0,0.0339571]),
			#("Locator_Leg_R_Knee", [-0.101871,-0.022638,0.271657]),
			#("Locator_Leg_R_Hip", [-0.101871,0.00565951,0.515016]),
			#("Locator_Leg_R_Heel", [-0.101871,0.0339571,0]),
			#("Locator_Leg_R_Toe", [-0.101871,-0.0665388,0.00670209]),
			#("Locator_Leg_R_Tip", [-0.101871,-0.107531,0]),
			#("Locator_Leg_R_SideOut", [-0.124509,-0.0254678,0]),
			#("Locator_Leg_R_SideIn", [-0.0679141,-0.0254678,0]),
			#("Locator_Leg_L_Ankle", [0.101871,0,0.0339571]),
			#("Locator_Leg_L_Knee", [0.101871,-0.022638,0.271657]),
			#("Locator_Leg_L_Hip", [0.101871,0.00565951,0.515016]),
			#("Locator_Leg_L_Heel", [0.101871,0.0339571,0]),
			#("Locator_Leg_L_Toe", [0.101871,-0.0665388,0.00670209]),
			#("Locator_Leg_L_Tip", [0.101871,-0.107531,0]),
			#("Locator_Leg_L_SideOut", [0.124509,-0.0254678,0]),
			#("Locator_Leg_L_SideIn", [0.0679141,-0.0254678,0])		)
		
		local legLines = #( \	
			#("LocatorLine_Leg_R", "Locator_Leg_R_Ankle", "Locator_Leg_R_Knee"),
			#("LocatorLine_Leg_R", "Locator_Leg_R_Knee", "Locator_Leg_R_Hip"),
			#("LocatorLine_Leg_R", "Locator_Leg_R_Heel", "Locator_Leg_R_Tip"),
			#("LocatorLine_Leg_R", "Locator_Leg_R_Tip", "Locator_Leg_R_Toe"),
			#("LocatorLine_Leg_L", "Locator_Leg_L_Ankle", "Locator_Leg_L_Knee"),
			#("LocatorLine_Leg_L", "Locator_Leg_L_Knee", "Locator_Leg_L_Hip"),
			#("LocatorLine_Leg_L", "Locator_Leg_L_Heel", "Locator_Leg_L_Tip"),
			#("LocatorLine_Leg_L", "Locator_Leg_L_Tip", "Locator_Leg_L_Toe")		)
		
		local animalLegHelpers = #( \
			#("Locator_Leg_R_1_Clavicle", [-0.0703447,-0.123975,0.317081]),
			#("Locator_Leg_R_1_Shoulder", [-0.0822832,-0.0889209,0.204196]),
			#("Locator_Leg_R_1_Elbow", [-0.0613759,-0.131024,0.118919]),
			#("Locator_Leg_R_1_Ankle", [-0.0613759,-0.13251,0.0431999]),
			#("Locator_Leg_R_1_Tip", [-0.0613759,-0.185088,0.0]),
			#("Locator_Leg_R_1_Heel", [-0.0613759,-0.132707,0.0]),
			#("Locator_Leg_L_1_Clavicle", [0.0703447,-0.123975,0.317081]),
			#("Locator_Leg_L_1_Shoulder", [0.0822832,-0.0889209,0.204196]),
			#("Locator_Leg_L_1_Elbow", [0.0613759,-0.131024,0.118919]),
			#("Locator_Leg_L_1_Ankle", [0.0613759,-0.13251,0.0431999]),
			#("Locator_Leg_L_1_Tip", [0.0613759,-0.185088,0.0]),
			#("Locator_Leg_L_1_Heel", [0.0613759,-0.132707,0.0])		)
			
		local animalLegLines = #( \		
			#("LocatorLine_Leg_R", "Locator_Leg_R_1_Clavicle", "Locator_Leg_R_1_Shoulder"),
			#("LocatorLine_Leg_R", "Locator_Leg_R_1_Shoulder", "Locator_Leg_R_1_Elbow"),
			#("LocatorLine_Leg_R", "Locator_Leg_R_1_Elbow", "Locator_Leg_R_1_Ankle"),
			#("LocatorLine_Leg_R", "Locator_Leg_R_1_Tip", "Locator_Leg_R_1_Heel"),
			#("LocatorLine_Leg_L", "Locator_Leg_L_1_Clavicle", "Locator_Leg_L_1_Shoulder"),
			#("LocatorLine_Leg_L", "Locator_Leg_L_1_Shoulder", "Locator_Leg_L_1_Elbow"),
			#("LocatorLine_Leg_L", "Locator_Leg_L_1_Elbow", "Locator_Leg_L_1_Ankle"),
			#("LocatorLine_Leg_L", "Locator_Leg_L_1_Tip", "Locator_Leg_L_1_Heel")		)
			
		local locators = #()
		local locatorLines = #()
		local existingParts = 0
			
		if bodyPart == "head" then ( locators = headHelpers; locatorLines = headLines )
		if bodyPart == "spine" then ( locators = spineHelpers; locatorLines = spineLines )
		if bodyPart == "arms" then ( locators = armHelpers; locatorLines = armLines )
		if bodyPart == "legs" then ( locators = legHelpers; locatorLines = legLines )
		if bodyPart == "animallegs" then 
		( 
			locators = animalLegHelpers
			locatorLines = animalLegLines 
			existingParts = findLocators charName "Leg" side:"R"
		)
			
		--	We create a rig layer, and make it active. 
		local storedCurrent = LayerManager.current
		local locatorLayerName = charName + " Rig Locators"
		local locatorLayer = layermanager.getlayerfromname locatorLayerName
		if locatorLayer == undefined then locatorLayer = layermanager.newLayerfromname locatorLayerName
		locatorLayer.current = true	
		
		-- Create a TOP node if no one exists
		local locatorParent = execute ("$'" + charName + "_Locators_TOP'")
		if locatorParent == undefined then 
		(
			locatorParent = SSR_rig.makeHelper [0,0,0] 20.0 (charName + "_Locators_TOP") centerMarker:true
		)
		
		for c = 1 to locators.count do
		(
			local newLocator = SSR_rig.makeHelper (locators[c][2]*rigSize) 1.0 (locators[c][1]) cross:true
			newLocator.parent = locatorParent
			
			-- If other parts exist, we offset the new locator and rename it
			if existingParts > 0 then
			(
				newLocator.name = substituteString newLocator.name "_1_" ("_" + (existingParts + 1) as string + "_")
				newLocator.pos = newLocator.pos + ( [0,-(rigSize*0.1),0] * (existingParts+1) )
			)
		)
		
		for c = 1 to locatorLines.count do
		(
			local startLoc = execute ("$'" + (substituteString locatorLines[c][3] "_1_" ("_" + (existingParts + 1) as string + "_")) + "'")
			local endLoc = execute ("$'" + (substituteString locatorLines[c][2] "_1_" ("_" + (existingParts + 1) as string + "_")) + "'")
			
			local newLine = connectLine startLoc endLoc locatorLines[c][1]
			newLine.parent = locatorParent
		)

		--	Restore original current layer
		storedCurrent.current = true
		SRT_newLine = undefined
		completeRedraw() 
		
		return true
	),

	-----------------------------------------------------------------------------
	--	createExtraLocators: 
	--	creates locators for extra FK parts, such as ears and tails
	-----------------------------------------------------------------------------
	function createExtraLocators bodyPart rigSize charName symmetric:false debug:false =
	(
		-- If something is selected we will create relative to the selection
		local startPos = [0,0,0]
		local extend = false
		if $selection.count == 1 then 
		(
			startPos = $selection[1].pos + [0,0,rigSize*0.1]
			-- Are we extending an existing extra bodypart?
			if matchPattern $selection[1].name pattern:"Locator_Xtra_*" == true then
				extend = true
		)
		
		--	We create a rig layer, and make it active. 
		local storedCurrent = LayerManager.current
		local locatorLayerName = charName + " Rig Locators"
		local locatorLayer = layermanager.getlayerfromname locatorLayerName
		if locatorLayer == undefined then locatorLayer = layermanager.newLayerfromname locatorLayerName
		locatorLayer.current = true	
		
		-- Create a TOP node if no one exists
		local locatorParent = execute ("$'" + charName + "_Locators_TOP'")
		if locatorParent == undefined then
			locatorParent = SSR_rig.makeHelper [0,0,0] 20.0 (charName + "_Locators_TOP") centerMarker:true
		
		-- Are we doing a symmetric or a straight bodypart
		if symmetric == false then
		(
			if debug == true then SSR_gen.verbose "Creating extra non-symmetric locator."
			-- Create centered extra locator
			if extend == false then
			(
				local startLoc = SSR_rig.makeHelper (startPos+[0,0,rigSize*0.1]) 1.0 "Locator_Xtra_Tail1" cross:true
				local endLoc = SSR_rig.makeHelper (startPos+[0,0,rigSize*0.2]) 1.0 "Locator_Xtra_Tail2" cross:true
				local locatorLine = SSR_loc.connectLine startLoc endLoc "LocatorLine_Xtra_Tail2"
				
				if $selection.count == 1 then
				(
					if classOf $selection[1] == point and (filterString $selection[1].name "_")[1] == "Locator" then
					local rigConnection = SSR_loc.connectLine startLoc $selection[1] "LocatorLine_Xtra_TailtoRig"
				)
				
				startLoc.parent = endLoc.parent = locatorLine.parent = locatorParent
			) else (
				-- Find the last locator in this segment
				local nameRoot = substring $selection[1].name 1 ($selection[1].name.count-1)
				local sameLocs = execute ("$'" + nameRoot + "*'")
				
				local endLoc = $selection[1]
				
				for l in sameLocs do
				(
					curNum = (substring l.name ($selection[1].name.count) ($selection[1].name.count)) as integer
					endNum = (substring endLoc.name (endLoc.name.count) (endLoc.name.count)) as integer

					if curNum > endNum then endLoc = l
				)
				
				local thisNum = (substring endLoc.name (endLoc.name.count) (endLoc.name.count)) as integer
				local prevLoc = execute ("$'" + nameRoot + (thisNum - 1) as string + "'")
				
				local newLoc = SSR_rig.makeHelper (endLoc.pos * 2 - prevLoc.pos) 1.0 ("Locator_Xtra_Tail" + (thisNum + 1) as string) cross:true
				local locatorLine = SSR_loc.connectLine endLoc newLoc ("LocatorLine_Xtra_Tail" + (thisNum + 1) as string)
				
				locatorLine.parent = newLoc.parent = locatorParent
			)
		) else (
			-- Create symmetric extra locators
			if debug == true then SSR_gen.verbose "Creating extra symmetric locators."
			
			-- Right side locators
			local startLoc = SSR_rig.makeHelper (startPos+[-rigSize*0.1,0,rigSize*0.1]) 1.0 "Locator_Xtra_R_Ear1" cross:true
			local endLoc = SSR_rig.makeHelper (startPos+[-rigSize*0.2,0,rigSize*0.2]) 1.0 "Locator_Xtra_R_Ear2" cross:true
			local locatorLine = SSR_loc.connectLine startLoc endLoc "LocatorLine_Xtra_R_Ear2"
			
			startLoc.parent = endLoc.parent = locatorLine.parent = locatorParent
			
			-- Left side locators
			startLoc = SSR_rig.makeHelper (startPos+[rigSize*0.1,0,rigSize*0.1]) 1.0 "Locator_Xtra_L_Ear1" cross:true
			endLoc = SSR_rig.makeHelper (startPos+[rigSize*0.2,0,rigSize*0.2]) 1.0 "Locator_Xtra_L_Ear2" cross:true
			locatorLine = SSR_loc.connectLine startLoc endLoc "LocatorLine_Xtra_L_Ear2"
			
			startLoc.parent = endLoc.parent = locatorLine.parent = locatorParent
			
			--
			--	Needs a function to extend symmetric helpers!
			--
		)
		
		storedCurrent.current = true
		completeRedraw() 
		return true
	),
	
	-----------------------------------------------------------------------------
	-- symmetryHelpers:
	-- Function to mirror all locators around the X axis
	-----------------------------------------------------------------------------
	function symmetryHelpers fromSide toSide debug:false =
	(
		local fromSide
		local toSide
		
		local fromHelpers = #()
		local toHelpers = #()
		
		for o in $helpers do
		(
			if matchPattern o.name pattern:("Locator_*_" + fromSide + "_*") == true then append fromHelpers o
			else if matchPattern o.name pattern:("Locator_*_" + toSide + "_*") == true then append toHelpers o
		)
		
		if debug == true then 
		(
			SSR_gen.verbose ("fromHelpers: " + fromHelpers.count as string)
			SSR_gen.verbose ("toHelpers: " + toHelpers.count as string)
		)
		
		for f in fromHelpers do
		(
			for t in toHelpers do
			(
				local fromStripped = substituteString f.name ("_" + fromSide + "_" ) "_"
				local toStripped = substituteString t.name ("_" + toSide + "_" ) "_"

				if fromStripped == toStripped then
				(
					if debug == true then SSR_gen.verbose (f.name + " matched to " + t.name)
					t.pos = [-f.pos.x, f.pos.y, f.pos.z]
				)
			)
		)
		
		return true
	),
	
	----------------------------------------------------------
	-- straightenHelper: 
	-- locks a helper to the line inbetween two helpers
	-- useful to make sure arms and legs are a straight line
	----------------------------------------------------------
	function straightenHelper targetObj firstObj secondObj =
	(
		local tShape = SplineShape pos:[0, 0, 0]
		addNewSpline tShape
		addKnot tShape 1 #corner #line firstObj.pos
		addKnot tShape 1 #corner #line secondObj.pos
		updateShape tShape

		local param = nearestPathParam tShape targetObj.pos
		local newLoc = pathInterp tShape param
		
		targetObj.pos = [newLoc.x, targetObj.pos.y, newLoc.z]
		delete tShape
	),
	
	----------------------------------------------------------
	-- averagePosition: 
	-- averages locator positions across an axis
	-- mainly to keep feet aligned
	----------------------------------------------------------
	function averagePosition objList axis:"X" =
	(
		local objPosition = [0,0,0]
		for o in objList do objPosition = objPosition + o.pos
		objPosition = objPosition / objList.count
		
		if axis == "X" then for o in objList do o.pos.x = objPosition.x
		if axis == "Y" then for o in objList do o.pos.y = objPosition.y
		if axis == "Z" then for o in objList do o.pos.z = objPosition.z
			
		return true
	),
	
	----------------------------------------------------------
	-- connectLine: 
	--	Use a spline to connect helpers
	----------------------------------------------------------
	function connectLine startObj endObj lineName =
	(
		local tShape = SplineShape pos:[0,0,0] name:lineName
		addNewSpline tShape
		addKnot tShape 1 #corner #line startObj.pos
		addKnot tShape 1 #corner #line endObj.pos
		updateShape tShape
		
		startCon = Point3_XYZ()
		endCon = Point3_XYZ()
		
		animateVertex tShape #all

		tShape[4][8].controller.Spline_1___Vertex_1.controller = startCon
		tShape[4][8].controller.Spline_1___Vertex_2.controller = endCon
		
		startCon[1].controller = startObj.pos.controller[1].controller
		startCon[2].controller = startObj.pos.controller[2].controller
		startCon[3].controller = startObj.pos.controller[3].controller
		
		endCon[1].controller = endObj.pos.controller[1].controller
		endCon[2].controller = endObj.pos.controller[2].controller
		endCon[3].controller = endObj.pos.controller[3].controller
		
		freeze tShape
		return tShape
	), 
	
	----------------------------------------------------------
	-- cleanLines: 
	--	Cleans out orphaned locator lines
	----------------------------------------------------------	
	function cleanLines charName = 
	(
		local cleanedLines = 0
		local charLocators = #()
		
		if (charName != "") and (execute ("$'" + charName + "_Locators_TOP'") == undefined) then
		(
			SSR_gen.verbose ("'$" + charName + "_Locators_TOP' does not exist.")
			return false
		)
		
		if charName != "" then charLocators = execute ("$'" + charName + "_Locators_TOP'.children")
		else charLocators = (execute "$'LocatorLine_*'") as array + (execute "$'Locator_*'") as array
		
		for o in charLocators do
		(
			if matchPattern o.name pattern:"LocatorLine_*" then
			(
				-- Check if controllers are no longer instanced
				helperCount = 0
				startCon = o[4][8].controller.Spline_1___Vertex_1.controller
				endCon = o[4][8].controller.Spline_1___Vertex_2.controller
				startPos = [startCon.x, startCon.y, startCon.z]
				endPos = [endCon.x, endCon.y, endCon.z]
				
				-- Cycle through all helpers, does anyone position at the ends?
				for h in charLocators do
				(
					--! this results in "Runtime error: Attempt to access deleted scene object"
					try (
						if matchPattern h.name pattern:"Locator_*" then
						(
							if startPos == h.pos then helperCount = helperCount + 1
							if endPos == h.pos then helperCount = helperCount + 1
						)
					) catch ()
				)
				
				if helperCount < 2 then 
				(
					delete o
					cleanedLines = cleanedLines + 1
				)
			)
		)
		
		SSR_gen.verbose ("Cleaned out " + cleanedLines as string + " line objects")
		return cleanedLines
	),
	
	-----------------------------------------------------------------------------
	--	killLocators: 
	--	clears all locators from the scene
	-----------------------------------------------------------------------------
	function killLocators charName =
	(
		if (charName != "") and (execute ("$'" + charName + "_Locators_TOP'") == undefined) then
		(
			SSR_gen.verbose ("'$" + charName + "_Locators_TOP' does not exist.")
			return false
		)
		
		if charName != "" then
		(
			local charLocators = execute ("$'" + charName + "_Locators_TOP'.children")
			for o in charLocators do delete o
			
			try	execute ("delete $'" + charName + "_Locators_TOP'") catch ()
			try	execute ("LayerManager.deleteLayerByName " + "\"" + charName + " Rig Locators\"") catch ()
		) else (
			-- Delete every single "Locator_*", "Locatorline_*", "*_Locators_TOP" and "* Rig Locators" layer
			try	execute ("delete $'Locator_*'") catch ()
			try	execute ("delete $'LocatorLine_*'") catch ()
			try	execute ("delete $'*_Locators_TOP'") catch ()
			SSR_gen.deleteLayer "* Rig Locators"
		)
		
		
		return true
	),
	
	-----------------------------------------------------------------------------
	--	findLocators: 
	--	reports how many instances there are of one specific bodypart
	-----------------------------------------------------------------------------
	function findLocators charName partName side:"R" = 
	(
		-- Collect all the children of the characters TOP node
		local charLocators = execute ("$'" + charName + "_Locators_TOP'.children")
		
		-- Search for nodes that match the limb root (clavicle or hip)
		local numLimbs = 0
		for o in charLocators do 
		(
			if (matchPattern o.name pattern:("Locator_" + partName + "_" + side + "_?_Hip")) == true or \
				(matchPattern o.name pattern:("Locator_" + partName + "_" + side + "_?_Clavicle")) == true then
			(
				numLimbs = numLimbs + 1
			)	
		)
		--SSR_gen.verbose ("Found " + numLimbs as string + " " + side + " side " + partName + "(s).")
		
		return numLimbs
	)

)

SSR_loc = SSR_loc()