-----------------------------------------------------------------------------
--
--	Functions needed for rigging a character
--
--	v1.0:	First version, merge of different files
--
-----------------------------------------------------------------------------


global SSR_rig

struct SSR_rig
(
	----------------------------------------------------------
	-- rigBodyPart: 
	--	Rigs the body part
	----------------------------------------------------------
	function rigBodyPart partsPath bodyPart charName extParams:undefined boneSize:1 verbose:false =
	(
		-- Load the specified rigging file
		partFiles = getFiles (partsPath + bodyPart + "*.*")
		fileIn partFiles[1]
		
		-- Create the rig!
		if extParams == undefined then
			local controllers = SSR_part.doPart charName boneSize:boneSize
		else
			local controllers = SSR_part.doPart charName extParams:extParams boneSize:boneSize
		
		return controllers
	),
	
	----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
	-- bodyPartVariants: 
	--	Builds an array of available body part rigs
	----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
	function listBodyParts partsPath bodyPart verbose:false =
	(
		bodyPartList = #()
		partsFiles = getFiles (partsPath + "\\" + bodypart + "_*.ms")
		
		for f in partsFiles do
		(
			append bodyPartList ( (filterString (getFilenameFile f) "_.")[2] + " " + (filterString (getFilenameFile f) "_.")[1] )
		)
		
		return bodyPartList as array
	),
	
	----------------------------------------------------------
	-- fillDropdown: 
	--	Fills a dropdown list with bodypart alternative rigs
	----------------------------------------------------------	
	function fillDropdown partsPath bodyPart dropTarget =
	(
		local rigTypes = SSR_rig.listBodyParts partsPath bodyPart
		dropTarget.items = for r in rigTypes collect r
	),
	
	----------------------------------------------------------
	-- makeHelper: 
	--	Functions to streamline the creation of helpers
	----------------------------------------------------------
	function makeHelper pos size objName centerMarker:false cross:false box:false axistripod:false drawontop:false =
	(
		local tCross = point pos:pos size:size centermarker:centermarker cross:cross box:box axistripod:axistripod drawontop:drawontop
		tCross.name = objName
		return tCross
	),
	
	----------------------------------------------------------
	-- makeBone: 
	--	Creates a bone
	----------------------------------------------------------
	function makeBone boneName startPos endPos zAxis boneSize frontFin:false =
	(
		ourBone = BoneSys.createBone startPos endPos zAxis
		ourBone.name = boneName
		ourBone.width = ourBone.height = boneSize
		ourBone.sidefins = false
		ourBone.backfin = false
		ourBone.frontfinsize = boneSize
		ourBone.taper = 90
		ourBone.frontfin = frontFin
		
		return ourBone
	),

	----------------------------------------------------------
	-- endBone: 
	--	creates an endBone
	----------------------------------------------------------
	function endBone capBone =
	(
		startPos = ((transMatrix [capBone.length,0,0])*capBone.transform).translationPart
		
		ourBone = BoneSys.createBone startPos (startPos + [0, 0, capBone.width]) [1,0,0]
		ourBone.name = capBone.name + "_End"
		ourBone.width = ourBone.height = capBone.width
		
		SSR_rig.objAlign ourBone capBone
		ourBone.parent = capBone
		ourBone.frontfin = false
		
		return ourBone
	),

	----------------------------------------------------------
	-- boneHelper: 
	--	adds and parents a point object to a bone,
	--	use the position factor to set the distance into the bone
	----------------------------------------------------------
	function boneHelper parentObj refBone position coords objName =
	(
		local obj
		local position
		
		local myHelper = point name:objName size:5 centermarker:false cross:false box:true axistripod:false
		myHelper.transform = parentObj.transform
		move in coordsys parentObj move myHelper (coords * (refBone.length * position * 0.01))
		myHelper.parent = parentObj
		
		return myHelper
	),

	----------------------------------------------------------
	-- skinBones: 
	--	creates a box that fits around a bone. nice for skinning
	----------------------------------------------------------
	function skinBones boneList objName side =
	(
		local skinBones = #()
		
		for o in boneList do
		(
			boneBox = box transform:o.transform name:(objName + side + "01") width:o.width length:o.height height:o.length
			in coordsys local rotate boneBox (angleaxis 90 [0,1,0])
			boneBox.parent = o
			
			append skinBones boneBox
		)
		
		return skinBones
	),

	----------------------------------------------------------
	-- freezeTransform: 
	--	freezes transformation, nicked from the 3ds max maxscript folder
	----------------------------------------------------------
	function freezeTransform CurObj =
	(
		Try
		(	
			local CurObj 	

			if classof CurObj.rotation.controller != Rotation_Layer do
			(
				-- freeze rotation		
				CurObj.rotation.controller = Euler_Xyz() 		
				CurObj.rotation.controller = Rotation_list() 			
				CurObj.rotation.controller.available.controller = Euler_xyz() 		
				CurObj.rotation.controller.setname 1 "Frozen Rotation" 		
				CurObj.rotation.controller.setname 2 "Zero Euler XYZ" 		
				CurObj.rotation.controller.SetActive 2 		
			)
			if classof CurObj.position.controller != Position_Layer do
			(
				-- freeze position
				CurObj.position.controller = Bezier_Position() 			
				CurObj.position.controller = position_list() 			
				CurObj.position.controller.available.controller = Position_XYZ() 	
				CurObj.position.controller.setname 1 "Frozen Position" 	
				CurObj.position.controller.setname 2 "Zero Pos XYZ" 			
				CurObj.position.controller.SetActive 2 		

				-- position to zero
				CurObj.Position.controller[2].x_Position = 0
				CurObj.Position.controller[2].y_Position = 0
				CurObj.Position.controller[2].z_Position = 0
			)
		)	

		Catch(SSR_gen.verbose "A failure occurred while freezing an object's transform." title:"Freeze Transform")
	),

	----------------------------------------------------------
	-- nullTransform: 
	--	nulls out transform of an object using a point object
	--	parents the object to the null, 
	--	and parents null to original parent
	----------------------------------------------------------
	function nullTransform obj =
	(
		local oldParent = obj.parent
		local ntHelper = point name:(obj.name + "_null") centerMarker:true cross:false box:false axistripod:false size:5
		ntHelper.transform = obj.transform
		--ntHelper.rotation = obj.rotation
		--ntHelper.position = obj.position
		ntHelper.scale = [1,1,1]
		obj.parent = ntHelper
		ntHelper.parent = oldParent
		
		return ntHelper
	),

	----------------------------------------------------------
	-- resetRotation: 
	--	some objects end up with 360 rotations in random axis.
	--	This function is a cheap fix for that
	----------------------------------------------------------
	function resetRotation obj =
	(
		if classOf obj.rotation.controller == Euler_XYZ then
		(
			obj.rotation.controller.X_rotation = 0.0
			obj.rotation.controller.Y_rotation = 0.0
			obj.rotation.controller.Z_rotation = 0.0
		)
	),

	----------------------------------------------------------
	-- posFactor: 
	--	aligns and moves object in another objects coord system - using a length factor
	--	Great for positioning objects along the length of, say, a bone
	----------------------------------------------------------
	function posFactor targetObj coordObj basePosition targetTrans transLength:1.0 transFactor:1.0 =
	(
		targetObj.transform = coordObj.transform
		targetObj.pos = basePosition
		in coordsys coordObj move targetObj (targetTrans * (transLength * transFactor))
		
		return true
	),
	
	----------------------------------------------------------
	-- drawVector: 
	--	Draws a line in a direction from a point, mostly for debugging
	----------------------------------------------------------
	function drawVector vDir vStart =
	(
		local tShape = SplineShape pos:[0, 0, 0]
		addNewSpline tShape
		addKnot tShape 1 #corner #line vStart
		addKnot tShape 1 #corner #line (vStart+vDir)
		updateShape tShape
		
		tShape.name = "tempVec"
		
		return tShape
	),

	-----------------------------------------------------------------------------
	--
	-- Functions related to control objects
	--
	-----------------------------------------------------------------------------

	----------------------------------------------------------
	-- colorControllers: 
	--	sets colors according to which side the controllers are on
	----------------------------------------------------------
	function colorControllers controllers rightColor centerColor leftColor =
	(
		for o in controllers do
		(
			in coordsys world
			(
				if o.pos.x > 0.001 then o.wirecolor = leftColor
				else if o.pos.x < -0.0001 then o.wirecolor = rightColor
				else o.wirecolor = centerColor
			)
		)
	),

	----------------------------------------------------------
	-- curveHelper: 
	--	locks a helper to the closest point on a curve
	----------------------------------------------------------
	function curveHelper hSize hSpline hPos =
	(
		local newHelp = point size:hSize
		local con = Path_Constraint path:hSpline
		newHelp.pos.controller = con
		deleteKeys con.percent.keys
		local param = nearestPathParam hSpline hPos
		con.percent = (pathToLengthParam hSpline param) * 100
		
		return newHelp
	),

	----------------------------------------------------------
	-- crossCTL: 
	--	creates a cross control object
	----------------------------------------------------------
	function crossCTL pos size thick:false =
	(
		local tShape = SplineShape pos:pos
		addNewSpline tShape
		addKnot tShape 1 #corner #line (pos + [size*.5, 0, 0])
		addKnot tShape 1 #corner #line (pos + [-size*.5, 0, 0])
		addNewSpline tShape
		addKnot tShape 2 #corner #line (pos + [0, size*.5, 0])
		addKnot tShape 2 #corner #line (pos + [0, -size*.5, 0])
		addNewSpline tShape
		addKnot tShape 3 #corner #line (pos + [0, 0, size*.5])
		addKnot tShape 3 #corner #line (pos + [0, 0, -size*.5])
		updateShape tShape

		if thick == "true" then
		(
			tShape.render_displayRenderMesh = true
			tShape.render_thickness = 1
		) 
		else if thick == "false" then
		(
			tShape.render_displayRenderMesh = false
		)
		
		return tShape
	),

	----------------------------------------------------------
	-- ringCTL: 
	--	creates a ring control object
	----------------------------------------------------------
	function ringCTL pos size alignAxis:"Z" thick:false steps:2 =
	(
		local tShape = circle pos:pos radius:(size*0.5) steps:steps
		
		if alignAxis == "X" then
		(
			local xMod = xform()
			addModifier tShape xMod
			xMod.gizmo.rotation = eulerAngles 0 90 0
		) 
		else if alignAxis == "Y" then
		(
			local xMod = xform()
			addModifier tShape xMod
			xMod.gizmo.rotation = eulerAngles 90 0 0
		)
		
		if thick == "true" then
		(
			tShape.render_displayRenderMesh = true
			tShape.render_thickness = 1
		) 
		else if thick == "false" then
		(
			tShape.render_displayRenderMesh = false
		)
		
		convertTo tShape SplineShape
		return tShape
	),

	----------------------------------------------------------
	-- ballCTL: 
	--	creates a ball control object
	----------------------------------------------------------
	function ballCTL pos size thick:false steps:2 =
	(
		-- First ring
		local tShape = circle pos:pos radius:(size*0.5) steps:steps
		convertTo tShape SplineShape
		
		-- Second and third ring
		local tRing1 = circle pos:pos radius:(size*0.5)
		in coordSys local rotate tRing1 (eulerAngles 0 90 0)
		convertTo tRing1 SplineShape
		
		local tRing2 = circle pos:pos radius:(size*0.5)
		in coordSys local rotate tRing2 (eulerAngles 90 0 0)
		convertTo tRing2 SplineShape
		
		-- Attach second and third ring to first
		addAndWeld tShape tRing1 -1
		addAndWeld tShape tRing2 -1

		local xMod = xform()
		addModifier tShape xMod
		xMod.gizmo.rotation = eulerAngles 0 90 0

		
		if thick == "true" then
		(
			tShape.render_displayRenderMesh = true
			tShape.render_thickness = 1
		) 
		else if thick == "false" then
		(
			tShape.render_displayRenderMesh = false
		)
		
		convertTo tShape SplineShape
		return tShape
	),

	----------------------------------------------------------
	-- pinCTL: 
	--	creates a pin control object
	----------------------------------------------------------
	function pinCTL startPos pinLength size alignAxis:"Y" thick:false steps:2 =
	(
		local tShape = SplineShape pos:startPos steps:steps
		local endPos = startPos + [0, pinLength, 0]
		addNewSpline tShape
		addKnot tShape 1 #corner #line startPos
		addKnot tShape 1 #corner #line endPos
		updateShape tShape
		
		local tPin = circle pos:endPos radius:(size*0.5)
		local xMod = xform()
		addModifier tPin xMod
		xMod.gizmo.rotation = eulerAngles 90 0 0
		convertTo tPin SplineShape
		
		addAndWeld tShape tPin -1
		
		if alignAxis == "X" then
		(
			local xMod = xForm()
			addModifier tShape xMod
			xMod.gizmo.rotation = eulerAngles 0 0 90
			convertTo tShape SplineShape
		)
		if alignAxis == "Z" then
		(
			local xMod = xForm()
			addModifier tShape xMod
			xMod.gizmo.rotation = eulerAngles 90 0 0
			convertTo tShape SplineShape
		)
		if alignAxis == "-Z" then
		(
			local xMod = xForm()
			addModifier tShape xMod
			xMod.gizmo.rotation = eulerAngles -90 0 0
			convertTo tShape SplineShape
		)
		
		if thick == "true" then
		(
			tShape.render_displayRenderMesh = true
			tShape.render_thickness = 1
		) 
		else if thick == "false" then
		(
			tShape.render_displayRenderMesh = false
		)
		
		return tShape
	),

	----------------------------------------------------------
	-- resizeCon: 
	--	resizes and moves the controller by inserting an xform modifier
	----------------------------------------------------------
	function resizeCon controller newSize newPos:[0,0,0] =
	(
		--! TODO - is there an existing rescale controller?
		
		local sizeCon = xform()
		addModifier controller sizeCon before:1
		sizeCon.gizmo.scale = newSize
		sizeCon.gizmo.position = newPos
		sizeCon.name = "XForm_CTL_Rescale"
		
		return controller
	),

	----------------------------------------------------------
	-- swapCon: 
	--	swaps one controller type with another
	----------------------------------------------------------
	function swapCon controller size type:"Pin" =
	(
		local startPos = controller.pos
		local endPos = startPos + [0, size*5, 0]
		
		local tShape = undefined
		
		-- Create a pin shape
		if type == "Pin" then
		(
			tShape = SplineShape pos:startPos
			addNewSpline tShape
			addKnot tShape 1 #corner #line startPos
			addKnot tShape 1 #corner #line endPos
			updateShape tShape
			
			local tPin = circle pos:endPos radius:(size*0.5)
			local xMod = xform()
			addModifier tPin xMod
			xMod.gizmo.rotation = eulerAngles 90 0 0
			convertTo tPin SplineShape
			
			addAndWeld tShape tPin -1
		)
		
		if type == "Circle" then
		(
			tShape = circle pos:startPos radius:size
			convertTo tShape SplineShape
		)
		
		-- Delete the previous shape, then merge in the new one.
		objAlign tShape controller
		
		for c = 1 to controller.numSplines do deleteSpline controller 1
		
		addAndWeld controller tShape -1
		updateShape controller
		
		return controller
	),


	-----------------------------------------------------------------------------
	--
	-- Functions related to alignment
	--
	-----------------------------------------------------------------------------

	-----------------------------------------------------------------------------
	-- closestAxis: 
	--	returns the closest world axis of a vector
	-----------------------------------------------------------------------------
	function closestAxis dirVec =
	(
		-- Compare angles, return the closest angle
		dirVec = normalize dirVec
		local vecBits = #(abs(dirVec[1]), abs(dirVec[2]), abs(dirVec[3]))
		local largestId = 0
		local largestVal = 0.0
		
		for c = 1 to 3 do
		(
			if vecBits[c] > largestVal then 
			(
				largestVal = vecBits[c]
				largestId = c
			)
		)
		return largestId
	),
	
	
	-----------------------------------------------------------------------------
	-- targetAlign: 
	--	aligns orientation of first object to look at  second object
	--	(using a controller method, should idealy just do it with math)
	-----------------------------------------------------------------------------
	function targetAlign sourceObj targetObj =
	(
		local tempHelper = SSR_rig.makeHelper sourceObj.pos 10 "alignTemp" axisTripod:true
		SSR_rig.lookCon tempHelper targetObj
		SSR_rig.objAlign sourceObj tempHelper 
		delete tempHelper
		
		return true
	),
	
	-----------------------------------------------------------------------------
	-- objAlign: 
	--	aligns orientation of first object to second object
	--	(but keeps its original position)
	-----------------------------------------------------------------------------
	function objAlign target source closestAxis:false =
	(
		if target == undefined or source == undefined then return false
		local tempPos = target.pos
		target.transform = source.transform
		
		if closestAxis == true then
		(
			-- Check the direction of the object x axis (transform row1)
			local closestAxis = SSR_rig.closestAxis source.transform.row1
			
			if closestAxis == 1 then target.rotation = (quat 0 0 0 1)
			if closestAxis == 2 then target.rotation = (quat 0 0 0.707107 0.707107)
			if closestAxis == 3 then target.rotation = (quat 0 -0.707107 0 0.707107)
		)
		target.pos = tempPos
		
		return true
	),

	-----------------------------------------------------------------------------
	-- pivotAlign: 
	--	aligns the pivot of first object to the position of second object
	-----------------------------------------------------------------------------
	function pivotAlign doAlign sourceAlign =
	(
		doAlign.pivot = sourceAlign
	),


	-----------------------------------------------------------------------------
	--
	-- Functions related to constraints
	--
	-----------------------------------------------------------------------------

	-----------------------------------------------------------------------------
	-- parentCon: 
	--	parent constraints source to target
	--	using both orientation and position constraints, keeping initial offset
	-----------------------------------------------------------------------------
	function parentCon source target =
	(
		local posCon = Position_Constraint()
		posCon.appendTarget target 100
		source.pos.controller = posCon
		posCon.relative = on

			
		local rotCon = Orientation_Constraint()
		rotCon.appendTarget target 100
		source.rotation.controller = rotCon
		rotCon.relative = on
		
		return #(posCon, rotCon)
	),

	function parentConTwo targetOne targetTwo =
	(
		local posCon = Position_Constraint()
		posCon.appendTarget targetOne 50.0
		posCon.appendTarget targetTwo 50.0
		source.pos.controller = posCon
		return posCon

			
		local rotCon = Orientation_Constraint()
		rotCon.appendTarget targetOne 50.0
		rotCon.appendTarget targetTwo 50.0
		source.rotation.controller = rotCon
		return rotCon
		
		return #(posCon, rotCon)
	),

	
	-----------------------------------------------------------------------------
	-- posCon: 
	--	position constraints source to target
	-----------------------------------------------------------------------------
	function posCon source target =
	(
		local con = Position_Constraint()
		con.appendTarget target 100
		source.pos.controller = con
		return con
	),

	function posConTwo source targetOne targetTwo =
	(
		local con = Position_Constraint()
		con.appendTarget targetOne 50.0
		con.appendTarget targetTwo 50.0
		source.pos.controller = con
		return con
	),

	-----------------------------------------------------------------------------
	-- rotCon: 
	--	orientation constraints source to target
	-----------------------------------------------------------------------------
	function rotCon source target =
	(
		local con = Orientation_Constraint()
		con.appendTarget target 100
		source.rotation.controller = con
		return con
	),

	function rotConTwo source targetOne targetTwo =
	(
		local con = Orientation_Constraint()
		con.appendTarget targetOne 50.0
		con.appendTarget targetTwo 50.0
		source.rotation.controller = con
		return con
	),

	-----------------------------------------------------------------------------
	-- linkCon: 
	--	link constraints source to target
	-----------------------------------------------------------------------------
	function linkCon source target =
	(
		local con = Link_Constraint()
		source.Transform.controller = con
		con.addTarget target 1
		return con
	),

	-----------------------------------------------------------------------------
	-- setUpNode: 
	--	sets up node on orientation constraint
	-----------------------------------------------------------------------------
	function setUpNode sourceCon target =
	(
		sourceCon.pickUpNode = target
		sourceCon.upnode_world = false
		sourceCon.upnode_ctrl = 0
	),

	-----------------------------------------------------------------------------
	-- lookCon: 
	--	lookat constraints source to target
	-----------------------------------------------------------------------------
	function lookCon source target =
	(
		local con = LookAt_Constraint()
		con.appendTarget target 100
		con.lookat_vector_length = 0.0
		con.viewline_length_abs = false
		source.rotation.controller = con
		
		return con
	),

	-----------------------------------------------------------------------------
	-- listCon: 
	--	creates a list controller on input channel
	-----------------------------------------------------------------------------
	function listCon inCon inObj =
	(
		print (classOf inCon)
		
		if classOf inCon == Euler_XYZ then
		(
			local rotWeight = 360.0/pi*0.005
			newCon = rotation_list()
			inObj.rotation.controller = newCon
			newCon.weight[1] = rotWeight
			
			return newCon
		)
		else if classOf inCon == Position_XYZ then
		(
			con = position_list()
			
			return newCon
		) else return false
	),
	
	-----------------------------------------------------------------------------
	--
	-- getChildBone: 
	--	checks for and returns a single child bone
	--	used in stretch twist function
	--
	-----------------------------------------------------------------------------
	function getChildBone masterBone =
	(
		if masterBone != undefined then
		(
			local childBones = #()
			
			for c in masterBone.children do
			(
				if classOf c == BoneGeometry then append childBones c
			)
			
			if childBones.count == 0 then
			(
				SSR_gen.verbose "Masterbone has no child bone."
				return false
			) else if childBones.count > 1 then
			(
				SSR_gen.verbose "Masterbone has too many child bones."
				return false
			) else (
				return childBones[1]
			)
		) else return false
	),

	-----------------------------------------------------------------------------
	--
	-- twistBones: 
	--	creates twist bones along a single bone, aligned to a target bone
	--
	-----------------------------------------------------------------------------
	function twistBones masterBone target numTwist twistUp:false debugState:false =
	(	
		-- An array containing the skin bones, will be returned at the end of the function
		skinBoneCollector = #()
		
		--	Length of each stretch bone
		local twistSize = masterBone.length / numTwist
		local twistLookat = undefined
		
		--	Up node that will drive the lookat on the main twist bone
		local targetUp = makeHelper target.pos (masterBone.width) (target.name + "_twistUp") cross:true
		targetUp.name = substitutestring (targetUp.name) "_skinBone_" "_"
		objAlign targetUp target
		targetUp.parent = target
		
		--	A helper that all twist bones will be parented to, otherwise stretching breaks
		local twistRoot = makeHelper masterBone.pos (masterBone.width) (masterBone.name + "_twistRoot") box:true
		twistRoot.name = substitutestring (twistRoot.name) "_skinBone_" "_"
		objAlign twistRoot masterBone
		twistRoot.parent = masterBone
		parentCon twistRoot masterBone
		
		-- Create start twist bone. Will be unconstrained
		local twistBoneStatic = makeBone (masterBone.name + "_twistStatic") masterBone.pos (masterBone.pos + [twistSize,0,0]) [0,0,1] (masterBone.width)
		objAlign twistBoneStatic masterBone
		if twistUp == false then
		(
			-- This bone is the first in the chain
		) else (
			-- This bone is last in the chain
			in coordsys masterBone move twistBoneStatic [twistSize*(numTwist-1), 0, 0] 
		)
		twistBoneStatic.parent = twistRoot
		
		-- Add to our skinBone array
		append skinBoneCollector twistBoneStatic
		
		-- Create last twist bone. Lookat constrained to the target bone
		local twistBoneLookat = makeBone (masterBone.name + "_twistLook") masterBone.pos (masterBone.pos + [twistSize,0,0]) [0,0,1] (masterBone.width)
		twistBoneLookat.frontfin = debugState
		objAlign twistBoneLookat masterBone
		if  twistUp == false then
		(
			-- This bone is last in the chain
			in coordsys masterBone move twistBoneLookat [twistSize*(numTwist-1), 0, 0]
			twistLookat = masterBone
		)  else (
			-- This bone is the first in the chain
			twistLookat = makeHelper masterBone.pos (masterBone.width) (masterBone.name + "_twistLookAt") cross:true
			twistLookat.name = substitutestring (twistLookat.name) "_skinBone_" "_"
			in coordsys masterBone move twistLookat [masterBone.length, 0, 0] 
			twistLookat.parent = twistRoot
		)
		twistBoneLookat.parent = twistRoot
		
		-- Add to our skinBone array
		append skinBoneCollector twistBoneLookat
		
		-- Constrain the twist bone
		local lookCon = SSR_rig.lookCon twistBoneLookat twistLookat
		lookCon.upnode_world = off
		lookCon.pickUpNode = targetUp
		if twistUp == false then lookCon.target_axisFlip = on else lookCon.target_axisFlip = off
		
		-- Fix axis to limit flipping
		lookCon.StoUp_axis = 1
		lookCon.upnode_axis = 1
		
		-- Create remaining twist bones
		for n = 2 to numTwist-1 do
		(
			local twistBone = makeBone (masterBone.name + "_twist" + n as string) masterBone.pos (masterBone.pos + [twistSize,0,0]) [0,0,1] (masterBone.width)
			twistBone.frontfin = debugState
			objAlign twistBone masterBone
			
			in coordsys masterBone move twistBone [twistSize*(n-1), 0, 0]
			twistBone.parent = twistRoot
			
			-- Constrain the twist bone
			local twistFactor = 100/(numTwist-1)*(n-1)
			local lookCon =  rotConTwo twistBone masterBone twistBoneLookat
			if twistUp == false then lookCon.setWeight 1 (100-twistFactor) else lookCon.setWeight 1 twistFactor
			if twistUp == false then lookCon.setWeight 2 twistFactor else lookCon.setWeight 2 (100-twistFactor)
			
			-- Add to our skinBone array
			append skinBoneCollector twistBone
		)
		
		-- Return the lookat controller, it sometimes needs flipping
		Return #(lookCon, skinBoneCollector)
	), -- function

	-----------------------------------------------------------------------------
	--
	-- stretchTwistBones: 
	--	creates stretchy twist bones along a single bone, aligned to a target bone
	--  	sub-controllers are created for control of bone curve
	--
	--	todo: optional sub-controllers
	--
	-----------------------------------------------------------------------------
	function stretchTwistBones masterBone targetAlign numTwist twistUp:false debug:false =
	(	
		-- An array containing the skin bones, this is what's returned at the completion of the function
		skinBoneList = #()
		controllerList = #()
		childBone = getChildBone masterBone
		
		-- Does the masterBone have a child bone? If not, exit the function.
		if childBone == undefined then
		(
			SSR_gen.verbose ("'" + masterBone.name + "' moes not match twist setup requirements. It must have a single bone child.")
			return false
		)
		
		-- New targetConstraint helper
		local targetConstraint = makeHelper targetAlign.pos (masterBone.width) (masterBone.name + "_targetConstraint") box:true
		targetConstraint.name = substitutestring (targetConstraint.name) "_skinBone_" "_"
		objAlign targetConstraint masterBone
		targetConstraint.parent = targetAlign		
		
		-- Length of each stretch bone
		local twistSize = masterBone.length / numTwist
		local twistLookat = undefined
		
		local prevHelper = undefined
		local prevBone = undefined
		local prevControl = undefined
		
		for n = 1 to numTwist do
		(			
			if n == 1 then
			(
				prevHelper = makeHelper masterBone.pos (masterBone.width) (masterBone.name + "_twistRoot") cross:true
				prevHelper.name = substitutestring (prevHelper.name) "_skinBone_" "_"
				objAlign prevHelper masterBone
				prevHelper.parent = masterBone		
			)
				
			-- Helpers that the bones will follow. Weighted to lie along the master bone 
			local twistHelper = makeHelper masterBone.pos (masterBone.width) (masterBone.name + "_twistHelper" + n as string) cross:true
			twistHelper.name = substitutestring (twistHelper.name) "_skinBone_" "_"
			objAlign twistHelper masterBone
			twistHelper.parent = masterBone
			
			-- Sub controllers for the middle bones
			local subControl = undefined
			if n < numTwist do
			(
				subControl = crossCTL twistHelper.pos (masterBone.width*4)
				subControl.name = masterBone.name + "_SUB_" + n as string + "_CTL"
				subControl.name = substitutestring (subControl.name) "_skinBone_" "_"
				subControl.parent = twistHelper
				objAlign subControl subControl.parent
				
				-- Add attributes to the controller
				SSR_rig.createPosAttributes subControl
				
				append controllerList subControl
			)
			
			-- Axis constraint helpers for the middle bones
			local lookatHelper = undefined
			if n > 1 and n < numTwist do
			(
				lookatHelper = makeHelper twistHelper.pos (masterBone.width) ("") axistripod:true
				lookatHelper.name = ((substitutestring (masterBone.name) "_skinBone" "_") + "lookatHelper" + n as string)
				lookatHelper.parent = twistHelper
				objAlign lookatHelper lookatHelper.parent
				
				local lookatCon = rotConTwo lookatHelper masterBone targetConstraint
				
				-- Weighting depends on whether we are constraining to a parent or child bone
				local weightFactor = ( 100.0/(numTwist-1) ) * (n-1)
				if debug == true then SSR_gen.verbose ("Twist weight on '" + masterBone.name + "', step " + n as string + ", is: " + weightFactor as string + ".")
				if twistUp == false then
				(
					lookatCon.weight[1] = 100.0 - weightFactor
					lookatCon.weight[2] = weightFactor
				) else (
					lookatCon.weight[2] = 100.0 - weightFactor
					lookatCon.weight[1] = weightFactor	
				)
			)
			
			-- Weight the helpers to be positioned along the masterBone
			local twistPosCon = posConTwo twistHelper masterBone childBone
			local weight = 100.0/numTwist*n
			twistPosCon.weight[1] = 100.0 - weight
			twistPosCon.weight[2] = weight
			
			-- We wedge stretchable bones inbetween these helpers
			local twistBone = makeBone (masterBone.name + "_twist" + n as string) masterBone.pos (masterBone.pos + [twistSize,0,0]) [0,0,1] (masterBone.width)
			if n == 1 then
			(
				--! Temporary fix for backwards compatability
				if twistUp == false then twistBone.name = (masterBone.name + "_twistStatic")
				else twistBone.name = (masterBone.name + "_twistLook")
			)
			if n == numTwist then
			(
				--! Temporary fix for backwards compatability
				if twistUp == false then twistBone.name = (masterBone.name + "_twistLook")
				else twistBone.name = (masterBone.name + "_twistStatic")
			)
			
			posFactor twistBone masterBone masterBone.pos [masterBone.length / numTwist*(n-1), 0.0, 0.0]		
			twistBone.frontfin = debug
			twistBone.boneAutoAlign = true
			twistBone.boneFreezeLength = false
			twistBone.boneScaleType = #scale
			objAlign twistBone masterBone
			
			append skinBoneList twistBone
			
			-- Position and orientation constraints for the twist bones
			if n > 1 and n < numTwist then (
				
				-- Constraints for middle twist bones
				twistBone.parent = prevBone
				local twistBoneCon = posCon twistBone prevControl
				
				local twistLookCon = lookCon twistBone subControl 
				twistLookCon.upnode_world = off
				twistLookCon.pickUpNode = lookatHelper
				
				
			) else if n == numTwist then (
				
				-- Constraints for end twist bone
				-- Create a nub bone to end the chain
				local endBone = makeBone (masterBone.name + "_twistEnd") masterBone.pos (masterBone.pos + [masterBone.width,0,0]) [0,0,1] (masterBone.width)
				endBone.name = substitutestring (endBone.name) "_skinBone_" "_"
				objAlign endBone masterBone
				
				local twistBoneCon = posCon twistBone prevControl
				local endBoneCon = posCon endBone twistHelper
				twistBone.parent = prevBone
				endBone.parent = twistBone
				
				local twistLookCon = lookCon twistBone twistHelper 
				twistLookCon.upnode_world = off
				if twistUp == false then twistLookCon.pickUpNode = targetConstraint
				else twistLookCon.pickUpNode = masterBone
				
			) else if n == 1 then (
			
				-- Constraints for first twist bone
				twistBone.parent = masterBone
				local twistBoneCon = posCon twistBone prevHelper
				
				local twistLookCon = lookCon twistBone subControl 
				twistLookCon.upnode_world = off
				if twistUp == false then twistLookCon.pickUpNode = masterBone
				else twistLookCon.pickUpNode = targetConstraint
				
			)
			
			-- Assign the helper and bone to variables for next loop of twist bone creation
			prevHelper = twistHelper
			prevBone = twistBone
			prevControl = subControl
		)
		
		Return #(controllerList, skinBoneList)
	),
	
	-----------------------------------------------------------------------------
	--
	-- IKSetup: 
	--	creates a soft IK chain with stretch
	--
	-- todo: optional soft IK and knee override position. maybe best as a separate function?
	--
	-----------------------------------------------------------------------------
	-- SSR_rig.IKSetup "Jack" "uparm" "elbow" "lowarm" $Locator_Arm_L_Shoulder.pos $Locator_Arm_L_Elbow.pos $Locator_Arm_L_Hand.pos debug:true
	function IKSetup prefix startName midName endName posStart posMid posEnd boneSize:1 zAxis:undefined boneColor:[255,128,128] doSoft:true debug:false =
	(
		local startLength = distance posStart posMid
		local endLength = distance posMid posEnd
		local IKlength = startLength + endLength
		
		if debug == true then SSR_gen.verbose ("start: " + startLength as string)
		if debug == true then SSR_gen.verbose ("end: " + endLength as string)
		if debug == true then SSR_gen.verbose ("total: " + IKLength as string)
		
		local IK_root = SSR_rig.makeHelper posStart (boneSize*4) (prefix + startName + "_IK_root") box:true
		local IK_target = SSR_rig.makeHelper posEnd (boneSize*4) (prefix + endName + "_IK_target") box:true
		
		-- Temporary bones. Used for alignment, and will be deleted immediately afterwards
		if zAxis == undefined then zAxis = normalize ( cross (posStart-posMid) (posEnd-posMid) )
		local tboneStart = SSR_rig.makeBone (prefix + "_T") posStart posMid zAxis boneSize
		local tboneEnd = SSR_rig.makeBone (prefix + "_T") posMid posEnd zAxis boneSize
		
		--	For this setup, we'll use helpers instead of bones in the IK system
		local startHelper = SSR_rig.makeHelper posStart boneSize (prefix + startName + "_IK_helper") cross:true
		local midHelper = SSR_rig.makeHelper posMid boneSize (prefix + midName + "_IK_helper") cross:true
		local endHelper = SSR_rig.makeHelper posEnd boneSize (prefix + endName + "_IK_helper") cross:true
		
		-- Align helpers (could this be done without temp bones? By MATH?
		SSR_rig.objAlign startHelper tboneStart
		SSR_rig.objAlign midHelper tboneEnd
		SSR_rig.objAlign endHelper tboneEnd
		
		startHelper.parent = IK_root
		endHelper.parent = midHelper
		midHelper.parent = startHelper

		--
		--	Knee Lookat with constraint system
		--
		local midVec =  normalize ( normalize(posMid - posStart) + normalize(posMid - posEnd) )
		local posMidLookat = ( midVec * (distance posStart posEnd) * 1.0) + posMid
		
		local midLookat = SSR_rig.makeHelper posMidLookat (boneSize*4) (prefix + midName + "_IK_lookat") box:true
		--in coordsys world move midLookat (normalize((posMid-posStart)-(posEnd-posMid))*IKlength*0.5)
		midLookat.parent = IK_root
		
		local midNull = SSR_rig.nullTransform midLookat
		
		--	Helpers parented to start and end
		local midStart = SSR_rig.makeHelper midLookat.pos (boneSize*2) (prefix + midName + "Constraint_" + startName) box:true
		midStart.parent = IK_root
		local midEnd = SSR_rig.makeHelper midLookat.pos (boneSize*1.5) (prefix + midName + "Constraint_" + endName) box:true
		midEnd.parent = IK_target
		local midStartEnd = SSR_rig.makeHelper midLookat.pos (boneSize*1.5) (prefix + midName + "Constraint_" + startName + endName) box:true
		midStartEnd.parent = IK_target
		local midWorld = SSR_rig.makeHelper midLookat.pos (boneSize*1.5) (prefix + midName + "Constraint_World") box:true
		
		local midLookCon = SSR_rig.posConTwo midStartEnd midEnd midStart
		
		local midWorldCon = SSR_rig.posConTwo midNull midStartEnd midWorld
		--SSR_rig.posCon midNull midLookat -- Causes 3ds Max to crash! Ehrm.. why was it here in the first place?

		--	Create IK system
		local IKSystem = IKSys.ikChain startHelper endHelper "IKLimb"
		IKSystem.parent = IK_target
		IKSystem.transform.controller.goalSize = boneSize * 2
		IKSystem.transform.controller.SAParent = 1
		IKSystem.name = prefix + "_IKChain"
		
		IKSystem.transform.controller.VHTarget = midLookat
		IKSystem.transform.controller.useVHTarget = true
		
		--
		--	Create soft stretchy system
		--
		if debug == true then SSR_gen.verbose ("midHelper: " + midHelper.Position.controller.X_position as string)
		if debug == true then SSR_gen.verbose ("endHelper: " + endHelper.Position.controller.X_position as string)
		
		-- soft IK
		local exposeControl = exposeTM pos:posStart size:(boneSize*1.5) name:(prefix + "IK_exposeControl")
		exposeControl.parent = IK_root
		exposeControl.exposeNode = IK_target
		exposeControl.useParent = false
		exposeControl.localReferenceNode = IK_root
		
		local softLookat = SSR_rig.makeHelper posStart boneSize (prefix + "IK_softLookat") axisTripod:true
		local lookCon = SSR_rig.lookCon softLookat IK_target
		lookCon.StoUp_axis = 1
		lookCon.upnode_axis = 1
		softLookat.parent = IK_root
		
		local softHelper = SSR_rig.makeHelper posEnd boneSize (prefix + "IK_softHelper") box:true
		SSR_rig.objAlign softHelper softLookat
		softHelper.parent = softLookat
		
		local limit = float_limit()
		softHelper.pos.controller.X_Position.controller = limit
		limit.SetLimitedControl exposeControl.distance.controller
		limit.upper_limit = IKLength -- Smooth will stop at total chain length
		limit.upper_smoothing = 0.0 -- Smooth falloff
		
		local exposeSoft = exposeTM pos:posStart size:(boneSize*1.8) name:(prefix + "IK_exposeSoft")
		exposeSoft.parent = exposeControl
		exposeSoft.exposeNode = IK_target
		exposeSoft.useParent = false
		exposeSoft.localReferenceNode = softHelper
		
		-- With some math, the leg will extend to fill the gap caused by the soft IK pulling
		--! At the moment, this is causing a weird snap with the IKHI controller. Why?
		local startStretchCon = Float_Expression ()
		midHelper.Position.controller.X_position.controller = startStretchCon
		startStretchCon.AddScalarConstant "len" startLength
		startStretchCon.AddScalarTarget "soft" exposeSoft.distance.controller
		startStretchCon.SetExpression "(soft*0.50)+len"

		local endStretchCon = Float_Expression ()
		endHelper.Position.controller.X_position.controller = endStretchCon
		endStretchCon.AddScalarConstant "len" endLength
		endStretchCon.AddScalarTarget "soft" exposeSoft.distance.controller
		endStretchCon.SetExpression "(soft*0.50)+len"
		
		--
		--	Create stretchy IK bones
		--
		local midMover = SSR_rig.makeHelper posMid (boneSize*4) (prefix + midName + "_IK_stretchHelper") box:true
		SSR_rig.objAlign midMover tBoneEnd
		midMover.parent = midHelper
		
		local boneStart = SSR_rig.muscleBone prefix startName posStart posMid boneSize:boneSize squashBone:false baseParent:startHelper endParent:midMover zAxis:zAxis
		local boneEnd = SSR_rig.muscleBone prefix endName posMid posEnd boneSize:boneSize squashBone:false baseParent:midMover endParent:endHelper zAxis:zAxis
		
		boneStart.wirecolor = boneEnd.wirecolor = boneColor
		
		boneEnd.rotation.controller.StoUp_axis = 1
		boneEnd.rotation.controller.StoUP_axisFlip = on
		boneEnd.rotation.controller.upnode_axis = 1
		
		delete tboneStart
		delete tboneEnd
		
		-- Create IK specific controllers, the calling function will do the rest
		-- Mid pin controller, will be used to offset the IK mid position
		local midPinControl = SSR_rig.crossCTL posmid (boneSize*10) alignAxis:"X"
		midPinControl.name = prefix + midName + "Pin_IK_CTL"
		SSR_rig.objAlign midPinControl boneEnd
		midPinControl.parent = midHelper
		midMover.parent = midPinControl
		setTransformLockFlags midPinControl #{4..9}
		SSR_rig.resetRotation midPinControl -- !Needs this to avoid flipping the bones. Investigate sometime...
		
		-- mid lookat controller, will be used as the IK lookat target
		local midLookatControl = SSR_rig.crossCTL midLookat.pos (boneSize*10) alignAxis:"X"
		midLookatControl.name = prefix + midName + "_IK_CTL"
		--SSR_rig.objAlign midLookatControl boneStart
		midLookatControl.parent = midNull
		midLookat.parent = midLookatControl
		setTransformLockFlags midLookatControl #{4..9}
		
		-----------------------------------------------------------------------------
		--
		--		IK Mid Controllers custom attributes
		--
		-----------------------------------------------------------------------------

		-- Mid Pin Controller
		local ca = attributes transforms
		(
			parameters params rollout:extraTrans
			(
				X_Pos		type:#float ui:PXSp
				Y_Pos		type:#float ui:PYSp
				Z_Pos		type:#float ui:PZSp
			)
			rollout extraTrans "Transforms"
			(
				spinner PXSp 	"Translate X " 	range:[-1000000, 1000000, 0] align:#right
				spinner PYSp 	"Translate Y " 	range:[-1000000, 1000000, 0] align:#right
				spinner PZSp 	"Translate Z " 	range:[-1000000, 1000000, 0] align:#right
			)
		)
		
		local penMod_KP = 'PEN_Attribute_Holder 2'()
		addModifier midPinControl penMod_KP
		custAttributes.add penMod_KP ca
		
		penMod_KP.transforms.X_Pos.controller = \
			midPinControl.position.controller.X_position.controller
		penMod_KP.transforms.Y_Pos.controller = \
			midPinControl.position.controller.Y_position.controller
		penMod_KP.transforms.Z_Pos.controller = \
			midPinControl.position.controller.Z_position.controller

		-- Mid Lookat Controllers attributes
		local ca_K_IK = attributes transforms
		(
			parameters params3 rollout:midR
			(
				X_Pos 		type:#float ui:kPXSp
				Y_Pos		type:#float ui:kPYSp
				Z_Pos		type:#float ui:kPZSp
			)
			parameters params4 rollout:conR
			(
				mid_Hip_Foot 	type:#float ui:(legHipSp, legHipSl)
				mid_World		type:#float ui:(worldSp, worldSl)
			)
			rollout midR "Joint Transforms"
			(
				spinner kPXSp 	"Translate X " 	range:[-1000000, 1000000, 0]
				spinner kPYSp 	"Translate Y " 	range:[-1000000, 1000000, 0]
				spinner kPZSp 	"Translate Z " 	range:[-1000000, 1000000, 0]
			)
			rollout conR "Joint Constraints"
			(
				label lbl01 		"Joint follows Start <> End" align:#centre offset:[0,5]
				spinner legHipSp	"" range:[-100, 100, 0] fieldwidth:40 align:#left offset:[10, 10] across:2
				slider legHipSl 	"" range:[-100, 100, 0]
				
				label lbl02 		"Force joint to world" align:#centre offset:[0,5]
				spinner worldSp		"" range:[0, 100, 0] fieldwidth:40 align:#left offset:[10, 10] across:2
				slider worldSl 		"" range:[0, 100, 0]

			)
		)	
		
		penMod_K_IK = 'PEN_Attribute_Holder 2'()
		addModifier midLookatControl penMod_K_IK
		custAttributes.add penMod_K_IK ca_K_IK
		
		penMod_K_IK.transforms.X_Pos.controller = midLookatControl.pos.controller.X_Position.controller
		penMod_K_IK.transforms.Y_Pos.controller = midLookatControl.pos.controller.Y_Position.controller
		penMod_K_IK.transforms.Z_Pos.controller = midLookatControl.pos.controller.Z_Position.controller

		paramWire.connect penMod_K_IK.transforms[#mid_Hip_Foot] \
			midStartEnd.position.controller[#Position_Weight_0] \
			"(mid_Hip_Foot*0.5)+50"
		paramWire.connect penMod_K_IK.transforms[#mid_Hip_Foot] \
			midStartEnd.position.controller[#Position_Weight_1] \
			"50-(mid_Hip_Foot*0.5)"
		
		paramWire.connect penMod_K_IK.transforms[#mid_world] \
			midNull.position.controller[#Position_Weight_0] \
			"100-mid_world"
		paramWire.connect penMod_K_IK.transforms[#mid_world] \
			midNull.position.controller[#Position_Weight_1] \
			"mid_world"	
		
		-- Returns an array of skinBones, topNodes, limitController, lookatHelper
		return #( #(midPinControl, midLookatControl), #(boneStart, boneEnd), #(IK_root, IK_target, midMover), limit, #(midLookat, midStartEnd, midNull, midStart) )
	),

	-----------------------------------------------------------------------------
	--
	-- FKSetup: 
	--	creates a stretchy FK chain, great for arms and legs
	--
	-----------------------------------------------------------------------------
	
	-- SSR_rig.FKSetup "Jack" "uparm" "elbow" "lowarm" $Locator_Arm_L_Shoulder.pos $Locator_Arm_L_Elbow.pos $Locator_Arm_L_Hand.pos flipSide:true
	function FKSetup prefix startName midName endName posStart posMid posEnd boneSize:1 zAxis:undefined boneColor:[128,255,128] flipSide:false debug:false =
	(
		local startLength = distance posStart posMid
		local endLength = distance posMid posEnd
		
		if debug == true then SSR_gen.verbose ("start: " + startLength as string)
		if debug == true then SSR_gen.verbose ("end: " + endLength as string)
		
		-- Our FK bone chain 
		if zAxis == undefined then zAxis = normalize ( cross (posStart-posMid) (posEnd-posMid) )
		
		local boneStart = SSR_rig.makeBone (prefix + "_" + startName + "_skinBone") posStart posMid zAxis boneSize
		local boneEnd 	= SSR_rig.makeBone (prefix + "_" + endName + "_skinBone") posMid posEnd zAxis boneSize
		local boneEndNub = SSR_rig.makeBone (prefix + "_" + endName + "_FK_endBone") posEnd (posEnd + [boneSize,0,0]) zAxis boneSize
		
		boneStart.wirecolor = boneEnd.wirecolor = boneColor
		
		SSR_rig.objAlign boneEndNub boneEnd
		boneEnd.parent = boneStart
		boneEndNub.parent = boneEnd
		
		--	Helpers that will do the stretching
		local startHelper = SSR_rig.makeHelper posMid (boneSize*2) (prefix + startName + "_FK_stretchHelper") cross:true
		local endHelper = SSR_rig.makeHelper posEnd (boneSize*2) (prefix + endName + "_FK_stretchHelper") cross:true
		SSR_rig.objAlign startHelper boneStart
		SSR_rig.objAlign endHelper boneEnd
		
		startNull = SSR_rig.nullTransform startHelper
		endNull = SSR_rig.nullTransform endHelper
		startNull.parent = boneStart
		endNull.parent = boneEnd
		
		SSR_rig.posCon boneEnd startHelper
		SSR_rig.posCon boneEndNub endHelper
		
		-- Create FK specific controllers
		-- Can be flipped on reverse side of character (still not sure if thats a good idea...)
		local startControl = SSR_rig.ringCTL posStart (boneSize*10) alignAxis:"X"
		startControl.name = prefix + "_" + startName + "_FK_CTL"
		SSR_rig.objAlign startControl boneStart
		if flipSide == true then in coordsys local rotate startControl (angleaxis 180 [0,0,-1])
		setTransformLockFlags startControl #{1..3,7..9}
		startConNull = SSR_rig.nullTransform startControl
		
		local endControl = SSR_rig.ringCTL posMid (boneSize*10) alignAxis:"X"
		endControl.name = prefix + "_" + endName + "_FK_CTL"
		SSR_rig.objAlign endControl boneEnd
		--if flipSide == true then in coordsys local rotate endControl (angleaxis 180 [0,0,-1])
		setTransformLockFlags endControl #{1..3,7..9}
		endConNull = SSR_rig.nullTransform endControl
		
		endConNull.parent = startHelper
		
		boneStart.parent = startControl
		--SSR_rig.rotCon boneStart startControl
		SSR_rig.rotCon boneEnd endControl
		
		-- Attributes
		SSR_rig.createRotAttributes startControl
		SSR_rig.createRotAttributes endControl
		
		-- Stretch attributes
		-- Attributes
		local ca = attributes stretch
		(
			parameters params rollout:extraTrans
			(
				strLength		type:#float ui:StrSp
			)
			rollout extraTrans "Bone Stretch"
			(
				spinner StrSp 	"Stretch Length " 	range:[-1000000, 1000000, 0] align:#right
			)
		)
		
		custAttributes.add startControl.modifiers[1] ca
		custAttributes.add endControl.modifiers[1] ca
		
		-- Attribute Connections
		startControl.modifiers[1].stretch.strLength.controller = \
			startHelper.position.controller.X_position.controller
		endControl.modifiers[1].stretch.strLength.controller = \
			endHelper.position.controller.X_position.controller

		return #(#(startControl, endControl), #(boneStart, boneEnd))
	),
	
	----------------------------------------------------------
	-- createFKController: 
	-- Applies a FK rotation controller to a single bone
	----------------------------------------------------------
	function createFKController charName rigBone limbName doCon:true =
	(
		if rigBone.parent == undefined then
		(
			SSR_gen.verbose "Bone must be parented!"
			return false
		)
		
		local topParent = rigBone.parent
		local prefix = charName + "_" + limbName + "_"
		
		-- Controller
		local rotControl = ringCTL rigBone.pos (rigBone.width*6) alignAxis:"X"
		rotControl.name = prefix + "CTL"
		objAlign rotControl rigBone
		if doCon == true then setTransformLockFlags rotControl #{7..9}
		else setTransformLockFlags rotControl #{1..3,7..9}
		
		rotControl.parent = topParent
		rigBone.parent = rotControl
		
		-- World/Local constraints
		if doCon == true then
		(
			local rotWorld = SSR_rig.makeHelper (rigBone.pos) (rigBone.width) (prefix + "World") box:true
			local rotLocal = SSR_rig.makeHelper (rigBone.pos) (rigBone.width*1.3) (prefix + "Local") box:true
			
			SSR_rig.objAlign rotWorld rotControl
			SSR_rig.objAlign rotLocal rotControl
			
			local rotNull = SSR_rig.makeHelper (rigBone.pos) (rigBone.width) (prefix + "Constraint") cross:true
			objAlign rotNull rotControl
			
			rotLocal.parent = topParent
			rotNull.parent = topParent	

			makeRotCon = SSR_rig.rotConTwo rotNull rotLocal rotWorld
			rotControl.parent = rotNull
		)
			
		-- Attributes
		if doCon == true then
		(
			local ca = attributes transforms
			(
				parameters params rollout:extraTrans
				(
					X_Pos		type:#float ui:PXSp
					Y_Pos		type:#float ui:PYSp
					Z_Pos		type:#float ui:PZSp
					X_Rot		type:#angle ui:RXSp
					Y_Rot		type:#angle ui:RYSp
					Z_Rot		type:#angle ui:RZSp
					Rot_WL 		type:#float ui:(RotSp, RotSl)
				)
				rollout extraTrans "Transforms"
				(
					spinner PXSp 	"Translate X " 	range:[-1000000, 1000000, 0] align:#right
					spinner PYSp 	"Translate Y " 	range:[-1000000, 1000000, 0] align:#right
					spinner PZSp 	"Translate Z " 	range:[-1000000, 1000000, 0] align:#right
					spinner RXSp 	"Rotate X " 	range:[-1000000, 1000000, 0] align:#right
					spinner RYSp 	"Rotate Y " 	range:[-1000000, 1000000, 0] align:#right
					spinner RZSp 	"Rotate Z " 	range:[-1000000, 1000000, 0] align:#right
					label lbl02 	"Orient to Local <> World" align:#centre offset:[0,10]
					spinner RotSp 	"" range:[0, 100, 0] fieldwidth:40 align:#left offset:[10, 10] across:2
					slider RotSl 	"" range:[0, 100, 0]
				)
			)
		) else (
			local ca = attributes transforms
			(
				parameters params rollout:extraTrans
				(
					X_Rot		type:#angle ui:RXSp
					Y_Rot		type:#angle ui:RYSp
					Z_Rot		type:#angle ui:RZSp
				)
				rollout extraTrans "Transforms"
				(
					spinner RXSp 	"Rotate X " 	range:[-1000000, 1000000, 0] align:#right
					spinner RYSp 	"Rotate Y " 	range:[-1000000, 1000000, 0] align:#right
					spinner RZSp 	"Rotate Z " 	range:[-1000000, 1000000, 0] align:#right
				)
			)
		) -- if doCon == true 
		
		local penMod = 'PEN_Attribute_Holder 2'()
		addModifier rotControl penMod
		custAttributes.add penMod ca
		
		-- Attribute Connections
		penMod.transforms.X_Rot.controller = \
			rotControl.rotation.controller.X_rotation.controller
		penMod.transforms.Y_Rot.controller = \
			rotControl.rotation.controller.Y_rotation.controller
		penMod.transforms.Z_Rot.controller = \
			rotControl.rotation.controller.Z_rotation.controller

		if doCon == true then (
			penMod.transforms.X_Pos.controller = \
				rotControl.position.controller.X_position.controller
			penMod.transforms.Y_Pos.controller = \
				rotControl.position.controller.Y_position.controller
			penMod.transforms.Z_Pos.controller = \
				rotControl.position.controller.Z_position.controller
			
			paramWire.connect penMod.transforms[#Rot_WL] \
				rotNull.rotation.controller[#Orientation_Weight_0] \
				"100-Rot_WL"
			paramWire.connect penMod.transforms[#Rot_WL] \
				rotNull.rotation.controller[#Orientation_Weight_1] \
				"Rot_WL"
		)
			
		return rotControl
	),
	
	----------------------------------------------------------
	-- createPosAttributes: 
	-- Adds position transform attributes to an object
	----------------------------------------------------------
	function createPosAttributes controller =
	(
		-- Attributes
		local ca = attributes transforms
		(
			parameters params rollout:extraTrans
			(
				X_Pos		type:#float ui:PXSp
				Y_Pos		type:#float ui:PYSp
				Z_Pos		type:#float ui:PZSp
			)
			rollout extraTrans "Transforms"
			(
				spinner PXSp 	"Translate X " 	range:[-1000000, 1000000, 0] align:#right
				spinner PYSp 	"Translate Y " 	range:[-1000000, 1000000, 0] align:#right
				spinner PZSp 	"Translate Z " 	range:[-1000000, 1000000, 0] align:#right
			)
		)
		
		local penMod = 'PEN_Attribute_Holder 2'()
		addModifier controller penMod
		custAttributes.add penMod ca
		
		-- Attribute Connections
		penMod.transforms.X_Pos.controller = \
			controller.position.controller.X_position.controller
		penMod.transforms.Y_Pos.controller = \
			controller.position.controller.Y_position.controller
		penMod.transforms.Z_Pos.controller = \
			controller.position.controller.Z_position.controller
		
		return true
	),
	
	----------------------------------------------------------
	-- createRotAttributes: 
	-- Adds rotation attributes to an object
	----------------------------------------------------------
	function createRotAttributes controller =
	(
		-- Attributes
		local ca = attributes transforms
		(
			parameters params rollout:extraTrans
			(
				X_Rot		type:#angle ui:RXSp
				Y_Rot		type:#angle ui:RYSp
				Z_Rot		type:#angle ui:RZSp
			)
			rollout extraTrans "Transforms"
			(
				spinner RXSp 	"Rotate X " 	range:[-1000000, 1000000, 0] align:#right
				spinner RYSp 	"Rotate Y " 	range:[-1000000, 1000000, 0] align:#right
				spinner RZSp 	"Rotate Z " 	range:[-1000000, 1000000, 0] align:#right
			)
		)
		
		local penMod = 'PEN_Attribute_Holder 2'()
		addModifier controller penMod
		custAttributes.add penMod ca
		
		-- Attribute Connections
		penMod.transforms.X_Rot.controller = \
			controller.rotation.controller.X_rotation.controller
		penMod.transforms.Y_Rot.controller = \
			controller.rotation.controller.Y_rotation.controller
		penMod.transforms.Z_Rot.controller = \
			controller.rotation.controller.Z_rotation.controller
		
		return true
	),
	
	----------------------------------------------------------
	-- createAllAttributes: 
	-- Adds all transform attributes to an object
	----------------------------------------------------------
	function createAllAttributes controller =
	(
		-- Attributes
		local ca = attributes transforms
		(
			parameters params rollout:extraTrans
			(
				X_Pos		type:#float ui:PXSp
				Y_Pos		type:#float ui:PYSp
				Z_Pos		type:#float ui:PZSp
				X_Rot		type:#angle ui:RXSp
				Y_Rot		type:#angle ui:RYSp
				Z_Rot		type:#angle ui:RZSp
				X_Sca		type:#float ui:RXSc
				Y_Sca		type:#float ui:RYSc
				Z_Sca		type:#float ui:RZSc
			)
			rollout extraTrans "Transforms"
			(
				spinner PXSp 	"Translate X " 	range:[-1000000, 1000000, 0] align:#right
				spinner PYSp 	"Translate Y " 	range:[-1000000, 1000000, 0] align:#right
				spinner PZSp 	"Translate Z " 	range:[-1000000, 1000000, 0] align:#right
				spinner RXSp 	"Rotate X " 	range:[-1000000, 1000000, 0] align:#right
				spinner RYSp 	"Rotate Y " 	range:[-1000000, 1000000, 0] align:#right
				spinner RZSp 	"Rotate Z " 	range:[-1000000, 1000000, 0] align:#right
				spinner RXSc 	"Scale X " 	range:[-1000000, 1000000, 1.0] align:#right offset:[0,10]
				spinner RYSc 	"Scale Y " 	range:[-1000000, 1000000, 1.0] align:#right
				spinner RZSc 	"Scale Z " 	range:[-1000000, 1000000, 1.0] align:#right
				
			)
		)
		
		local penMod = 'PEN_Attribute_Holder 2'()
		addModifier controller penMod
		custAttributes.add penMod ca
		
		-- Attribute Connections
		penMod.transforms.X_Pos.controller = \
			controller.position.controller.X_position.controller
		penMod.transforms.Y_Pos.controller = \
			controller.position.controller.Y_position.controller
		penMod.transforms.Z_Pos.controller = \
			controller.position.controller.Z_position.controller
		
		penMod.transforms.X_Rot.controller = \
			controller.rotation.controller.X_rotation.controller
		penMod.transforms.Y_Rot.controller = \
			controller.rotation.controller.Y_rotation.controller
		penMod.transforms.Z_Rot.controller = \
			controller.rotation.controller.Z_rotation.controller
		
		controller.scale.controller = ScaleXYZ ()
		
		penMod.transforms.X_Sca.controller = \
			controller.scale.controller.X_scale.controller
		penMod.transforms.Y_Sca.controller = \
			controller.scale.controller.Y_scale.controller
		penMod.transforms.Z_Sca.controller = \
			controller.scale.controller.Z_scale.controller
		
		return true
	),

	
	-----------------------------------------------------------------------------
	--	halfBone: 
	--	creates a bone that is halfway between two other bones
	--	reverseTwo: new bone positioned at end of targetTwo, otherwise positioned at targetTwo root
	--	inverted:
	-----------------------------------------------------------------------------
	function halfBone charName prefix targetOne targetTwo boneSize factor:50.0 reverseTwo:false inverted:false alignSecond:false =
	(
		-- End points for the bones will be needed for our math
		local oneEndpos = ((transMatrix [targetOne.length,0,0])*targetOne.transform).translationPart
		local twoEndpos = ((transMatrix [targetTwo.length,0,0])*targetTwo.transform).translationPart
		
		-- Normalized bone vectors
		local oneVector = normalize (oneEndpos - targetOne.pos)
		local twoVector = normalize (twoEndpos - targetTwo.pos)
		
		-- Bone root will be at the start of the targetTwo bone, which would normally be the connection
		-- point between the two bones. If this bone is reversed, we will have to find a new bone position
		local bonePos = targetTwo.pos
		if reverseTwo == true then bonePos = twoEndpos
		local halfPoint = point pos:bonePos size:boneSize name:(charName + "_" + prefix + "_halfLock")
		
		local halfCon = Orientation_Constraint()
		halfCon.appendTarget targetOne factor
		halfCon.appendTarget targetTwo (100.0-factor)
		halfPoint.rotation.controller = halfCon
		
		-- Parent this helper to the second bone
		halfPoint.parent = targetTwo
		
		-- Then create the skinBone itself
		zAxis = cross oneVector twoVector
		local endPos = undefined
		if alignSecond == true then
		(
			if reverseTwo == true then endPos = targetTwo.pos
			else endPos = twoEndPos	
		) else (
			endPos = bonePos + ( (oneVector - twoVector) * 10.0 )
		)
		
		local halfBone = makeBone (charName + "_" + prefix + "_skinBone") bonePos endPos zAxis boneSize
		halfBone.parent = halfPoint
		
		local halfEnd = makeBone (charName + "_" + prefix + "_boneEnd") endPos [endPos.x, endPos.y, endPos.z + boneSize] zAxis boneSize
		objAlign halfEnd halfBone
		halfEnd.parent = halfBone
		
		-- Put the created objects into respective layers
		local skinLayerName = charName + " Skinning Bones"
		local skinLayer = layermanager.getlayerfromname skinLayerName
		skinLayer.addNode halfBone
		
		local rigLayerName = charName + " Rig"
		local rigLayer = layermanager.getlayerfromname rigLayerName
		rigLayer.addNode halfPoint
		rigLayer.addNode halfEnd
		
		return halfPoint
	),

	-----------------------------------------------------------------------------
	--	muscleBone: 
	--	creates a stretchable bone between two helpers
	-----------------------------------------------------------------------------
	function muscleBone prefix boneName basePos endPos boneSize:1 squashBone:true baseParent:undefined endParent:undefined zAxis:false debug:false =
	(
		-- Figure out the best bone zAxis. If the initial bone direction is close to the y axis, we will set
		-- the bone zAxis to along the x axis.
		if zAxis == false then
		(	
			zAxis = [1,0,0]
			local boneDir = normalize ( endPos - basePos )
			local zAngle = acos (dot zAxis boneDir)
			
			if zAngle < 30 or zAngle > 150 then 
			(
				zAxis = [0,1,0]
				if debug == true then SSR_gen.verbose ("zAxis along Y.")
			)
		)
			
		-- Create the skinBone and endBone
		local stretchBone = makeBone (prefix + boneName + "_skinBone") basePos endPos zAxis boneSize
		if squashBone == true then stretchBone.boneScaleType = #squash
		
		local endBone = makeBone (prefix + boneName + "_endBone") endPos (endPos+[0,0,boneSize]) zAxis boneSize
		objAlign endBone stretchBone
		endBone.parent = stretchBone
		
		-- Helpers that will take care of the constraining
		local baseHelper = baseParent
		if baseHelper == undefined then 
		(
			baseHelper = makeHelper basePos boneSize (prefix + boneName + "_baseHelper") box:true
			objAlign baseHelper stretchBone
		)
		
		local endHelper = endParent
		if endHelper == undefined then 
		(
			endHelper = makeHelper endPos boneSize (prefix + boneName + "_endHelper") box:true
			objAlign endHelper stretchBone
		)
		
		-- Constraints and final parenting
		stretchBone.parent = baseHelper
		local lookController = lookCon stretchBone endHelper
		posCon endBone endHelper
		lookController.upnode_world = off
		lookController.pickUpNode = baseHelper
		
		if debug == true then 
		(
			SSR_gen.verbose ("baseParent: " + baseParent as string)
			SSR_gen.verbose ("endParent: " + endParent as string)
		)
		
		-- Layer assignment
		try
		(
			local rigLayerName = prefix + " Rig"
			local rigLayer = layermanager.getlayerfromname rigLayerName
			local skinLayerName = prefix + " Skinning Bones"
			local skinLayer = layermanager.getlayerfromname skinLayerName
			
			skinLayer.addNode stretchBone
			rigLayer.addNode endBone
			rigLayer.addNode baseHelper
			rigLayer.addNode endHelper
		)
		catch()
		
		return stretchBone
	),

	-----------------------------------------------------------------------------
	--
	--	stretchyChain: 
	--	Makes a chain of bones into a stretchy chain using constraints.
	--
	-----------------------------------------------------------------------------
	function stretchyChain boneList helperSize:1 alignHelpers:true doParenting:true = 
	(
		local helperObjs = #()
		
		for o in boneList do
		(
			curHelper = SSR_rig.makeHelper o.pos helperSize (o.name + "_stretchHelper") box:true
			if alignHelpers == true then SSR_rig.objAlign curHelper o
			append helperObjs curHelper
		)
		
		for i = 1 to boneList.count do
		(
			if i > 1 and doParenting == true then helperObjs[i].parent = helperObjs[i-1]
				
			SSR_rig.posCon boneList[i] helperObjs[i]
			
			if i < boneList.count then
			(
				stretchLookat = SSR_rig.lookCon boneList[i] helperObjs[i+1]
				
				stretchLookat.pickUpNode = helperObjs[i]
				stretchLookat.lookat_vector_length = 0
				stretchLookat.upnode_world = false
			)
		)
		
		return helperObjs
	),

	-----------------------------------------------------------------------------
	--
	--	cleanName: 
	--	cleans the character name by removing underscores and spaces
	--
	-----------------------------------------------------------------------------
	function cleanName charName =
	(
		charName = substituteString charName "_" "" 
		charName = substituteString charName " " "" 
		charName = substituteString charName ":" "" 
		
		return charName
	),

	-----------------------------------------------------------------------------
	--
	--	deleteLayer: 
	--	deletes layer and the objects it contains
	--
	-----------------------------------------------------------------------------
	function deleteLayer layerName =
	(
		local rigLayerName = layerName
		local rigLayer = layerManager.getlayerfromname rigLayerName
		if rigLayer != undefined then
		(
			rigLayer.nodes &rigNodes
			theNodes = rigNodes as array
			for o in theNodes do 
			( 
				try (delete o)
				catch()
			)
			layerManager.deleteLayerByName rigLayerName
		) else ( 
			SSR_gen.verbose ("No layer found: " + rigLayerName )
		)
		return true
	),

	-----------------------------------------------------------------------------
	--
	--	hideLayer: 
	--	hide a named layer
	--
	-----------------------------------------------------------------------------
	function hideLayer layerName =
	(
		local theLayer = layerManager.getlayerfromname layerName
		theLayer.isHidden = true
	),
	
	
	-----------------------------------------------------------------------------
	--
	--	deleteRig: 
	--	deletes the entire character rig and all its layers
	--
	-----------------------------------------------------------------------------
	function deleteRig charName =
	(
		SSR_gen.verbose ("Deleting all rig components belonging to: " + charName + "")
		
		SSR_rig.deleteLayer (charName + " Skinning Bones")
		SSR_rig.deleteLayer (charName + " Rig")
		SSR_rig.deleteLayer (charName + " Controllers")
		
		return true
	),

	-----------------------------------------------------------------------------
	--
	--	Functions to upgrade the rig from pre 1.0-versions
	--
	-----------------------------------------------------------------------------
	function upgradeSkinBones charName debug:true =
	(
		-- Old names											--New Names
		boneNames = #( \
		"_L_Index_skinBone_A",						"_L_IndexFinger_skinBone_A", \
		"_L_Index_skinBone_B",						"_L_IndexFinger_skinBone_B", \
		"_L_Index_skinBone_C",						"_L_IndexFinger_skinBone_C", \
		"_L_leg_skinBone_Ankle",					"_L_leg_skinBone_Foot", \
		"_L_Middle_skinBone_A",					"_L_MiddleFinger_skinBone_A", \
		"_L_Middle_skinBone_B",					"_L_MiddleFinger_skinBone_B", \
		"_L_Middle_skinBone_C",					"_L_MiddleFinger_skinBone_C", \
		"_L_Pinkie_skinBone_A",						"_L_PinkieFinger_skinBone_A", \
		"_L_Pinkie_skinBone_B",						"_L_PinkieFinger_skinBone_B", \
		"_L_Pinkie_skinBone_C",						"_L_PinkieFinger_skinBone_C", \
		"_L_Ring_skinBone_A",						"_L_RingFinger_skinBone_A", \
		"_L_Ring_skinBone_B",						"_L_RingFinger_skinBone_B", \
		"_L_Ring_skinBone_C",						"_L_RingFinger_skinBone_C", \
		"_L_Thumb_skinBone_A",					"_L_ThumbFinger_stretch_skinBone", \
		"_L_Thumb_skinBone_B",					"_L_ThumbFinger_skinBone_A", \
		"_L_Thumb_skinBone_C",					"_L_ThumbFinger_skinBone_B", \
		"_R_Index_skinBone_A",						"_R_IndexFinger_skinBone_A", \
		"_R_Index_skinBone_B",						"_R_IndexFinger_skinBone_B", \
		"_R_Index_skinBone_C",						"_R_IndexFinger_skinBone_C", \
		"_R_leg_skinBone_Ankle",					"_R_leg_skinBone_Foot", \
		"_R_Middle_skinBone_A",					"_R_MiddleFinger_skinBone_A", \
		"_R_Middle_skinBone_B",					"_R_MiddleFinger_skinBone_B", \
		"_R_Middle_skinBone_C",					"_R_MiddleFinger_skinBone_C", \
		"_R_Pinkie_skinBone_A",					"_R_PinkieFinger_skinBone_A", \
		"_R_Pinkie_skinBone_B",						"_R_PinkieFinger_skinBone_B", \
		"_R_Pinkie_skinBone_C",					"_R_PinkieFinger_skinBone_C", \
		"_R_Ring_skinBone_A",						"_R_RingFinger_skinBone_A", \
		"_R_Ring_skinBone_B",						"_R_RingFinger_skinBone_B", \
		"_R_Ring_skinBone_C",						"_R_RingFinger_skinBone_C", \
		"_R_Thumb_skinBone_A",					"_R_ThumbFinger_stretch_skinBone", \
		"_R_Thumb_skinBone_B",					"_R_ThumbFinger_skinBone_A", \
		"_R_Thumb_skinBone_C",					"_R_ThumbFinger_skinBone_B" )
		
		-- Remember toes, goddamit!
		
		-- Rename skinBones
		if debug==true then SSR_gen.verbose ("boneNames.count: " + boneNames.count as string)
		
		for c = 1 to (boneNames.count/2) do
		(
			try (
				fromName = charName + boneNames[(c*2)-1]
				toName = charName + boneNames[(c*2)]

				renameObj = execute ("$'" + fromName + "'")
				renameObj.name = toName
			)
			catch SSR_gen.verbose ("Error renaming $'" + fromName + "'. Does the object exist?")
		)
		
		return true
	)
	
)

SSR_rig = SSR_rig()