-----------------------------------------------------------------------------
--
-- 	Functions needed by the animate/select user interface
--
--	v1.0:	First version, merge of different files
--
-----------------------------------------------------------------------------

global SSR_skin

struct SSR_skin
(	
	-- Forward declarations
	function parentClosest = (),
	function lineDist = (),
	function segDist = (),
	function buildMatList = (),
	function materialSplit = (),
	function splitByMaterial = (),
	function hardSkin = (),
	function cutObjects = (),
	
	-----------------------------------------------------------------------------
	--	parentClosest: 
	--	parents objects in one array to the closest one in a second array
	--	using distance to object centers as the comparison
	-----------------------------------------------------------------------------
	function parentClosest parentList closestList doNames:true =
	(
		for o = 1 to parentList.count do
		(
			local closestObj = closestList[1]
			for c in closestList do
			(
				-- Last distance
				local closestDistance = distance (parentList[o].center) (closestObj.center)
				local newDistance = distance (parentList[o].center) c.center
				
				if newDistance < closestDistance then closestObj = c
			)
			parentList[o].parent = closestObj
			
			if doNames == true then
			(
				--! Insert renaming scheme
			)
		)
		return true
	),
	
	-----------------------------------------------------------------------------
	--	lineDist: 
	--	finds the shortest distance from a point to a line
	-----------------------------------------------------------------------------
	function lineDist lPosA lPosB pos = 
	(
		local vectorAB = lPosB-lPosA
		local vectorAC = pos-lPosA
		return (length (cross vectorAB vectorAC))/(length vectorAB)
	),

	-----------------------------------------------------------------------------
	--	segDist: 
	--	finds the shortest distance from a point to a line segment
	-----------------------------------------------------------------------------

	function segDist lPosA lPosB pos =
	(
		local c = pos - lPosA	-- Vector from a to Point
		local v = normalize (lPosB - lPosA)	-- Unit Vector from a to b
		local d = distance lPosB lPosA	-- Length of the line segment
		local t = dot v c	-- Intersection point Distance from a

		-- Check to see if the point is on the line
		-- if not then return the endpoint
		if (t < 0) then return distance lPosA pos;
		else if(t > d) then return distance lPosB pos;
		else return lineDist lPosA lPosB pos
	),

	
	-----------------------------------------------------------------------------
	--	buildMatList 
	--	returns a list of material IDs and the polygons that belong to them
	-----------------------------------------------------------------------------
	function buildMatList splitMesh maxID:200 =
	(
		local IDList = #()
		local polysLeft = #()
		for c = 1 to (polyop.getNumFaces splitMesh) do append polysLeft c

		progressStart ("Building material ID list" + splitMesh.name + "...")
		
		-- Count through the MatIDs we are searching for
		for m = 1 to maxID do
		(
			progressUpdate (100.0 * (m as float) / (maxID as float))
			
			local curID = #()
			for p = polysLeft.count to 1 by -1 do
			(
				local thisID = polyOp.getFaceMatID splitMesh polysLeft[p]
				if thisID == m then 
				(
					append curID polysLeft[p]
					deleteItem polysLeft p	
				)
			)
			if curID.count > 0 then
			(
				insertItem m curID 1
				append IDList curID
			)
		)
		
		progressEnd()

		return IDList
		
	),

	-----------------------------------------------------------------------------
	--	materialSplit: 
	--	creates MatIDs based on bone distance.
	-----------------------------------------------------------------------------
	function materialSplit splitMesh bonesList debug:false =
	(
		-- Create poly-index holding array and a start-end position array, both based on the bones
		local polyList = #()
		local boneEndList = #()
		for c = 1 to bonesList.count do
		(
			polyList[c] = #()
			boneEndList[c] = ((transMatrix [bonesList[c].length,0,0])*bonesList[c].transform).translationPart	
			--sphere pos:(boneEndList[c])
		)
		
		local faceNum = polyop.getNumFaces splitMesh
		
		-- This operation can take a long time, so we need to be able to cancel
		escapeEnable = true
		local notCancelled = true
		progressStart ("Scanning " + splitMesh.name + "...")
		local startTime = timestamp() as float
		
		-- For each face, find distance to nearest bone
		for n = 1 to faceNum while notCancelled do
		(
			--Update progress bar, and check for cancellation
			notCancelled = progressUpdate (100.0 * n / faceNum)
			if (keyboard.escPressed) do ( print n; notCancelled = false ) 
			
			local faceCent = polyOp.getFaceCenter splitMesh n
			local closestBone = undefined
			local closestDist = undefined
			
			-- For each bone, check which one is closest to current poly
			for c = 1 to bonesList.count while notCancelled do
			(
				local newDist = segDist (bonesList[c].pos) (boneEndList[c]) faceCent
				
				-- Only store the closest bone
				if closestDist != undefined then
				(
					if newDist < closestDist then
					(
						closestBone = c
						closestDist = newDist
					)				
				) else (
					closestBone = c			
					closestDist = newDist
				)
			)
			
			-- Append this polygon to the array referring to the closest bone
			append polyList[closestBone] n
		)
		
		-- End progress bar
		progressEnd()
		escapeEnable = false
		if notCancelled == false then return false

		if debug == true then format "-- Finished scanning in % minutes\n" ((timestamp()-startTime)/60000.0)
		
		-- Go through each bone, processing its assigned polygons
		for c = 1 to polyList.count do
		(
			if polyList[c].count > 0 then
			(
				-- Assign increasing MatIDs to polygons
				polyOp.setFaceMatID splitMesh (polyList[c]) c
				if debug == true then redrawViews()
			)
		)
		return true
	),
	
	
	-----------------------------------------------------------------------------
	--	createSplitMaterial: 
	--	creates a multi-sub material with random colours for each MatID in an object
	-----------------------------------------------------------------------------
	function createSplitMaterial splitMesh bonesList charName debug:false =
	(
		local matIDList = buildMatList splitMesh

		-- Create material
		local splitMat = multiMaterial name:(splitMesh.name + "_multiMat")
		splitMat.materialList.count = matIDlist.count
		splitMesh.material = splitMat
		
		print ("NumMats: " + matIDlist.count as string)
		
		for m = 1 to matIDlist.count do
		(
			splitMat.materialIDList[m] = matIDList[m][1]
			splitMat.names[m] = substring bonesList[m].name (charName.count + 2) (bonesList[m].name.count)
			splitMat.materialList[m].Diffuse = color (random 64 255) (random 64 255) (random 64 255)
			--splitMat.materialList[m].Diffuse = bonesList[m].wirecolor 
		)
		
		return matIDList
	),
	
	-----------------------------------------------------------------------------
	--	splitByMaterial: 
	--	splits into separate objects based on materialIDs
	-----------------------------------------------------------------------------
	function splitByMaterial splitMesh bonesList charName =
	(
		local splitObjects = #()
		local matIDlist = buildMatList splitMesh

		progressStart ("Splitting " + splitMesh.name + "...")
		
		for m = 1 to matIDlist.count do
		(
			-- strip material index from IDList:
			deleteItem matIDlist[m] 1
			local splitObjName = charName + "_proxy" + m as string + "_" + (subString bonesList[m].name (charName.count + 2) bonesList[m].name.count)
			polyOp.detachFaces splitMesh (matIDlist[m]) delete:false asNode:true name:splitObjName
			
			progressUpdate (100.0 * (m as float) / (matIDlist.count as float))
			
			local splitObj = getNodeByName splitObjName
			splitObj.isHidden = false
			append splitObjects splitObj
		) 
		
		progressEnd()
		
		splitMesh.isHidden = true
		return splitObjects
	),
	
	
	function hardSkin charName =
	(
		-- Hard skin mesh
		
		
	),
	
	-----------------------------------------------------------------------------
	--	cutObjects: 
	--	splits character mesh into colours based on bone positions
	-----------------------------------------------------------------------------
	function cutObjects charName meshType:1 boneType:2 doPrep:true =
	(
		skinBones = #()
		skinMesh = #()
		
		-- boneType 1 is current selection, boneType 2 is skinBone layer
		if boneType == 1 then
		(
			for o in $selection do
				if classOf o == BoneGeometry then append skinBones o
		) else if boneType == 2 then
		(
			boneLayer = LayerManager.getLayerFromName (charName + " Skinning Bones")
			boneLayer.nodes &layerNodes
			for o in layerNodes do
				if classOf o == BoneGeometry then append skinBones o
		)
		
		-- skinMesh 1 is current selection, boneType 2 is "[Name] Render" layer
		if meshType == 1 then
		(
			for o in $selection do
				if classOf o == Editable_Poly then append skinMesh o
		) else if meshType == 2 then
		(
			charLayer = LayerManager.getLayerFromName (charName + " Render")
			charLayer.nodes &layerNodes
			for o in layerNodes do
				if classOf o == Editable_Poly then append skinMesh o
		)
		
		SSR_gen.verbose ("About to process " + skinMesh.count as string + " meshes and " + skinBones.count as string + " skinBones.")
		
		-- Exit if there are no bones or skin meshes
		if skinMesh.count == 0 then
		(
			SSR_gen.verbose ("Hey, there's nothing to skin!")
			return false
		)
		if skinBones.count == 0 then 
		(
			SSR_gen.verbose ("Hey, where's my bones!")
			return false
		)
		
		for o in skinMesh do
		(
			if doPrep == true then
			(
				materialSplit o skinBones debug:true
				createSplitMaterial o skinBones charName
			) else (
				-- Kill previous proxies
				delete ( execute ("$'" + charName + "_proxy*'") )
				
				splitObjects = splitByMaterial o skinBones charName
				parentClosest splitObjects skinBones doNames:true
			)
		)

		
		return true
	)
	
	
) -- struct SSR_skin

SSR_skin = SSR_skin()